Package: vetiver
Title: Version, Share, Deploy, and Monitor Models
Version: 0.2.7
Authors@R: c(
    person("Julia", "Silge", , "julia.silge@posit.co", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-3671-836X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: The goal of 'vetiver' is to provide fluent tooling to
    version, share, deploy, and monitor a trained model. Functions handle
    both recording and checking the model's input data prototype, and
    predicting from a remote API endpoint. The 'vetiver' package is
    extensible, with generics that can support many kinds of models.
License: MIT + file LICENSE
URL: https://vetiver.posit.co, https://rstudio.github.io/vetiver-r/,
        https://github.com/rstudio/vetiver-r/
BugReports: https://github.com/rstudio/vetiver-r/issues
Depends: R (>= 4.1)
Imports: bundle (>= 0.1.3), butcher (>= 0.4.0), cereal, cli, fs,
        generics, glue, hardhat, lifecycle, pins (>= 1.3.0), purrr,
        rapidoc, readr (>= 1.4.0), rlang (>= 1.1.0), tibble, vctrs,
        withr
Suggests: arrow, callr, caret, clustMixType, covr, curl, dplyr,
        flexdashboard, ggplot2, httpuv, httr, jsonlite, keras, knitr,
        LiblineaR, luz, mgcv, mlr3 (>= 0.17.0), mlr3data, mlr3learners,
        mockery, modeldata, parsnip, paws.machine.learning (>= 0.2.0),
        pingr, plotly, plumber (>= 1.0.0), probably (>= 1.2.0),
        quantregForest, ranger, recipes (>= 1.1.0), reticulate,
        rmarkdown, rpart, rsample, rsconnect, slider (>= 0.2.2),
        smdocker (>= 0.1.2), stacks, tensorflow, testthat (>= 3.1.8),
        tidyselect, torch, tune, vdiffr, workflows, xgboost, yardstick
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-12-11
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2025-12-12 20:39:29 UTC; juliasilge
Author: Julia Silge [cre, aut] (ORCID: <https://orcid.org/0000-0002-3671-836X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Julia Silge <julia.silge@posit.co>
Repository: CRAN
Date/Publication: 2025-12-13 06:10:02 UTC
Built: R 4.6.0; ; 2026-04-29 09:40:20 UTC; unix

Package: vetiver
Title: Version, Share, Deploy, and Monitor Models
Version: 0.2.8
Authors@R: c(
    person("Julia", "Silge", , "julia.silge@posit.co", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-3671-836X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: The goal of 'vetiver' is to provide fluent tooling to
    version, share, deploy, and monitor a trained model. Functions handle
    both recording and checking the model's input data prototype, and
    predicting from a remote API endpoint. The 'vetiver' package is
    extensible, with generics that can support many kinds of models.
License: MIT + file LICENSE
URL: https://vetiver.posit.co, https://rstudio.github.io/vetiver-r/,
        https://github.com/rstudio/vetiver-r/
BugReports: https://github.com/rstudio/vetiver-r/issues
Depends: R (>= 4.1)
Imports: bundle (>= 0.1.3), butcher (>= 0.4.0), cereal, cli, fs,
        generics, glue, hardhat, lifecycle, pins (>= 1.3.0), purrr,
        rapidoc, readr (>= 1.4.0), rlang (>= 1.1.0), tibble, vctrs,
        withr
Suggests: arrow, callr, caret, clustMixType, covr, curl, dplyr,
        flexdashboard, ggplot2, httpuv, httr, jsonlite, keras, knitr,
        LiblineaR, luz, mgcv, mlr3 (>= 0.17.0), mlr3data, mlr3learners,
        mockery, modeldata, parsnip, paws.machine.learning (>= 0.2.0),
        pingr, plotly, plumber (>= 1.0.0), probably (>= 1.2.0),
        quantregForest, ranger, recipes (>= 1.1.0), reticulate,
        rmarkdown, rpart, rsample, rsconnect, slider (>= 0.2.2),
        smdocker (>= 0.1.2), stacks, tensorflow, testthat (>= 3.1.8),
        tidyselect, torch, tune, vdiffr, workflows, xgboost, yardstick
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-12-11
Encoding: UTF-8
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-22 23:25:03 UTC; juliasilge
Author: Julia Silge [cre, aut] (ORCID: <https://orcid.org/0000-0002-3671-836X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Julia Silge <julia.silge@posit.co>
Repository: CRAN
Date/Publication: 2026-08-23 06:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(augment,vetiver_endpoint)
S3method(augment,vetiver_endpoint_sagemaker)
S3method(augment,vetiver_model)
S3method(format,vetiver_endpoint)
S3method(format,vetiver_endpoint_sagemaker)
S3method(format,vetiver_model)
S3method(glue_spec_summary,array)
S3method(glue_spec_summary,data.frame)
S3method(glue_spec_summary,default)
S3method(handler_predict,Learner)
S3method(handler_predict,default)
S3method(handler_predict,gam)
S3method(handler_predict,glm)
S3method(handler_predict,int_conformal_cv)
S3method(handler_predict,int_conformal_full)
S3method(handler_predict,int_conformal_quantile)
S3method(handler_predict,int_conformal_split)
S3method(handler_predict,keras.engine.training.Model)
S3method(handler_predict,kproto)
S3method(handler_predict,lm)
S3method(handler_predict,luz_module_fitted)
S3method(handler_predict,model_stack)
S3method(handler_predict,ranger)
S3method(handler_predict,recipe)
S3method(handler_predict,train)
S3method(handler_predict,workflow)
S3method(handler_predict,xgb.Booster)
S3method(handler_startup,Learner)
S3method(handler_startup,default)
S3method(handler_startup,gam)
S3method(handler_startup,int_conformal_cv)
S3method(handler_startup,int_conformal_full)
S3method(handler_startup,int_conformal_quantile)
S3method(handler_startup,int_conformal_split)
S3method(handler_startup,keras.engine.training.Model)
S3method(handler_startup,luz_module_fitted)
S3method(handler_startup,model_stack)
S3method(handler_startup,ranger)
S3method(handler_startup,recipe)
S3method(handler_startup,train)
S3method(handler_startup,workflow)
S3method(handler_startup,xgb.Booster)
S3method(map_request_body,array)
S3method(map_request_body,data.frame)
S3method(map_request_body,default)
S3method(predict,vetiver_endpoint)
S3method(predict,vetiver_endpoint_sagemaker)
S3method(predict,vetiver_model)
S3method(print,vetiver_endpoint)
S3method(print,vetiver_endpoint_sagemaker)
S3method(print,vetiver_model)
S3method(vetiver_create_description,Learner)
S3method(vetiver_create_description,default)
S3method(vetiver_create_description,gam)
S3method(vetiver_create_description,glm)
S3method(vetiver_create_description,int_conformal_cv)
S3method(vetiver_create_description,int_conformal_full)
S3method(vetiver_create_description,int_conformal_quantile)
S3method(vetiver_create_description,int_conformal_split)
S3method(vetiver_create_description,keras.engine.training.Model)
S3method(vetiver_create_description,kproto)
S3method(vetiver_create_description,lm)
S3method(vetiver_create_description,luz_module_fitted)
S3method(vetiver_create_description,model_stack)
S3method(vetiver_create_description,ranger)
S3method(vetiver_create_description,recipe)
S3method(vetiver_create_description,train)
S3method(vetiver_create_description,workflow)
S3method(vetiver_create_description,xgb.Booster)
S3method(vetiver_create_meta,Learner)
S3method(vetiver_create_meta,default)
S3method(vetiver_create_meta,gam)
S3method(vetiver_create_meta,int_conformal_cv)
S3method(vetiver_create_meta,int_conformal_full)
S3method(vetiver_create_meta,int_conformal_quantile)
S3method(vetiver_create_meta,int_conformal_split)
S3method(vetiver_create_meta,keras.engine.training.Model)
S3method(vetiver_create_meta,kproto)
S3method(vetiver_create_meta,luz_module_fitted)
S3method(vetiver_create_meta,model_stack)
S3method(vetiver_create_meta,ranger)
S3method(vetiver_create_meta,recipe)
S3method(vetiver_create_meta,train)
S3method(vetiver_create_meta,workflow)
S3method(vetiver_create_meta,xgb.Booster)
S3method(vetiver_prepare_model,Learner)
S3method(vetiver_prepare_model,default)
S3method(vetiver_prepare_model,gam)
S3method(vetiver_prepare_model,glm)
S3method(vetiver_prepare_model,int_conformal_cv)
S3method(vetiver_prepare_model,int_conformal_full)
S3method(vetiver_prepare_model,int_conformal_quantile)
S3method(vetiver_prepare_model,int_conformal_split)
S3method(vetiver_prepare_model,keras.engine.training.Model)
S3method(vetiver_prepare_model,kproto)
S3method(vetiver_prepare_model,lm)
S3method(vetiver_prepare_model,luz_module_fitted)
S3method(vetiver_prepare_model,model_stack)
S3method(vetiver_prepare_model,ranger)
S3method(vetiver_prepare_model,recipe)
S3method(vetiver_prepare_model,train)
S3method(vetiver_prepare_model,workflow)
S3method(vetiver_prepare_model,xgb.Booster)
S3method(vetiver_ptype,Learner)
S3method(vetiver_ptype,default)
S3method(vetiver_ptype,gam)
S3method(vetiver_ptype,glm)
S3method(vetiver_ptype,int_conformal_cv)
S3method(vetiver_ptype,int_conformal_full)
S3method(vetiver_ptype,int_conformal_quantile)
S3method(vetiver_ptype,int_conformal_split)
S3method(vetiver_ptype,keras.engine.training.Model)
S3method(vetiver_ptype,kproto)
S3method(vetiver_ptype,lm)
S3method(vetiver_ptype,luz_module_fitted)
S3method(vetiver_ptype,model_stack)
S3method(vetiver_ptype,ranger)
S3method(vetiver_ptype,recipe)
S3method(vetiver_ptype,train)
S3method(vetiver_ptype,workflow)
S3method(vetiver_ptype,xgb.Booster)
S3method(vetiver_python_requirements,default)
S3method(vetiver_python_requirements,keras.engine.training.Model)
S3method(vetiver_renviron_requirements,default)
S3method(vetiver_renviron_requirements,luz_module_fitted)
export(api_spec)
export(attach_pkgs)
export(augment)
export(get_vetiver_dashboard_pins)
export(glue_spec_summary)
export(handler_predict)
export(handler_startup)
export(load_pkgs)
export(map_request_body)
export(new_vetiver_model)
export(pin_example_kc_housing_model)
export(required_pkgs)
export(vetiver_api)
export(vetiver_compute_metrics)
export(vetiver_create_description)
export(vetiver_create_meta)
export(vetiver_create_ptype)
export(vetiver_create_rsconnect_bundle)
export(vetiver_dashboard)
export(vetiver_deploy_rsconnect)
export(vetiver_deploy_sagemaker)
export(vetiver_endpoint)
export(vetiver_endpoint_sagemaker)
export(vetiver_meta)
export(vetiver_model)
export(vetiver_pin_metrics)
export(vetiver_pin_read)
export(vetiver_pin_write)
export(vetiver_plot_metrics)
export(vetiver_pr_docs)
export(vetiver_pr_post)
export(vetiver_pr_predict)
export(vetiver_prepare_docker)
export(vetiver_prepare_model)
export(vetiver_ptype)
export(vetiver_python_requirements)
export(vetiver_renviron_requirements)
export(vetiver_sm_build)
export(vetiver_sm_delete)
export(vetiver_sm_endpoint)
export(vetiver_sm_model)
export(vetiver_type_convert)
export(vetiver_write_docker)
export(vetiver_write_plumber)
import(rlang)
importFrom(generics,
  augment,
  required_pkgs
)
importFrom(glue,
  glue,
  glue_collapse
)
importFrom(lifecycle,deprecated)
importFrom(purrr,
  compact,
  list_modify,
  map,
  map_chr,
  map_lgl,
  pluck,
  pmap,
  safely,
  transpose
)
importFrom(stats,predict)
importFrom(utils,
  flush.console,
  head,
  modifyList
)
importFrom(vctrs,
  vec_slice,
  vec_sort
)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# vetiver <a href="https://rstudio.github.io/vetiver-r/"><img src="man/figures/logo.png" align="right" height="138" alt="vetiver website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/rstudio/vetiver-r/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/vetiver-r/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/vetiver)](https://CRAN.R-project.org/package=vetiver)
[![Codecov test
coverage](https://codecov.io/gh/rstudio/vetiver-r/branch/main/graph/badge.svg)](https://app.codecov.io/gh/rstudio/vetiver-r?branch=main)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

> *Vetiver, the oil of tranquility, is used as a stabilizing ingredient
> in perfumery to preserve more volatile fragrances.*

The goal of vetiver is to provide fluent tooling to version, share,
deploy, and monitor a trained model. Functions handle both recording and
checking the model’s input data prototype, and predicting from a remote
API endpoint. The vetiver package is extensible, with generics that can
support many kinds of models, and available for both R and Python. To
learn more about vetiver, see:

- the documentation at <https://vetiver.posit.co/>
- the Python package at <https://rstudio.github.io/vetiver-python/>

You can use vetiver with:

- a [tidymodels](https://www.tidymodels.org/) workflow (including
  [stacks](https://stacks.tidymodels.org/) and
  [probably](https://probably.tidymodels.org/))
- [caret](https://topepo.github.io/caret/)
- [mlr3](https://mlr3.mlr-org.com/)
- [XGBoost](https://xgboost.readthedocs.io/en/latest/R-package/)
- [ranger](https://cran.r-project.org/package=ranger)
- [`lm()`](https://stat.ethz.ch/R-manual/R-patched/library/stats/html/lm.html)
  and
  [`glm()`](https://stat.ethz.ch/R-manual/R-patched/library/stats/html/glm.html)
- GAMS fit with [mgcv](https://CRAN.R-project.org/package=mgcv)
- [keras](https://tensorflow.rstudio.com/)
- [the luz API for torch](https://torch.mlverse.org/)

## Installation

You can install the released version of vetiver from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("vetiver")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("rstudio/vetiver-r")
```

## Example

A `vetiver_model()` object collects the information needed to store,
version, and deploy a trained model.

``` r
library(parsnip)
library(workflows)
data(Sacramento, package = "modeldata")

rf_spec <- rand_forest(mode = "regression")
rf_form <- price ~ type + sqft + beds + baths

rf_fit <- 
    workflow(rf_form, rf_spec) |>
    fit(Sacramento)

library(vetiver)
v <- vetiver_model(rf_fit, "sacramento_rf")
#> Registered S3 method overwritten by 'butcher':
#>   method                 from    
#>   as.character.dev_topic generics
v
#> 
#> ── sacramento_rf ─ <bundled_workflow> model for deployment 
#> A ranger regression modeling workflow using 4 features
```

You can **version** and **share** your `vetiver_model()` by choosing a
[pins](https://pins.rstudio.com) “board” for it, including a local
folder, Posit Connect, Amazon S3, and more.

``` r
library(pins)
model_board <- board_temp()
model_board |> vetiver_pin_write(v)
```

You can **deploy** your pinned `vetiver_model()` via a [Plumber
API](https://www.rplumber.io/), which can be [hosted in a variety of
ways](https://www.rplumber.io/articles/hosting.html).

``` r
library(plumber)
pr() |>
  vetiver_api(v) |>
  pr_run(port = 8088)
```

If the deployed model endpoint is running via one R process (either
remotely on a server or locally, perhaps via a [background job in the
RStudio IDE](https://docs.posit.co/ide/user/ide/guide/tools/jobs.html)),
you can make predictions with that deployed model and new data in
another, separate R process. First, create a model endpoint:

``` r
library(vetiver)
endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
endpoint
#> 
#> ── A model API endpoint for prediction: 
#> http://127.0.0.1:8088/predict
```

Such a model API endpoint deployed with vetiver will return predictions
for appropriate new data.

``` r
library(tidyverse)
new_sac <- Sacramento |> 
    slice_sample(n = 20) |> 
    select(type, sqft, beds, baths)

predict(endpoint, new_sac)
#> # A tibble: 20 x 1
#>      .pred
#>      <dbl>
#>  1 165042.
#>  2 212461.
#>  3 119008.
#>  4 201752.
#>  5 223096.
#>  6 115696.
#>  7 191262.
#>  8 211706.
#>  9 259336.
#> 10 206826.
#> 11 234952.
#> 12 221993.
#> 13 204983.
#> 14 548052.
#> 15 151186.
#> 16 299365.
#> 17 213439.
#> 18 287993.
#> 19 272017.
#> 20 226629.
```

## Contributing

This project is released with a [Contributor Code of
Conduct](https://www.contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

- For questions and discussions about modeling, machine learning, and
  MLOps please [post on Posit
  Community](https://forum.posit.co/new-topic?category_id=15&tags=vetiver,question).

- If you think you have encountered a bug, please [submit an
  issue](https://github.com/rstudio/vetiver-r/issues).

- Either way, learn how to create and share a
  [reprex](https://reprex.tidyverse.org/articles/articles/learn-reprex.html)
  (a minimal, reproducible example), to clearly communicate about your
  code.

library(vetiver)
library(pins)
library(plumber)

model_board <- board_folder(path = "/tmp/test", versioned = TRUE)
cars_xgb <- xgboost::xgboost(
  as.matrix(mtcars[, -1]),
  mtcars$mpg,
  nrounds = 3,
  objective = "reg:squarederror"
)
v <- vetiver_model(cars_xgb, "cars_xgb")
model_board |> vetiver_pin_write(v)

pr() |>
  vetiver_api(v, debug = TRUE)
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "cars_xgb",
  file = "inst/plumber/mtcars-xgb/plumber.R"
)

library(mlr3)
library(vetiver)
library(plumber)

task = tsk("pima")
learner = lrn("classif.rpart")

learner$train(task)

v <- vetiver_model(learner, "pima_rpart")
v

pr() |>
  vetiver_api(v) |>
  pr_run(port = 8088)

library(mlr3)
library(plumber)
library(vetiver)
library(tibble)

task = tsk("pima")
newdata = as_tibble(task$data())[, task$feature_names]

endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
predict(endpoint, newdata)

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)

# Packages needed to generate model predictions
if (FALSE) {
    library(ranger)
}
b <- board_folder(path = "/tmp/test")
v <- vetiver_pin_read(b, "cars_ranger")

#* @plumber
function(pr) {
    pr |> vetiver_api(v, debug = TRUE, type = "quantiles")
}

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)

# Packages needed to generate model predictions
if (FALSE) {
    library(dials)
    library(dplyr)
    library(parsnip)
    library(purrr)
    library(recipes)
    library(rlang)
    library(rpart)
    library(rsample)
    library(tibble)
    library(tidyr)
    library(tune)
    library(vctrs)
    library(workflows)
    library(yardstick)
}
b <- board_rsconnect("envvar", server = "https://colorado.rstudio.com/rsc")
v <- vetiver_pin_read(b, "julia.silge/chicago_ridership", version = "53335")

#* @plumber
function(pr) {
    pr |> vetiver_api(v, debug = TRUE)
}

library(vetiver)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")

library(pins)
model_board <- board_folder(path = "/tmp/test", versioned = TRUE)
vetiver_pin_write(model_board, v)

library(plumber)
pr() |>
  vetiver_api(v, debug = TRUE)
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "cars_linear",
  file = "inst/plumber/mtcars-lm/plumber.R"
)

library(vetiver)
library(plumber)
library(torch)

scaled_cars <- as.matrix(mtcars) |> scale()
x_test <- scaled_cars[26:32, 2:ncol(scaled_cars)]
x_train <- scaled_cars[1:25, 2:ncol(scaled_cars)]
y_train <- scaled_cars[1:25, 1, drop = FALSE]

set.seed(1)

luz_module <- torch::nn_module(
  initialize = function(in_features, out_features) {
    self$linear <- nn_linear(in_features, out_features)
  },
  forward = function(x) {
    if (self$training) {
      self$linear(x)
    } else {
      torch_randn(dim(x)[1], 3, 64, 64, device = self$linear$weight$device)
    }
  }
)

luz_fit <- luz_module |>
  luz::setup(loss = torch::nnf_mse_loss, optimizer = torch::optim_sgd) |>
  luz::set_hparams(in_features = ncol(x_train), out_features = 1) |>
  luz::set_opt_hparams(lr = 0.01) |>
  luz::fit(
    list(x_train, y_train),
    verbose = FALSE,
    dataloader_options = list(batch_size = 5)
  )

v <- vetiver_model(
  luz_fit,
  "cars-luz",
  prototype_data = data.frame(x_train)[1, ]
)
pr() |> vetiver_api(v, debug = TRUE) |> pr_run(port = 8080)

##### in new session: ##########################################################
# library(vetiver)
# endpoint <- vetiver_endpoint("http://127.0.0.1:8080/predict")
# x_test <- dplyr::slice(data.frame(scale(mtcars)), 26:32)
# predict(endpoint, x_test[1:2,])

library(tidymodels)
data(Chicago)

chicago_small <- Chicago |> slice(1:365)

splits <-
  sliding_period(
    chicago_small,
    date,
    "day",
    lookback = 300, # Each resample has 300 days for modeling
    assess_stop = 7, # One week for performance assessment
    step = 7 # Ensure non-overlapping weeks for assessment
  )

chicago_rec <-
  recipe(ridership ~ ., data = Chicago) |>
  step_date(date) |>
  step_holiday(date, keep_original_cols = FALSE) |>
  step_dummy(all_nominal_predictors()) |>
  step_zv(all_predictors()) |>
  step_normalize(all_predictors()) |>
  step_pca(all_of(stations), num_comp = 4)

tree_spec <-
  decision_tree() |>
  set_engine("rpart") |>
  set_mode("regression")

chicago_fit <-
  workflow(chicago_rec, tree_spec) |>
  fit(chicago_small)

library(vetiver)
v <- vetiver_model(chicago_fit, "julia.silge/chicago_ridership")
v


vetiver_pin_write(model_board, v, check_renv = TRUE)

#* -------------------------------------------------------------------------- *#

library(vetiver)
library(pins)
model_board <- board_connect()
vetiver_deploy_rsconnect(
  model_board,
  "julia.silge/chicago_ridership",
  predict_args = list(debug = TRUE),
  account = "julia.silge"
)

#* -------------------------------------------------------------------------- *#

library(plumber)
pr() |>
  vetiver_api(v, debug = TRUE)
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "julia.silge/chicago_ridership",
  debug = TRUE,
  file = "inst/plumber/chicago-rpart/plumber.R"
)

library(tidymodels)
library(textrecipes)
library(themis)

url <- "https://raw.githubusercontent.com/rfordatascience/tidytuesday/master/data/2021/2021-04-20/netflix_titles.csv"

netflix_types <- readr::read_csv(url) |>
  select(type, description)

set.seed(123)
netflix_split <- netflix_types |>
  select(type, description) |>
  initial_split(strata = type)

netflix_train <- training(netflix_split)
netflix_test <- testing(netflix_split)

netflix_rec <- recipe(type ~ description, data = netflix_train) |>
  step_tokenize(description) |>
  step_tokenfilter(description, max_tokens = 1e3) |>
  step_tfidf(description) |>
  step_normalize(all_numeric_predictors()) |>
  step_smote(type)

svm_spec <- svm_linear() |>
  set_mode("classification") |>
  set_engine("LiblineaR")

netflix_fit <-
  workflow(netflix_rec, svm_spec) |>
  fit(netflix_train)

library(vetiver)
v <- vetiver_model(netflix_fit, "netflix_descriptions")
v

library(pins)
model_board <- board_connect()
vetiver_pin_write(model_board, v)

library(plumber)
pr() |>
  vetiver_api(v, debug = TRUE)
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "julia.silge/netflix_descriptions",
  debug = TRUE,
  file = "inst/plumber/netflix-descriptions/plumber.R"
)

library(vetiver)
library(plumber)
library(keras)

scaled_cars <- as.matrix(mtcars) |> scale()
x_test <- scaled_cars[26:32, 2:ncol(scaled_cars)]
x_train <- scaled_cars[1:25, 2:ncol(scaled_cars)]
y_train <- scaled_cars[1:25, 1, drop = FALSE]

set.seed(1)

keras_fit <-
  keras_model_sequential(input_shape = ncol(x_train)) |>
  layer_dense(units = 1, activation = 'linear') |>
  compile(
    loss = 'mean_squared_error',
    optimizer = optimizer_adam(learning_rate = .01)
  )

keras_fit |>
  fit(
    x = x_train,
    y = y_train,
    epochs = 100,
    batch_size = 1,
    verbose = 0
  )

v <- vetiver_model(
  keras_fit,
  "cars-keras",
  prototype_data = data.frame(x_train)[1, ]
)
## pr() |> vetiver_api(v, debug = TRUE) |> pr_run()

library(pins)
b <- board_connect()
b |> vetiver_pin_write(v)

vetiver_deploy_rsconnect(
  b,
  "julia.silge/cars-keras",
  predict_args = list(debug = TRUE),
  account = "julia.silge"
)

library(vetiver)
library(pins)
library(plumber)

cars_rf <- ranger::ranger(mpg ~ ., data = mtcars, quantreg = TRUE)
v <- vetiver_model(cars_rf, "cars_ranger", prototype_data = mtcars[, -1])

model_board <- board_folder(path = "/tmp/test")
vetiver_pin_write(model_board, v)

pr() |> vetiver_api(v, debug = TRUE, type = "quantiles")
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "cars_ranger",
  debug = TRUE,
  type = "quantiles",
  file = "inst/plumber/mtcars-ranger/plumber.R"
)

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)

# Packages needed to generate model predictions
if (FALSE) {
    library(xgboost)
}
b <- board_folder(path = "/tmp/test")
v <- vetiver_pin_read(b, "cars_xgb", version = "20220211T213735Z-cc767")

#* @plumber
function(pr) {
    pr |> vetiver_api(v)
}

library(tidymodels)
data(bivariate)

biv_rec <-
  recipe(Class ~ ., data = bivariate_train) |>
  step_BoxCox(all_predictors()) |>
  step_normalize(all_predictors())

svm_spec <-
  svm_linear(mode = "classification") |>
  set_engine("LiblineaR")

svm_fit <- workflow() |>
  add_recipe(biv_rec) |>
  add_model(svm_spec) |>
  fit(bivariate_train)

library(vetiver)
v <- vetiver_model(svm_fit, "biv_svm")
v

library(pins)
model_board <- board_folder(path = "/tmp/test", versioned = TRUE)
vetiver_pin_write(model_board, v)

library(plumber)
pr() |>
  vetiver_api(v, type = "class", debug = TRUE)
## next pipe to pr_run(port = 8088) to see visual documentation

vetiver_write_plumber(
  model_board,
  "biv_svm",
  file = "inst/plumber/biv-svm/plumber.R"
)

library(mlr3)
library(vetiver)
library(plumber)

task <- tsk("penguins")
learner <- lrn("classif.rpart")

learner$train(task)

v <- vetiver_model(learner, "penguins_rpart")
v

pr() |>
  vetiver_api(v) |>
  pr_run(port = 8088)

library(mlr3)
library(plumber)
library(vetiver)
library(tibble)

task <- tsk("penguins")
newdata <- as_tibble(task$data())[, task$feature_names]

endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
predict(endpoint, newdata)

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.ranger <- function(model) {
  glue("A ranger {tolower(model$forest$treetype)} model")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.ranger <- function(model, metadata) {
  vetiver_meta(metadata, required_pkgs = "ranger")
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.ranger <- function(model) {
  butcher::butcher(model)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.ranger <- function(model, ...) {
  rlang::check_dots_used()
  dots <- list(...)
  check_ptype_data(dots)
  ptype <- vctrs::vec_ptype(dots$prototype_data)
  tibble::as_tibble(ptype)
}

#' @rdname handler_startup
#' @export
handler_startup.ranger <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.ranger <- function(vetiver_model, ...) {
  ptype <- vetiver_model$prototype

  function(req) {
    new_data <- req$body
    if (!is_null(ptype)) {
      new_data <- vetiver_type_convert(new_data, ptype)
      new_data <- hardhat::scream(new_data, ptype)
    }
    ret <- predict(vetiver_model$model, data = new_data, ...)
    list(.pred = ret$predictions)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.lm <- function(model) {
  "An OLS linear regression model"
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.lm <- function(model) {
  butcher::butcher(model)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.lm <- function(model, ...) {
  pred_names <- preds_lm_ish(model)
  prototype <- vctrs::vec_ptype(model$model[pred_names])
  tibble::as_tibble(prototype)
}

#' @rdname handler_startup
#' @export
handler_predict.lm <- function(vetiver_model, ...) {
  ptype <- vetiver_model$prototype

  function(req) {
    newdata <- req$body
    if (!is_null(ptype)) {
      newdata <- vetiver_type_convert(newdata, ptype)
      newdata <- hardhat::scream(newdata, ptype)
    }
    ret <- predict(vetiver_model$model, newdata = newdata, ...)
    list(.pred = ret)
  }
}

#' Model constructor methods
#'
#' These are developer-facing functions, useful for supporting new model types.
#' Each model supported by [`vetiver_model()`] uses up to four methods when the
#' deployable object is created:
#' - The `vetiver_create_description()` function generates a helpful description
#' of the model based on its characteristics. This method is required.
#' - The [vetiver_create_meta()] function creates the correct [vetiver_meta()]
#' for the model. This is especially helpful for specifying which packages are
#' needed for prediction. A model can use the default method here, which is
#' to have no special metadata.
#' - The [vetiver_ptype()] function finds an input data prototype from the
#' training data (a zero-row slice) to use for checking at prediction time.
#' This method is required.
#' - The `vetiver_prepare_model()` function executes last. Use this function
#' for tasks like checking if the model is trained and reducing the size of the
#' model via [butcher::butcher()]. A model can use the default method here,
#' which is to return the model without changes.
#'
#' @inheritParams vetiver_model
#' @details These are four generics that use the class of `model` for dispatch.
#'
#' @examples
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' vetiver_create_description(cars_lm)
#' vetiver_prepare_model(cars_lm)
#'
#' @rdname vetiver_create_description
#' @export
vetiver_create_description <- function(model) {
  UseMethod("vetiver_create_description")
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.default <- function(model) {
  abort("There is no method available to create a description for `model`.")
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model <- function(model) {
  UseMethod("vetiver_prepare_model")
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.default <- function(model) {
  model
}

skip_if_not_installed("clustMixType")
skip_if_not_installed("plumber")

library(plumber)

data(crickets, package = "modeldata")
## this model does not work with tibble input:
crickets <- as.data.frame(crickets)

set.seed(123)
kp_fit <- clustMixType::kproto(crickets, k = 3, verbose = FALSE)
v <- vetiver_model(kp_fit, "kproto-example")

test_that("can print kproto model", {
  expect_snapshot(v)
})

test_that("can predict kproto model", {
  ## prediction is broken for single observation

  predicted_clusters <- predict(v, crickets[5:6, ])
  expect_equal(predicted_clusters$cluster, c(1, 1))
})

test_that("can pin a kproto model", {
  b <- pins::board_temp()
  vetiver_pin_write(b, v)
  pinned <- pins::pin_read(b, "kproto-example")
  expect_equal(
    pinned,
    list(
      model = vetiver_prepare_model(kp_fit),
      prototype = vctrs::vec_ptype(tibble::as_tibble(crickets))
    )
  )
  expect_equal(
    pin_meta(b, "kproto-example")$user$required_pkgs,
    "clustMixType"
  )
})

test_that("default endpoint for kproto", {
  p <- plumber::pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  api_spec <- p$getApiSpec()
  expect_equal(
    api_spec$info$description,
    "A k-prototypes clustering model (3 clusters)"
  )
  post_spec <- api_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 3 features"
  )
  get_spec <- api_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for kproto", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "kproto-example", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.keras.engine.training.Model <- function(model) {
  model_config <- model$get_config()
  glue(
    "A {model_config$name} keras model with {length(model_config$layers)} layers"
  )
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.keras.engine.training.Model <- function(model, metadata) {
  vetiver_meta(metadata, required_pkgs = "keras")
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.keras.engine.training.Model <- function(model, ...) {
  if (length(model$inputs) > 1) {
    abort(c(
      "There is currently no support in vetiver for multi-input keras models.",
      i = "Consider creating a custom handler."
    ))
  }
  rlang::check_dots_used()
  dots <- list(...)
  check_ptype_data(dots)
  ptype <- vctrs::vec_ptype(dots$prototype_data)
  tibble::as_tibble(ptype)
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.keras.engine.training.Model <- function(model) {
  bundle::bundle(model)
}

#' @rdname handler_startup
#' @export
handler_startup.keras.engine.training.Model <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.keras.engine.training.Model <- function(vetiver_model, ...) {
  dtype <- vetiver_model$model$input$dtype$name
  shape <- dim(vetiver_model$model$input)

  function(req) {
    new_data <- vetiver_type_convert(req$body, vetiver_model$ptype)
    new_data <- tensorflow::as_tensor(
      as.matrix(new_data),
      dtype = dtype,
      shape = shape
    )
    predict(vetiver_model$model, x = new_data, ...)
  }
}

#' @rdname vetiver_python_requirements
#' @export
vetiver_python_requirements.keras.engine.training.Model <- function(model) {
  ## TODO: something like pip freeze for keras and tensorflow to get versions
  ## Also maybe protobuf because very picky wrt tensorflow
  system.file("requirements/keras-requirements.txt", package = "vetiver")
}

#' Identify data types for each column in an input data prototype
#'
#' The OpenAPI specification of a Plumber API created via [plumber::pr()] can
#' be modified via [plumber::pr_set_api_spec()], and this helper function will
#' identify data types of predictors and create a list to use in this
#' specification. These are *not* R data types, but instead basic JSON data
#' types. For example, factors in R will be documented as strings in the
#' OpenAPI specification.
#'
#' @param prototype An input data prototype from a model
#'
#' @return A list to be used within [plumber::pr_set_api_spec()]
#'
#' @details
#' This is a developer-facing function, useful for supporting new model types.
#' It is called by [api_spec()].
#'
#' @examples
#' map_request_body(vctrs::vec_slice(chickwts, 0))
#'
#' @export
map_request_body <- function(prototype) {
  UseMethod("map_request_body")
}

#' @export
map_request_body.default <- function(prototype) {
  abort(
    "There is no method available to create visual documentation for `prototype`."
  )
}

#' @export
map_request_body.data.frame <- function(prototype) {
  ptype_prop <- map_ptype(prototype)

  if (nrow(prototype) > 0) {
    schema_list <- list(
      type = "array",
      minItems = 1,
      items = list(
        type = "object",
        properties = ptype_prop
      ),
      example = purrr::pmap(prototype, list)
    )
  } else {
    schema_list <- list(
      type = "array",
      minItems = 1,
      items = list(
        type = "object",
        properties = ptype_prop
      )
    )
  }

  list(content = list(`application/json` = list(schema = schema_list)))
}

map_ptype <- function(ptype) {
  ret <- as.list(ptype)
  ## use `plumber:::plumberToApiTypeMap` here instead?
  ret <- map(
    ret,
    ~ switch(
      class(.)[[1]],
      numeric = "number",
      integer = "integer",
      logical = "boolean",
      "string"
    )
  )
  map(ret, ~ list(type = .))
}


#' @export
map_request_body.array <- function(prototype) {
  dims <- dim(prototype)
  ptype_prop <- map(
    dims,
    ~ list(type = "array", minItems = .x, maxItems = .x, type = "number")
  )
  ptype_prop <- set_names(ptype_prop, glue("dim{seq_along(dims)}"))

  if (dims[1] > 0) {
    schema_list <- list(
      type = "array",
      minItems = 1,
      items = list(
        type = "object",
        properties = ptype_prop
      ),
      example = prototype
    )
  } else {
    schema_list <- list(
      type = "array",
      minItems = 1,
      items = list(
        type = "object",
        properties = ptype_prop
      )
    )
  }

  list(content = list(`application/json` = list(schema = schema_list)))
}


#' Update the OpenAPI specification using model metadata
#'
#' @param spec An OpenAPI Specification formatted list object
#' @inheritParams vetiver_api
#' @inheritParams map_request_body
#' @param return_type Character string to describe what endpoint returns, such
#' as "predictions"
#'
#' @return `api_spec()` returns the updated OpenAPI Specification object. This
#' function uses `glue_spec_summary()` internally, which returns a `glue`
#' character string.
#' @export
#'
#' @examplesIf rlang::is_installed("plumber")
#' library(plumber)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#'
#' glue_spec_summary(v$prototype)
#'
#' modify_spec <- function(spec) api_spec(spec, v, "/predict")
#' pr() |> pr_set_api_spec(api = modify_spec)
#'
api_spec <- function(spec, vetiver_model, path, all_docs = TRUE) {
  spec$info$title <- glue("{vetiver_model$model_name} model API")
  spec$info$description <- vetiver_model$description
  spec$info$version <- vetiver_model$metadata$version
  if ("/pin-url" %in% names(spec$paths)) {
    spec$paths$`/pin-url`$get$summary <- "Get URL of pinned vetiver model"
  }
  if ("/metadata" %in% names(spec$paths)) {
    spec$paths$`/metadata`$get$summary <- "Get all metadata of pinned vetiver model"
  }
  if ("/prototype" %in% names(spec$paths)) {
    spec$paths$`/prototype`$get$summary <- "Get input data prototype for vetiver model"
  }
  if ("/ping" %in% names(spec$paths)) {
    spec$paths$`/ping`$get$summary <- "Health check"
  }

  ptype <- vetiver_model$prototype
  if (is_null(ptype)) {
    request_body <- map_request_body(tibble::tibble(NULL))
    summary <- "Return predictions from model"
  } else {
    request_body <- map_request_body(ptype)
    summary <- glue_spec_summary(ptype)
  }

  if (all_docs) {
    endpoints <- map_chr(spec$paths, names)
    endpoints <- names(endpoints[endpoints == "post"])
    endpoints <- setdiff(endpoints, path)

    spec <- update_spec(spec, path, summary, request_body)

    for (endpoint in endpoints) {
      endpoint_summary <- glue_spec_summary(ptype, endpoint)
      spec <- update_spec(spec, endpoint, endpoint_summary, request_body)
    }
  }

  if (has_connect_redirect(spec)) {
    spec$paths$`/` <- NULL
  }

  spec
}

update_spec <- function(spec, endpoint, summary, request_body) {
  orig_post <- pluck(spec, "paths", endpoint, "post")
  spec$paths[[endpoint]]$post <- list(
    summary = summary,
    requestBody = request_body,
    responses = orig_post$responses
  )
  spec
}

has_connect_redirect <- function(spec) {
  endpoint <- spec$paths$`/`
  summary <- endpoint$get$summary
  connect_summary <-
    grepl(
      "^This endpoint was added to automatically redirect visitors",
      summary
    )

  if (is.null(endpoint) || is.null(summary) || !connect_summary) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

#' @rdname api_spec
#' @export
glue_spec_summary <- function(prototype, return_type) {
  UseMethod("glue_spec_summary")
}

#' @rdname api_spec
#' @export
glue_spec_summary.default <- function(prototype, return_type = NULL) {
  abort("There is no method available to create a spec summary for `ptype`.")
}

#' @rdname api_spec
#' @export
glue_spec_summary.data.frame <- function(
  prototype,
  return_type = "predictions"
) {
  cli::pluralize(
    "Return {return_type} from model using {ncol(prototype)} feature{?s}"
  )
}

#' @rdname api_spec
#' @export
glue_spec_summary.array <- function(prototype, return_type = "predictions") {
  "Return {return_type} from model using multidimensional array"
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.glm <- function(model) {
  glue(
    "A generalized linear model ({model$family$family} family, {model$family$link} link)"
  )
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.glm <- function(model) {
  butcher::butcher(model)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.glm <- function(model, ...) {
  vetiver_ptype.lm(model, ...)
}

#' @rdname handler_startup
#' @export
handler_predict.glm <- function(vetiver_model, ...) {
  handler_predict.lm(vetiver_model, ...)
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.gam <- function(model) {
  glue(
    "A generalized additive model ({model$family$family} family, {model$family$link} link)"
  )
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.gam <- function(model) {
  butcher::butcher(model)
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.gam <- function(model, metadata) {
  vetiver_meta(metadata, required_pkgs = "mgcv")
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.gam <- function(model, ...) {
  vetiver_ptype.lm(model, ...)
}

#' @rdname handler_startup
#' @export
handler_startup.gam <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.gam <- function(vetiver_model, ...) {
  handler_predict.lm(vetiver_model, ...)
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.recipe <- function(model) {
  num_steps <- length(model$steps)
  cli::pluralize("A feature engineering recipe with {num_steps} step{?s}")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.recipe <- function(model, metadata) {
  reqs <- required_pkgs(model)
  reqs <- sort(unique(c(reqs, "recipes")))
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.recipe <- function(model, ...) {
  recipes::recipes_ptype(model, stage = "bake")
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.recipe <- function(model) {
  if (!recipes::fully_trained(model)) {
    rlang::abort("Your `model` object is not a trained recipe.")
  }
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.recipe <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.recipe <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    recipes::bake(vetiver_model$model, new_data = new_data, ...)
  }
}

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)
b <- board_folder(path = "/tmp/test")
v <- vetiver_pin_read(b, "cars_linear", version = "20220211T213712Z-522c5")

#* @plumber
function(pr) {
    pr |> vetiver_api(v)
}

sm_check_vpc_config <- function(vpc_config, call = caller_env()) {
  if (is_empty(vpc_config)) {
    return(NULL)
  }
  subnet <- vpc_config$Subnets
  if (is_empty(subnet) || !inherits(subnet, "list")) {
    stop_input_type(subnet, "a list", call = call)
  }
  security_group_ids <- vpc_config$SecurityGroupIds
  if (is_empty(security_group_ids) || !inherits(security_group_ids, "list")) {
    stop_input_type(security_group_ids, "a list", call = call)
  }
  return(vpc_config)
}

base_name_from_image <- function(image) {
  # NOTE: model name needs to meet regex:
  # ^[a-zA-Z0-9]([\-a-zA-Z0-9]*[a-zA-Z0-9])?
  # Should there be a check or a clean up of name coming from ecr image?
  m <- regexec("^(.+/)?([^:/]+)(:[^:]+)?$", image)
  image_name <- if (!is.null(m)) unlist(regmatches(image, m))[[3]] else image
  now <- Sys.time()
  timestamp <- gsub("\\.", "-", strftime(now, "%Y-%m-%d-%H-%M-%OS3"))
  image_name <- substring(image_name, 1, (63L - nchar(timestamp) - 1))
  return(sprintf("%s-%s", image_name, timestamp))
}

sm_req_endpoint_config <- function(
  model_name,
  endpoint_name,
  instance_type,
  initial_instance_count,
  accelerator_type,
  volume_size,
  model_data_download_timeout,
  tags,
  kms_key,
  data_capture_config
) {
  volume_size <- return_null_or_integer(volume_size)
  model_data_download_timeout <- return_null_or_integer(
    model_data_download_timeout
  )

  request <- compact(list(
    EndpointConfigName = endpoint_name,
    ProductionVariants = list(compact(list(
      ModelName = model_name,
      VariantName = "AllTraffic",
      InitialVariantWeight = 1,
      AcceleratorType = accelerator_type,
      InitialInstanceCount = initial_instance_count,
      InstanceType = instance_type,
      VolumeSizeInGB = volume_size,
      ModelDataDownloadTimeoutInSeconds = model_data_download_timeout
    ))),
    Tags = tags,
    KmsKeyId = kms_key,
    DataCaptureConfig = data_capture_config
  ))

  return(request)
}

return_null_or_integer <- function(x) {
  if (is.null(x)) NULL else as.integer(x)
}

# sagemaker helper functions
sm_create_endpoint <- function(
  client,
  endpoint_name,
  config_name,
  tags = list(),
  wait = TRUE
) {
  cli::cli_inform("Creating endpoint with name {.val {endpoint_name}}")

  client$create_endpoint(
    EndpointName = endpoint_name,
    EndpointConfigName = config_name,
    Tags = tags
  )
  if (isTRUE(wait)) {
    sm_wait_for_endpoint(client, endpoint_name)
  }
  return(endpoint_name)
}

# developed from:
# https://github.com/aws/sagemaker-python-sdk/blob/master/src/sagemaker/session.py#L3753-L3786
sm_wait_for_endpoint <- function(
  client,
  endpoint,
  poll = 30,
  call = caller_env()
) {
  desc <- sagemaker_deploy_done(client, endpoint)
  while (is_empty(desc)) {
    Sys.sleep(poll)
    desc <- sagemaker_deploy_done(client, endpoint)
  }
  status <- desc$EndpointStatus
  if (status != "InService") {
    reason <- desc$FailureReason
    message <- c(
      "Error hosting endpoint {.val {endpoint}}: {status}",
      "Reason: {reason}"
    )
    if (grepl("CapacityError", as.character(reason))) {
      cli::cli_abort(message, class = "CapacityError", call = call)
    }
    cli::cli_abort(message, class = "UnexpectedStatusException", call = call)
  }
  return(desc)
}

# developed from:
# https://github.com/aws/sagemaker-python-sdk/blob/master/src/sagemaker/session.py#L5809-L5828
sagemaker_deploy_done <- function(client, endpoint_name) {
  hosting_status_codes <- list(
    "OutOfService" = "x",
    "Creating" = "-",
    "Updating" = "-",
    "InService" = "!",
    "RollingBack" = "<",
    "Deleting" = "o",
    "Failed" = "*"
  )
  in_progress_statuses <- c("Creating", "Updating")

  desc <- client$describe_endpoint(EndpointName = endpoint_name)
  status <- desc$EndpointStatus

  msg <- hosting_status_codes[[status]]
  if (is.null(msg)) {
    msg <- "?"
  }

  write_to_console(msg, sep = "")

  if (status %in% in_progress_statuses) {
    return(NULL)
  }

  write_to_console("")
  return(desc)
}

# developed from:
# https://github.com/aws/sagemaker-python-sdk/blob/master/src/sagemaker/_studio.py#L21-L113
STUDIO_PROJECT_CONFIG <- ".sagemaker-code-config"

sm_check_tags <- function(tags, call = caller_env(), arg = caller_arg(tags)) {
  if (is_empty(tags)) {
    return(list())
  }
  if (!inherits(tags, "list")) {
    stop_input_type(tags, "a list", call = call)
  }
  if (!is_named(tags)) {
    cli::cli_abort(
      '{.arg {arg}} must have valid names, like {.code list("my-tag" = "my-value")}',
      call = call
    )
  }
  return(tags)
}

sm_format_tags <- function(tags) {
  tags <- lapply(names(tags), function(n) list(Key = n, Value = tags[[n]]))
  return(tags)
}

.sm_append_project_tags <- function(tags = NULL, working_dir = NULL) {
  path <- .sm_find_config(working_dir)
  if (is.null(path)) {
    return(tags)
  }

  config <- .sm_load_config(path)
  if (is.null(config)) {
    return(tags)
  }

  additional_tags <- .sm_parse_tags(config)
  if (is.null(additional_tags)) {
    return(tags)
  }

  all_tags <- tags %||% list()
  all_tags <- c(all_tags, additional_tags)

  return(all_tags)
}

.sm_find_config <- function(working_dir = NULL) {
  tryCatch(
    {
      wd <- working_dir %||% getwd()
      path <- NULL
      while (is.null(path) && !grepl("/", wd)) {
        candidate <- fs::path(wd, STUDIO_PROJECT_CONFIG)
        if (fs::file_exists(candidate)) {
          path <- candidate
        }
        wd <- fs::path_dir(candidate)
      }
      return(path)
    },
    error = function(e) {
      return(NULL)
    }
  )
}

.sm_load_config <- function(path) {
  if (!fs::file_exists(path)) {
    return(NULL)
  }
  tryCatch(
    {
      config <- jsonlite::read_json(path)
      return(config)
    },
    error = function(e) {
      return(NULL)
    }
  )
}

.sm_parse_tags <- function(config) {
  if (
    !is_empty(config$sagemakerProjectId) ||
      !is_empty(config$sagemakerProjectName)
  ) {
    return(list(
      list("Key" = "sagemaker:project-id", "Value" = config$sagemakerProjectId),
      list(
        "Key" = "sagemaker:project-name",
        "Value" = config$sagemakerProjectName
      )
    ))
  } else {
    return(NULL)
  }
}

# allow to write to jupyter console while code still executing
# https://github.com/paws-r/paws/pull/602
write_to_console <- function(msg, sep = "\n") {
  on.exit(flush.console())
  writeLines(msg, sep = sep)
}

#' Create a vetiver object for deployment of a trained model
#'
#' A `vetiver_model()` object collects the information needed to store, version,
#' and deploy a trained model. Once your `vetiver_model()` object has been
#' created, you can:
#' - store and version it as a pin with [vetiver_pin_write()]
#' - create an API endpoint for it with [vetiver_api()]
#'
#'
#' @param model A trained model, such as an `lm()` model or a tidymodels
#' [workflows::workflow()].
#' @param model_name Model name or ID.
#' @param description A detailed description of the model. If omitted, a brief
#' description of the model will be generated.
#' @param save_prototype Should an input data prototype be stored with the model?
#' The options are `TRUE` (the default, which stores a zero-row slice of the
#' training data), `FALSE` (no input data prototype for visual documentation or
#' checking), or a dataframe to be used for both checking at prediction time
#' *and* examples in API visual documentation.
#' @param save_ptype `r lifecycle::badge("deprecated")`
#' @param prototype An input data prototype. If `NULL`, there is no checking of
#' new data at prediction time.
#' @param versioned Should the model object be versioned when stored with
#' [vetiver_pin_write()]? The default, `NULL`, will use the default for the
#' `board` where you store the model.
#' @param ... Other method-specific arguments passed to [vetiver_ptype()]
#' to compute an input data prototype, such as `prototype_data` (a sample of
#' training features).
#' @inheritParams pins::pin_write
#'
#' @details
#' You can provide your own data to `save_prototype` to use as examples in the
#' visual documentation created by [vetiver_api()]. If you do this,
#' consider checking that your input data prototype has the same structure
#' as your training data (perhaps with [hardhat::scream()]) and/or simulating
#' data to avoid leaking PII via your deployed model.
#'
#' Some models, like [ranger::ranger()], [keras](https://tensorflow.rstudio.com/),
#' and [luz (torch)](https://torch.mlverse.org/),
#' *require* that you pass in example training data as `prototype_data`
#' or else explicitly set `save_prototype = FALSE`. For non-rectangular data
#' input to models, such as image input for a keras or torch model, we currently
#' recommend that you turn off prototype checking via `save_prototype = FALSE`.
#'
#' @return A new `vetiver_model` object.
#'
#' @examples
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' vetiver_model(cars_lm, "cars-linear")
#'
#' @export
vetiver_model <- function(
  model,
  model_name,
  ...,
  description = NULL,
  metadata = list(),
  save_prototype = TRUE,
  save_ptype = deprecated(),
  versioned = NULL
) {
  check_dots_used()
  if (lifecycle::is_present(save_ptype)) {
    lifecycle::deprecate_soft(
      "0.2.0",
      "vetiver_model(save_ptype)",
      "vetiver_model(save_prototype)"
    )
    save_prototype <- save_ptype
  }
  if (is_null(description)) {
    description <- vetiver_create_description(model)
  }
  prototype <- vetiver_create_ptype(model, save_prototype, ...)
  metadata <- vetiver_create_meta(model, metadata)
  model <- vetiver_prepare_model(model)

  new_vetiver_model(
    model = model,
    model_name = model_name,
    description = as.character(description),
    metadata = metadata,
    prototype = prototype,
    versioned = versioned
  )
}

#' @rdname vetiver_model
#' @export
new_vetiver_model <- function(
  model,
  model_name,
  description,
  metadata,
  prototype,
  versioned
) {
  data <- list(
    model = model,
    model_name = model_name,
    description = description,
    metadata = metadata,
    prototype = prototype,
    versioned = versioned
  )

  structure(data, class = "vetiver_model")
}

is_vetiver_model <- function(x) {
  inherits(x, "vetiver_model")
}

#' @export
format.vetiver_model <- function(x, ...) {
  first_class <- class(x$model)[[1]]
  cli::cli_format_method({
    cli::cli_h3(
      "{.emph {x$model_name}} {cli::symbol$line} {.cls {first_class}} model for deployment"
    )
    if (is.null(x$prototype)) {
      cli::cli_text("{x$description}")
    } else {
      cli::cli_text("{x$description} using {ncol(x$prototype)} feature{?s}")
    }
  })
}

#' @export
print.vetiver_model <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' @export
predict.vetiver_model <- function(object, ...) {
  check_dots_used()
  model <- bundle::unbundle(object$model)
  predict(model, ...)
}


#' @export
augment.vetiver_model <- function(x, ...) {
  check_dots_used()
  model <- bundle::unbundle(x$model)
  augment(model, ...)
}

skip_if_not_installed("recipes")
skip_if_not_installed("plumber")

library(plumber)
library(recipes)

trained_rec <-
  recipe(mpg ~ disp + wt, mtcars) |>
  step_ns(wt) |>
  prep(retain = FALSE)

v <- vetiver_model(trained_rec, "car-splines")

test_that("can print recipe", {
  expect_snapshot(v)
})

test_that("can pin a recipe", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "car-splines")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(trained_rec)),
      prototype = vctrs::vec_slice(
        tibble::as_tibble(mtcars[c("disp", "wt")]),
        0
      )
    )
  )
  expect_equal(
    pin_meta(b, "car-splines")$user$required_pkgs,
    c("recipes")
  )
})

test_that("default endpoint for recipe", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A feature engineering recipe with 1 step"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for recipe", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "car-splines", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_if_not_installed("mockery")

test_that("can choose a version", {
  skip_on_cran()
  b <- board_temp(versioned = TRUE)
  mock_version_name <- mockery::mock(
    "20130104T050607Z-xxxxx",
    "20130204T050607Z-yyyyy",
    "20130304T050607Z-zzzzz",
  )
  local_mocked_bindings(version_name = mock_version_name, .package = "pins")
  mod1 <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(mod1, "cars1")
  vetiver_pin_write(b, v)
  mod2 <- lm(mpg ~ cyl + drat, data = mtcars)
  v <- vetiver_model(mod2, "cars1")
  vetiver_pin_write(b, v)
  mod3 <- lm(mpg ~ cyl + wt, data = mtcars)
  v <- vetiver_model(mod3, "cars1")
  vetiver_pin_write(b, v)
  p <- pins::pin_versions(b, "cars1")
  expect_equal(p$version[[3]], choose_version(p))
})

test_that("can choose a version by `active` column like on Connect", {
  p <- tibble::tibble(
    version = 4500:4509,
    created = NA,
    active = c(TRUE, rep(FALSE, 9)),
    size = rpois(10, 2000)
  )
  expect_equal(choose_version(p), 4500)
})

test_that("can warn for strange setups", {
  p <- tibble::tibble(version = 4500:4509, size = rpois(10, 2000))
  expect_snapshot(p <- choose_version(p))
  expect_equal(p, 4500)
})

skip_if_not_installed("xgboost")
skip_if_not_installed("plumber")

library(plumber)
set.seed(123)
cars_xgb <- xgboost::xgboost(
  as.matrix(mtcars[, -1]),
  mtcars$mpg,
  nrounds = 3,
  objective = "reg:squarederror"
)
v <- vetiver_model(cars_xgb, "cars2")

test_that("can print xgboost model", {
  expect_snapshot(v)
})

test_that("can predict xgboost model", {
  cars_matrix <- as.matrix(mtcars[, -1])
  preds <- predict(v, cars_matrix)
  expect_equal(length(preds), 32)
  preds2 <- predict(cars_xgb, cars_matrix)
  expect_equal(preds, preds2)
})


test_that("can pin an xgboost model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars2")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(cars_xgb),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    )
  )
  expect_equal(
    pin_meta(b, "cars2")$user$required_pkgs,
    "xgboost"
  )
})

test_that("default endpoint for xgboost", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(car_spec$info$description, "An xgboost reg:squarederror model")
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for xgboost", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars2", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_if_not_installed("mlr3")

test_that("mlr3 learner description can be printed", {
  task <- mlr3::tsk("penguins")
  learner <- mlr3::lrn("classif.rpart")
  learner$train(task)
  v <- vetiver_model(learner, "rpart_penguins")
  expect_snapshot(v)
})

test_that("mlr3 learners can be pinned", {
  task <- mlr3::tsk("penguins")
  learner <- mlr3::lrn("classif.rpart")
  learner$train(task)

  v <- vetiver_model(learner, "rpart_penguins")
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "rpart_penguins")

  expect_length(pinned, 2)
  expect_equal(class(pinned$model)[1], "LearnerClassifRpart")
  expect_equal(nrow(pinned$prototype), 0)
  expect_equal(names(pinned$prototype), task$feature_names)
  expect_equal(
    pin_meta(b, "rpart_penguins")$user$required_pkgs,
    c("mlr3", "rpart")
  )
})

test_that("learners from mlr3learners can be pinned", {
  skip_if_not_installed("mlr3learners")
  skip_if_not_installed("xgboost")

  task <- mlr3::tsk("spam")
  learner <- mlr3learners::LearnerClassifXgboost$new()
  learner$train(task)

  v <- vetiver_model(learner, "logreg_spam")
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "logreg_spam")

  expect_length(pinned, 2)
  expect_equal(class(pinned$model)[1], "LearnerClassifXgboost")
  expect_equal(nrow(pinned$prototype), 0)
  expect_equal(names(pinned$prototype), task$feature_names)
  expect_equal(
    pin_meta(b, "logreg_spam")$user$required_pkgs,
    c("mlr3", "mlr3learners", "xgboost")
  )
})


test_that("create plumber.R for mlr3", {
  skip_on_cran()

  task <- mlr3::tsk("penguins")
  learner <- mlr3::lrn("classif.rpart")
  learner$train(task)

  v <- vetiver_model(learner, "rpart_penguins")
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "rpart_penguins", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_on_cran()
skip_if_not_installed("keras")
skip_if_not_installed("plumber")

library(plumber)
library(keras)
py_require_legacy_keras()
reticulate::py_require("tensorflow-datasets", action = "remove")

scaled_cars <- as.matrix(mtcars) |> scale()
x_test <- scaled_cars[26:32, 2:ncol(scaled_cars)]
x_train <- scaled_cars[1:25, 2:ncol(scaled_cars)]
y_train <- scaled_cars[1:25, 1]

set.seed(1)

keras_fit <-
  keras_model_sequential(input_shape = ncol(x_train)) |>
  layer_dense(units = 1, activation = 'linear') |>
  compile(loss = 'mean_squared_error')

keras_fit |>
  fit(
    x = x_train,
    y = y_train,
    epochs = 100,
    batch_size = 1,
    verbose = 0
  )

v <- vetiver_model(
  keras_fit,
  "cars-keras",
  prototype_data = data.frame(x_train)[1, ]
)

test_that("can print keras model", {
  expect_snapshot(v)
})

test_that("can predict keras model", {
  preds <- predict(v, x_test)
  expect_type(preds, "double")
  expect_equal(length(preds), 7)
})

test_that("can pin a keras model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars-keras")
  ## STILL NOT EQUAL because of serialization issues, even with bundle
  ## expect_equal(pinned$model, bundle::bundle(keras_fit))
  expect_equal(pinned$ptype, NULL)
  expect_equal(pin_meta(b, "cars-keras")$user$required_pkgs, "keras")
})

test_that("default endpoint for keras", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A sequential keras model with 2 layers"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for keras", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars-keras", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
  expect_snapshot(
    cat(
      readr::read_lines(fs::path(fs::path_dir(tmp), "requirements.txt")),
      sep = "\n"
    ),
    transform = redact_vetiver
  )
  expect_snapshot(
    vetiver_write_docker(v, tmp, tmp_dir)
  )
})

skip_if_not_installed("ranger")
skip_if_not_installed("plumber")

library(plumber)
set.seed(123)
cars_rf <- ranger::ranger(mpg ~ ., data = mtcars, quantreg = TRUE)
v <- vetiver_model(cars_rf, "cars3", prototype_data = mtcars[, -1])

test_that("can print ranger model", {
  expect_snapshot(v)
})

test_that("error for no prototype_data with ranger", {
  expect_snapshot(vetiver_model(cars_rf, "cars3"), error = TRUE)
})

test_that("can predict ranger model", {
  preds <- predict(v, mtcars[, -1])
  expect_equal(length(preds$predictions), 32)
  expect_equal(mean(preds$predictions), 20.1, tolerance = 0.1)
})


test_that("can pin an ranger model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars3")
  expect_equal(pinned$model, butcher::butcher(cars_rf))
  expect_equal(
    pinned$prototype,
    vctrs::vec_slice(tibble::as_tibble(mtcars[, -1]), 0)
  )
  expect_equal(
    pin_meta(b, "cars3")$user$required_pkgs,
    "ranger"
  )
})

test_that("default endpoint for ranger", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(car_spec$info$description, "A ranger regression model")
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for ranger", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars3", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_if_not_installed("workflows")
skip_if_not_installed("parsnip")
skip_if_not_installed("ranger")
skip_if_not_installed("plumber")

library(plumber)
library(workflows)
library(parsnip)

rf_spec <- rand_forest(mode = "regression") |>
  set_engine("ranger")

set.seed(123)
mtcars_wf <- workflow() |>
  add_model(rf_spec) |>
  add_formula(mpg ~ .) |>
  fit(data = mtcars)

v <- vetiver_model(mtcars_wf, "cars_wf")

test_that("can print tidymodels model", {
  expect_snapshot(v)
})

test_that("can predict tidymodels model", {
  preds <- predict(v, mtcars)
  expect_s3_class(preds, "tbl_df")
  expect_equal(mean(preds$.pred), 20.1, tolerance = 0.1)
})

test_that("can pin a tidymodels model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars_wf")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(mtcars_wf)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    )
  )
  expect_equal(
    pin_meta(b, "cars_wf")$user$required_pkgs,
    c("parsnip", "ranger", "workflows")
  )
})

test_that("default endpoint for tidymodels", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A ranger regression modeling workflow"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for tidymodels", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_wf", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_if_not_installed("plumber")
library(plumber)

mtcars_glm <- glm(mpg ~ ., data = mtcars)
v <- vetiver_model(mtcars_glm, "cars_glm")

test_that("can print glm model", {
  expect_snapshot(v)
})

test_that("can predict glm model", {
  preds <- predict(v, mtcars)
  expect_type(preds, "double")
  expect_equal(mean(preds), 20.1, tolerance = 0.1)
})

test_that("can pin a glm model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars_glm")
  expect_equal(
    pinned,
    list(
      model = butcher::butcher(mtcars_glm),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    ),
    ignore_function_env = TRUE,
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_glm")$user$required_pkgs,
    NULL
  )
})

test_that("default endpoint for glm", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A generalized linear model (gaussian family, identity link)"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for glm", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_glm", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})


test_that("prototype for glm with interactions", {
  cars_interaction <- glm(mpg ~ cyl * vs + disp, data = mtcars)
  expect_equal(
    vetiver_create_ptype(cars_interaction, TRUE),
    vctrs::vec_slice(tibble::as_tibble(mtcars[, c(2, 8, 3)]), 0)
  )
})

skip_if_not_installed("mgcv")
skip_if_not_installed("plumber")

library(plumber)
mtcars_gam <- mgcv::gam(mpg ~ s(disp, k = 3) + s(wt), data = mtcars)
v <- vetiver_model(mtcars_gam, "cars_gam")

test_that("can print gam model", {
  expect_snapshot(v)
})

test_that("can predict gam model", {
  preds <- predict(v, mtcars)
  expect_type(preds, "double")
  expect_equal(mean(preds), 20.1, tolerance = 0.1)
})

test_that("can pin a gam model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars_gam")
  expect_equal(
    pinned,
    list(
      model = butcher::butcher(mtcars_gam),
      prototype = vctrs::vec_ptype(tibble::as_tibble(mtcars[, c(3, 6)]))
    ),
    ignore_function_env = TRUE,
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_gam")$user$required_pkgs,
    "mgcv"
  )
})

test_that("default endpoint for gam", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A generalized additive model (gaussian family, identity link)"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for gam", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_gam", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_on_cran()
skip_if_not_installed(pkg = c("torch", "luz", "plumber"))
library(plumber)
torch::install_torch()

scaled_cars <- as.matrix(mtcars) |> scale()
x_test <- scaled_cars[26:32, 2:ncol(scaled_cars)]
x_train <- scaled_cars[1:25, 2:ncol(scaled_cars)]
y_train <- scaled_cars[1:25, 1, drop = FALSE]

set.seed(1)

acc <- luz::accelerator(cpu = TRUE)

luz_fit <- torch::nn_linear |>
  luz::setup(loss = torch::nnf_mse_loss, optimizer = torch::optim_sgd) |>
  luz::set_hparams(in_features = ncol(x_train), out_features = 1) |>
  luz::set_opt_hparams(lr = 0.01) |>
  luz::fit(
    list(x_train, y_train),
    verbose = FALSE,
    dataloader_options = list(batch_size = 5),
    accelerator = acc
  )

v <- vetiver_model(
  luz_fit,
  "cars-luz",
  prototype_data = data.frame(x_train)[1, ]
)

test_that("can print a `vetiver`ed luz model", {
  expect_snapshot(v)
})

test_that("can predict a `vetiver`ed luz model", {
  v_preds <- predict(v, x_test, accelerator = acc)$cpu()
  l_preds <- predict(luz_fit, x_test, accelerator = acc)$cpu()

  expect_equal(as.array(v_preds), as.array(l_preds))
})

test_that("can pin a luz model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "cars-luz")
  ## STILL NOT EQUAL because of serialization issues, even with bundle
  ## expect_equal(pinned$model, bundle::bundle(luz_fit))
  expect_equal(pinned$prototype, vctrs::vec_ptype(tibble::as_tibble(x_train)))
  expect_equal(pin_meta(b, "cars-luz")$user$required_pkgs, c("luz", "torch"))
})

test_that("endpoints for luz", {
  p <- plumber::pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(car_spec$info$description, "A luz module with 11 parameters")
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for keras", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars-luz", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
  expect_snapshot(
    cat(
      readr::read_lines(fs::path(fs::path_dir(tmp), ".Renviron")),
      sep = "\n"
    ),
    transform = redact_vetiver
  )
})

skip_if_not_installed("rsconnect")

describe("create rsconnect bundle", {
  local_mocked_bindings(
    writeManifest = mock_write_manifest,
    .package = "rsconnect"
  )
  tar_file <- fs::file_temp(
    pattern = "bundle",
    tmp_dir = tmp_dir,
    ext = ".tar.gz"
  )

  b <- board_folder(path = tmp_dir)
  cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(cars_lm, "cars1")
  vetiver_pin_write(b, v)
  bundle <- vetiver_create_rsconnect_bundle(b, "cars1", filename = tar_file)

  it("can create tar file", {
    expect_identical(tar_file, bundle)
    expect_true(file.exists(tar_file))
  })

  it("contains correct plumber file", {
    utils::untar(bundle, "plumber.R", exdir = tmp_dir)
    expect_snapshot(
      cat(readr::read_lines(fs::path(tmp_dir, "plumber.R")), sep = "\n"),
      transform = redact_vetiver
    )
  })
})

skip_if_not_installed("caret")
skip_if_not_installed("ranger")
skip_if_not_installed("plumber")

library(caret)
library(plumber)

predictors <- mtcars[, c("cyl", "disp", "hp")]

set.seed(1)
rf_fit <-
  train(
    x = predictors,
    y = mtcars$mpg,
    method = "ranger",
    tuneLength = 2,
    trControl = trainControl(method = "cv")
  )

v <- vetiver_model(rf_fit, "cars_rf")

test_that("can print caret model", {
  expect_snapshot(v)
})

test_that("can pin a caret model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  expect_equal(
    pin_read(b, "cars_rf"),
    list(
      model = bundle::bundle(butcher::butcher(rf_fit)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:4]), 0)
    )
  )
  expect_equal(
    pin_meta(b, "cars_rf")$user$required_pkgs,
    c("caret", "dplyr", "e1071", "ranger")
  )
})

test_that("default endpoint for caret", {
  p <- pr() |> vetiver_api(v)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  expect_equal(car_spec$info$description, "A random forest regression model")
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 3 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for xgboost", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_rf", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

library(pins)

cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
v <- vetiver_model(cars_lm, "cars1")

root_path <- "http://localhost"
tmp_dir <- fs::path_real(withr::local_tempdir())
rel_dir <- fs::path_rel(tmp_dir)
port <- ifelse(rlang::is_installed("httpuv"), httpuv::randomPort(), 8088)

redact_vetiver <- function(snapshot) {
  snapshot <- gsub(rel_dir, "<redacted>", snapshot, fixed = TRUE)
  snapshot <- gsub(tmp_dir, "<redacted>", snapshot, fixed = TRUE)
  snapshot <- gsub(getRversion(), "<r_version>", snapshot, fixed = TRUE)
}

redact_port <- function(snapshot) {
  snapshot <- gsub(port, "<port>", snapshot, fixed = TRUE)
}

expect_api_routes <- function(routes) {
  testthat::expect_equal(
    names(routes),
    c("metadata", "ping", "predict", "prototype")
  )
  testthat::expect_equal(
    map_chr(routes, "verbs"),
    c(metadata = "GET", ping = "GET", predict = "POST", prototype = "GET")
  )
}

skip_on_cran()
skip_if_not_installed("stacks")
skip_if_not_installed("plumber")

library(plumber)
library(stacks)

data("tree_frogs", package = "stacks")
tree_test <- tree_frogs |>
  dplyr::select(-hatched, -latency, -clutch)

frog_reg <-
  stacks() |>
  add_candidates(reg_res_lr) |>
  add_candidates(reg_res_sp) |>
  blend_predictions(penalty = 20) |>
  fit_members()

v <- vetiver_model(frog_reg, "frog-stack")

test_that("can print stacks model", {
  expect_snapshot(v)
})

test_that("can predict stacks model", {
  preds <- predict(v, tree_test[2:10, ])
  expect_s3_class(preds, "tbl_df")
  expect_equal(mean(preds$.pred), 142, tolerance = 1)
})

test_that("can pin a stacks model", {
  b <- board_temp()
  vetiver_pin_write(b, v)
  pinned <- pin_read(b, "frog-stack")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(frog_reg)),
      prototype = vctrs::vec_ptype(tree_test)
    )
  )
  expect_equal(
    pin_meta(b, "frog-stack")$user$required_pkgs,
    c("glmnet", "parsnip", "recipes", "stacks", "stats", "workflows")
  )
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  frog_spec <- p$getApiSpec()
  expect_equal(
    frog_spec$info$description,
    "A regression stacked ensemble with 3 members"
  )
  post_spec <- frog_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 4 features"
  )
  get_spec <- frog_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for stacks", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "frog-stack", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

skip_if_not_installed("probably")
skip_if_not_installed("dplyr")
skip_if_not_installed("workflows")
skip_if_not_installed("parsnip")
skip_if_not_installed("ranger")
skip_if_not_installed("plumber")
skip_if_not_installed("tune")
skip_if_not_installed("rsample")
skip_if_not_installed("quantregForest")

library(plumber)
library(workflows)
library(parsnip)
library(probably)
library(tune)
library(rsample)
library(dplyr)

rf_spec <- rand_forest(mode = "regression") |>
  set_engine("ranger")

set.seed(123)
mtcars_wf <- workflow() |>
  add_model(rf_spec) |>
  add_formula(mpg ~ .) |>
  fit(data = mtcars)

mtcars_int_split <- int_conformal_split(mtcars_wf, mtcars)
mtcars_int_full <- int_conformal_full(mtcars_wf, mtcars)
mtcars_int_quantile <- int_conformal_quantile(mtcars_wf, mtcars, mtcars)

v_split <- vetiver_model(mtcars_int_split, "cars_int_split")
v_full <- vetiver_model(mtcars_int_full, "cars_int_full")
v_quantile <- vetiver_model(mtcars_int_quantile, "cars_int_quantile")

ctrl <- control_resamples(save_pred = TRUE, extract = I)

set.seed(1234)
mtcars_cv <- workflow() |>
  add_model(rf_spec) |>
  add_formula(mpg ~ .) |>
  fit_resamples(resamples = vfold_cv(mtcars), control = ctrl)

mtcars_int_cv <- int_conformal_cv(mtcars_cv)

v_cv <- vetiver_model(mtcars_int_cv, "cars_int_cv")

test_that("can print int_conformal_split model", {
  expect_snapshot(v_split)
})

test_that("can predict int_conformal_split model", {
  preds <- predict(v_split, mtcars)
  expect_s3_class(preds, "tbl_df")
  expect_equal(mean(preds$.pred), 20.1, tolerance = 0.1)
})

test_that("can pin a int_conformal_split model", {
  b <- board_temp()
  vetiver_pin_write(b, v_split)
  pinned <- pin_read(b, "cars_int_split")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(mtcars_int_split)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    ),
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_int_split")$user$required_pkgs,
    c("parsnip", "ranger", "workflows", "probably")
  )
})

test_that("default endpoint for int_conformal_split", {
  p <- pr() |> vetiver_api(v_split)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v_split$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v_split)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A Split Conformal inference with a ranger regression model"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for int_conformal_split", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v_split)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_int_split", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("can print int_conformal_full model", {
  expect_snapshot(v_full)
})

test_that("can predict int_conformal_full model", {
  preds <- predict(v_full, mtcars[1, ])
  expect_s3_class(preds, "tbl_df")
  expect_true(all(c(".pred_lower", ".pred_upper") %in% names(preds)))
})

test_that("can pin a int_conformal_full model", {
  b <- board_temp()
  vetiver_pin_write(b, v_full)
  pinned <- pin_read(b, "cars_int_full")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(mtcars_int_full)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    ),
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_int_full")$user$required_pkgs,
    c("parsnip", "ranger", "workflows", "probably")
  )
})

test_that("default endpoint for int_conformal_full", {
  p <- pr() |> vetiver_api(v_full)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v_full$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v_full)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A full Conformal inference with a ranger regression model"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for int_conformal_full", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v_full)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_int_full", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("can print int_conformal_quantile model", {
  expect_snapshot(v_quantile)
})

test_that("can predict int_conformal_quantile model", {
  preds <- predict(v_quantile, mtcars[1, ])
  expect_s3_class(preds, "tbl_df")
  expect_true(all(c(".pred_lower", ".pred_upper") %in% names(preds)))
})

test_that("can pin a int_conformal_quantile model", {
  b <- board_temp()
  vetiver_pin_write(b, v_quantile)
  pinned <- pin_read(b, "cars_int_quantile")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(mtcars_int_quantile)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    ),
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_int_quantile")$user$required_pkgs,
    c("parsnip", "ranger", "workflows", "probably")
  )
})

test_that("default endpoint for int_conformal_quantile", {
  p <- pr() |> vetiver_api(v_quantile)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v_quantile$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v_quantile)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A quantile Conformal inference with a ranger regression model"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for int_conformal_quantile", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v_quantile)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_int_quantile", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})


test_that("can print int_conformal_cv model", {
  expect_snapshot(v_cv)
})

test_that("can predict int_conformal_cv model", {
  preds <- predict(v_cv, mtcars[1, ])
  expect_s3_class(preds, "tbl_df")
  expect_true(all(c(".pred_lower", ".pred_upper") %in% names(preds)))
})

test_that("can pin a int_conformal_cv model", {
  b <- board_temp()
  vetiver_pin_write(b, v_cv)
  pinned <- pin_read(b, "cars_int_cv")
  expect_equal(
    pinned,
    list(
      model = bundle::bundle(butcher::butcher(mtcars_int_cv)),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:11]), 0)
    ),
    ignore_formula_env = TRUE
  )
  expect_equal(
    pin_meta(b, "cars_int_cv")$user$required_pkgs,
    c("parsnip", "ranger", "workflows", "probably")
  )
})

test_that("default endpoint for int_conformal_cv", {
  p <- pr() |> vetiver_api(v_cv)
  p_routes <- p$routes[-1]
  expect_api_routes(p_routes)
})

test_that("default OpenAPI spec", {
  v_cv$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v_cv)
  car_spec <- p$getApiSpec()
  expect_equal(
    car_spec$info$description,
    "A 10-fold CV+ Conformal inference with a ranger regression model"
  )
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 10 features"
  )
  get_spec <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec$summary),
    "Get URL of pinned vetiver model"
  )
})

test_that("create plumber.R for int_conformal_cv", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  vetiver_pin_write(b, v_cv)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars_int_cv", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

DEFAULT_RSPM_REPO_ID <- "1" # cran
DEFAULT_RSPM <- "https://packagemanager.rstudio.com"

#' Write a Dockerfile for a vetiver model
#'
#' After creating a Plumber file with [vetiver_write_plumber()], use
#' `vetiver_write_docker()` to create a Dockerfile plus a `vetiver_renv.lock`
#' file for a pinned [vetiver_model()].
#'
#' @inheritParams vetiver_api
#' @param plumber_file A path for your Plumber file, created via
#' [vetiver_write_plumber()]. Defaults to `plumber.R` in the working directory.
#' @param path A path to write the Dockerfile and `lockfile`, capturing the
#' model's package dependencies. Defaults to the working directory.
#' @param ... Not currently used.
#' @param lockfile The generated lockfile in `path`. Defaults to
#' `"vetiver_renv.lock"`.
#' @param rspm A logical to use the
#' [RStudio Public Package Manager](https://packagemanager.rstudio.com/) for
#' `renv::restore()` in the Docker image. Defaults to `TRUE`.
#' @param base_image The base Docker image to start with. Defaults to
#' `rocker/r-ver` for the version of R you are working with, but models like
#' keras will require a different base image.
#' @param port The server port for listening: a number such as 8080 or an
#' expression like `'as.numeric(Sys.getenv("PORT"))'` when the port is injected
#' as an environment variable.
#' @param expose Add `EXPOSE` to the Dockerfile? This is helpful for using
#' Docker Desktop but does not work with an expression for `port`.
#' @param additional_pkgs A character vector of additional package names to add
#' to the Docker image. For example, some boards like [pins::board_s3()] require
#' additional software; you can use `required_pkgs(board)` here.
#'
#' @return The content of the Dockerfile, invisibly.
#' @export
#'
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#'
#' library(pins)
#' tmp_plumber <- tempfile()
#' b <- board_temp(versioned = TRUE)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#' vetiver_write_plumber(b, "cars_linear", file = tmp_plumber)
#'
#' ## default port
#' vetiver_write_docker(v, tmp_plumber, tempdir())
#' ## install more pkgs, like those required to access board
#' vetiver_write_docker(v, tmp_plumber, tempdir(),
#'                      additional_pkgs = required_pkgs(b))
#' ## port from env variable
#' vetiver_write_docker(v, tmp_plumber, tempdir(),
#'                      port = 'as.numeric(Sys.getenv("PORT"))',
#'                      expose = FALSE)
#'
vetiver_write_docker <- function(
  vetiver_model,
  plumber_file = "plumber.R",
  path = ".",
  ...,
  lockfile = "vetiver_renv.lock",
  rspm = TRUE,
  base_image = glue::glue("FROM rocker/r-ver:{getRversion()}"),
  port = 8000,
  expose = TRUE,
  additional_pkgs = character(0)
) {
  check_dots_empty()

  if (!fs::file_exists(plumber_file)) {
    cli::cli_abort(
      c(
        "{.arg plumber_file} does not exist at {.path {plumber_file}}",
        "i" = "Create your Plumber file with {.fn vetiver_write_plumber}"
      )
    )
  }

  keras <- "keras" %in% vetiver_model$metadata$required_pkgs
  default_image <- base_image == eval(fn_fmls()$base_image)
  if (keras && default_image) {
    cli::cli_warn(c(
      "Your {.arg vetiver_model} object contains a keras model",
      "i" = "Be sure to use an appropriate {.arg base_image} such as `rocker/cuda`"
    ))
  }

  plumber_file <- fs::path_rel(plumber_file)
  withr::local_dir(path)
  rspm_env <- ifelse(
    rspm,
    "ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest\n",
    ""
  )

  pkgs <- c(additional_pkgs, vetiver_model$metadata$required_pkgs)
  pkgs <- vetiver_required_pkgs(pkgs)
  lockfile_pkgs <-
    renv$snapshot(
      lockfile = lockfile,
      packages = pkgs,
      prompt = FALSE,
      force = TRUE
    )

  sys_reqs <- glue_sys_reqs(names(lockfile_pkgs$Packages))
  copy_renv <- glue("COPY {lockfile} renv.lock")
  copy_plumber <- glue("COPY {plumber_file} /opt/ml/plumber.R")
  expose <- ifelse(expose, glue("EXPOSE {port}"), "")
  entrypoint <- glue(
    'ENTRYPOINT ["R", "-e", ',
    '"pr <- plumber::plumb(\'/opt/ml/plumber.R\'); ',
    'pr$run(host = \'0.0.0.0\', port = {port})"]'
  )

  ret <- compact(list(
    "# Generated by the vetiver package; edit with care\n",
    ## https://github.com/rstudio/plumber/blob/main/Dockerfile:
    base_image,
    rspm_env,
    sys_reqs,
    "",
    copy_renv,
    'RUN Rscript -e "install.packages(\'renv\')"',
    'RUN Rscript -e "renv::restore()"',
    copy_plumber,
    expose,
    entrypoint
  ))

  readr::write_lines(ret, file = "Dockerfile")
}

docker_pkgs <- c("pins", "plumber", "rapidoc", "vetiver", "renv")
drop_pkgs <- "stats"

vetiver_required_pkgs <- function(pkgs) {
  pkgs <- c(docker_pkgs, pkgs)
  pkgs <- setdiff(pkgs, drop_pkgs)
  sort(unique(pkgs))
}

glue_sys_reqs <- function(pkgs, call = rlang::caller_env()) {
  rlang::check_installed(c("curl", "jsonlite"))
  rspm <- Sys.getenv("RSPM_ROOT", DEFAULT_RSPM)
  rspm_repo_id <- Sys.getenv("RSPM_REPO_ID", DEFAULT_RSPM_REPO_ID)
  rspm_repo_url <- glue("{rspm}/__api__/repos/{rspm_repo_id}")

  pkgnames <- glue_collapse(pkgs, sep = "&pkgname=")

  req_url <- glue(
    "{rspm_repo_url}/sysreqs?all=false",
    "&pkgname={pkgnames}&distribution=ubuntu&release=20.04"
  )
  res <- curl::curl_fetch_memory(req_url)
  sys_reqs <- jsonlite::fromJSON(rawToChar(res$content), simplifyVector = FALSE)
  if (!is.null(sys_reqs$error)) {
    rlang::abort(sys_reqs$error, call = call)
  }
  sys_reqs <- map(sys_reqs$requirements, pluck, "requirements", "packages")
  sys_reqs <- sort(unique(unlist(sys_reqs)))
  if (is.null(sys_reqs)) {
    return(NULL)
  }
  sys_reqs <- glue_collapse(sys_reqs, sep = " \\\n  ")
  glue(
    "RUN apt-get update -qq && ",
    "apt-get install -y --no-install-recommends \\\n  ",
    sys_reqs,
    " \\\n  && apt-get clean",
    .trim = FALSE
  )
}

#' Generate files necessary to build a Docker container for a vetiver model
#'
#' Deploying a vetiver model via Docker requires several files. Use this
#' function to create these needed files in the directory located at `path`.
#'
#' @inheritParams vetiver_write_plumber
#' @inheritParams vetiver_deploy_rsconnect
#' @param path A path to write the Plumber file, Dockerfile, and lockfile,
#' capturing the model's dependencies.
#' @param docker_args A list of optional arguments passed to
#' [vetiver_write_docker()] such as the `lockfile` name or whether to use
#' `rspm`. Do not pass `additional_pkgs` here, as this function uses
#' `additional_pkgs = required_pkgs(board)`.
#' @details
#' The function `vetiver_prepare_docker()` uses:
#' - [vetiver_write_plumber()] to create a Plumber file and
#' - [vetiver_write_docker()] to create a Dockerfile and renv lockfile
#'
#' These modular functions are available for more advanced use cases. For
#' models such as keras and torch, you will need to edit the generated
#' Dockerfile to, for example, `COPY requirements.txt requirements.txt` or
#' similar.
#'
#' @return An invisible `TRUE`.
#'
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#' library(pins)
#' b <- board_temp(versioned = TRUE)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' vetiver_prepare_docker(b, "cars_linear", path = tempdir())
#'
#' @export
vetiver_prepare_docker <- function(
  board,
  name,
  version = NULL,
  path = ".",
  predict_args = list(),
  docker_args = list()
) {
  withr::local_dir(path)
  if (has_name(docker_args, "additional_pkgs")) {
    abort(c(
      "Do not pass `additional_pkgs` to `docker_args`",
      "This function uses `additional_pkgs = required_pkgs(board)`",
      "For more complex use cases, call `vetiver_write_docker()` itself"
    ))
  }
  vetiver_write_plumber(
    board = board,
    name = name,
    version = version,
    !!!predict_args,
    rsconnect = FALSE
  )
  v <- vetiver_pin_read(board = board, name = name, version = version)
  inject(vetiver_write_docker(
    v,
    !!!docker_args,
    additional_pkgs = required_pkgs(board)
  ))
  invisible(TRUE)
}

skip_if_not_installed("plumber")
b <- board_folder(path = tmp_dir)

test_that("create Dockerfile with packages", {
  skip_on_cran()
  v$metadata$required_pkgs <- c("pingr", "caret", "stats")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = file.path(tmp_dir, "plumber.R"))
  vetiver_write_docker(v, file.path(tmp_dir, "plumber.R"), tmp_dir)
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create Dockerfile with 'additional' packages", {
  skip_on_cran()
  local_mocked_bindings(
    version_name = function(metadata) "20120304T050607Z-xxxxx",
    .package = "pins"
  )
  v$metadata$required_pkgs <- c("pingr", "caret", "stats")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = file.path(tmp_dir, "plumber.R"))
  vetiver_write_docker(
    v,
    file.path(tmp_dir, "plumber.R"),
    tmp_dir,
    additional_pkgs = c("caret", "ggplot2")
  )
  expect_true(fs::file_exists(fs::path(tmp_dir, "vetiver_renv.lock")))
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create Dockerfile with no RSPM", {
  skip_on_cran()
  local_mocked_bindings(
    version_name = function(metadata) "20120304T050607Z-yyyyy",
    .package = "pins"
  )
  v$metadata$required_pkgs <- c("pingr", "caret")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = file.path(tmp_dir, "plumber.R"))
  vetiver_write_docker(
    v,
    file.path(tmp_dir, "plumber.R"),
    tmp_dir,
    rspm = FALSE
  )
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create Dockerfile with no packages", {
  skip_on_cran()
  mock_version_name <- mockery::mock("20130104T050607Z-xxxxx")
  local_mocked_bindings(version_name = mock_version_name, .package = "pins")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = file.path(tmp_dir, "plumber.R"))
  vetiver_write_docker(v, file.path(tmp_dir, "plumber.R"), tmp_dir)
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create Dockerfile with specific port", {
  skip_on_cran()
  mock_version_name <- mockery::mock("20130204T050607Z-yyyyy")
  local_mocked_bindings(version_name = mock_version_name, .package = "pins")
  v$metadata$required_pkgs <- c("pingr", "caret")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = file.path(tmp_dir, "plumber.R"))
  vetiver_write_docker(
    v,
    file.path(tmp_dir, "plumber.R"),
    tmp_dir,
    port = 'as.numeric(Sys.getenv("PORT"))',
    expose = FALSE
  )
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("No sys deps", {
  skip_on_cran()
  # Data package; should always have 0 sys reqs
  reqs <- glue_sys_reqs("ggplot2movies")
  expect_length(reqs, 0)
})

test_that("create all files needed for Docker", {
  skip_on_cran()
  mock_version_name <- mockery::mock("20130304T050607Z-zzzzz")
  local_mocked_bindings(version_name = mock_version_name, .package = "pins")
  v$metadata$required_pkgs <- c("pingr", "caret")
  vetiver_pin_write(b, v)
  vetiver_prepare_docker(
    b,
    "cars1",
    path = tmp_dir,
    predict_args = list(path = "cars"),
    docker_args = list(rspm = FALSE)
  )

  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "plumber.R")), sep = "\n"),
    transform = redact_vetiver
  )
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.workflow <- function(model) {
  spec <- workflows::extract_spec_parsnip(model)
  glue("A {spec$engine} {spec$mode} modeling workflow")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.workflow <- function(model, metadata) {
  reqs <- required_pkgs(model)
  reqs <- sort(unique(c(reqs, "workflows")))
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.workflow <- function(model, ...) {
  mold <- workflows::extract_mold(model)
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.workflow <- function(model) {
  if (!workflows::is_trained_workflow(model)) {
    rlang::abort("Your `model` object is not a trained workflow.")
  }
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.workflow <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.workflow <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' Create a vetiver input data prototype
#'
#' Optionally find and return an input data prototype for a model.
#'
#' @details
#' These are developer-facing functions, useful for supporting new model types.
#' A [vetiver_model()] object optionally stores an input data prototype for
#' checking at prediction time.
#'
#' - The default for `save_prototype`, `TRUE`, finds an input data prototype (a
#' zero-row slice of the training data) via [vetiver_ptype()].
#' - `save_prototype = FALSE` opts out of storing any input data prototype.
#' - You may pass your own data to `save_prototype`, but be sure to check that it
#' has the same structure as your training data, perhaps with
#' [hardhat::scream()].
#'
#' @inheritParams vetiver_model
#'
#' @return A `vetiver_ptype` method returns a zero-row dataframe, and
#' `vetiver_create_ptype()` returns either such a zero-row dataframe, `NULL`,
#' or the dataframe passed to `save_prototype`.
#'
#' @examples
#'
#' cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
#'
#' vetiver_create_ptype(cars_lm, TRUE)
#'
#' ## calls the right method for `model` via:
#' vetiver_ptype(cars_lm)
#'
#' ## can also turn off prototype
#' vetiver_create_ptype(cars_lm, FALSE)
#' @examplesIf rlang::is_installed("ranger")
#' ## some models require that you pass in training features
#' cars_rf <- ranger::ranger(mpg ~ ., data = mtcars)
#' vetiver_ptype(cars_rf, prototype_data = mtcars[,-1])
#'
#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype <- function(model, ...) {
  UseMethod("vetiver_ptype")
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.default <- function(model, ...) {
  abort(
    "There is no method available to create a 0-row input data prototype for `model`."
  )
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_create_ptype <- function(model, save_prototype, ...) {
  check_dots_used()
  if (isTRUE(save_prototype)) {
    ptype <- vetiver_ptype(model, ...)
  } else if (isFALSE(save_prototype)) {
    ptype <- NULL
  } else if (rlang::inherits_any(save_prototype, "data.frame")) {
    return(save_prototype)
  } else {
    abort("The `save_prototype` argument must be TRUE, FALSE, or a dataframe.")
  }
  ptype
}

check_ptype_data <- function(dots, call = rlang::caller_env()) {
  if (!rlang::has_name(dots, "prototype_data")) {
    abort(
      c(
        "No `prototype_data` available to create an input data prototype",
        "Pass at least one row of training features as `prototype_data`",
        "See the documentation for `vetiver_ptype()`"
      ),
      call = call
    )
  }
}


preds_lm_ish <- function(model) {
  .terms <- terms(model)
  terms_matrix <- attr(.terms, "factors")
  terms_names <- colnames(terms_matrix)
  terms_exprs <- parse_exprs(terms_names)
  has_interactions <- map_lgl(terms_exprs, expr_contains, what = as.name(":"))
  terms_names[!has_interactions]
}

expr_contains <- function(expr, what) {
  switch(
    typeof(expr),
    symbol = identical(expr, what),
    call = call_contains(expr, what),
    language = call_contains(expr, what),
    FALSE
  )
}

call_contains <- function(expr, what, call = rlang::caller_env()) {
  if (length(expr) == 0L) {
    abort("Internal error, `expr` should be at least length 1.", call = call)
  }

  # Recurse into elements
  contains <- map_lgl(as.list(expr), expr_contains, what = what)
  any(contains)
}

b <- board_temp()
cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)

test_that("default ptype", {
  expect_equal(
    vetiver_create_ptype(cars_lm, TRUE),
    vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:3]), 0)
  )
})

test_that("default ptype, check vetiver_slice_zero", {
  expect_equal(
    vetiver_create_ptype(cars_lm, TRUE),
    vetiver_ptype(cars_lm)
  )
})


test_that("ptype = FALSE", {
  expect_equal(
    vetiver_create_ptype(cars_lm, FALSE),
    NULL
  )
})

test_that("ptype for model with interactions", {
  cars_interaction <- lm(mpg ~ cyl * vs + disp, data = mtcars)
  expect_equal(
    vetiver_create_ptype(cars_interaction, TRUE),
    vctrs::vec_slice(tibble::as_tibble(mtcars[, c(2, 8, 3)]), 0)
  )
})

test_that("custom ptype", {
  expect_equal(
    vetiver_create_ptype(cars_lm, mtcars[3:10, 2:3]),
    mtcars[3:10, 2:3]
  )
})

test_that("bad ptype", {
  expect_snapshot(
    vetiver_create_ptype(cars_lm, "potato"),
    error = TRUE
  )
})

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)

# Packages needed to generate model predictions
if (FALSE) {
    library(dials)
    library(dplyr)
    library(LiblineaR)
    library(parsnip)
    library(purrr)
    library(recipes)
    library(rlang)
    library(rsample)
    library(tibble)
    library(tidyr)
    library(tune)
    library(vctrs)
    library(workflows)
    library(yardstick)
}
b <- board_folder(path = "/tmp/test")
v <- vetiver_pin_read(b, "biv_svm", version = "20220211T213615Z-182ed")

#* @plumber
function(pr) {
    pr |> vetiver_api(v)
}

#' Fully attach or load packages for making model predictions
#'
#' These are developer-facing functions, useful for supporting new model types.
#' Some models require one or more R packages to be fully attached to make
#' predictions, and some require only that the namespace of one or more R
#' packages is loaded.
#'
#' @details These two functions will attempt either to:
#'
#' - fully attach or
#' - load
#'
#' the namespace of the `pkgs` vector of package names, preserving the current
#' random seed.
#'
#' To learn more about load vs. attach, read the ["Dependencies" chapter of
#' *R Packages*](https://r-pkgs.org/dependencies-mindset-background.html#sec-dependencies-attach-vs-load).
#' For deploying a model, it is likely safer to fully attach needed packages
#' but that comes with the risk of naming conflicts between packages.
#'
#' @param pkgs A character vector of package names to load or fully attach.
#'
#' @return An invisible `TRUE`.
#' @family namespace
#' @export
#'
#' @examples
#' ## succeed
#' load_pkgs(c("knitr", "readr"))
#' attach_pkgs(c("knitr", "readr"))
#'
#' ## fail
#' try(load_pkgs(c("bloopy", "readr")))
#' try(attach_pkgs(c("bloopy", "readr")))
#'
attach_pkgs <- function(pkgs) {
  attached <- paste0("package:", pkgs) %in% search()
  pkgs <- pkgs[!attached]
  namespace_handling(pkgs, attachNamespace, "Package(s) could not be attached:")
}

#' @export
#' @rdname attach_pkgs
load_pkgs <- function(pkgs) {
  loaded <- map_lgl(pkgs, isNamespaceLoaded)
  pkgs <- pkgs[!loaded]
  namespace_handling(pkgs, loadNamespace, "Namespace(s) could not be loaded:")
}

namespace_handling <- function(
  pkgs,
  func,
  error_msg,
  call = rlang::caller_env()
) {
  safe_load <- safely(withr::with_preserve_seed(func))
  did_load <- map(pkgs, safe_load)
  bad <- compact(map(did_load, "error"))
  bad <- map_chr(bad, "package")
  if (length(bad) >= 1) {
    abort(c(error_msg, bad), call = call)
  }

  invisible(TRUE)
}

test_that("can attach pkgs", {
  skip_on_cran()
  expect_snapshot(attach_pkgs(c("knitr", "readr")))
})

test_that("can fail on a single pkg", {
  skip_on_cran()
  expect_snapshot(attach_pkgs(c("potato")), error = TRUE)
})

test_that("can fail on multiple pkgs", {
  skip_on_cran()
  expect_snapshot(attach_pkgs(c("potato", "bloopy")), error = TRUE)
  expect_snapshot(attach_pkgs(c("potato", "readr")), error = TRUE)
})

#' Write a deployable Plumber file for a vetiver model
#'
#' Use `vetiver_write_plumber()` to create a `plumber.R` file for a
#' [vetiver_model()] that has been versioned and stored via
#' [vetiver_pin_write()].
#'
#' @inheritParams pins::pin_read
#' @inheritParams vetiver_model
#' @param ... Other arguments passed to [vetiver_api()] such as the endpoint
#' `path` or prediction `type`.
#' @param file A path to write the Plumber file. Defaults to `plumber.R` in the
#' working directory. See [plumber::plumb()] for naming precedence rules.
#' @param rsconnect Create a Plumber file with features needed for [Posit
#' Connect](https://posit.co/products/enterprise/connect)? Defaults to `TRUE`.
#' @param additional_pkgs Any additional R packages that need to be **attached**
#' via [library()] to run your API, as a character vector.
#'
#' @details
#' By default, this function will find and use the latest version of your
#' vetiver model; the model API (when deployed) will be linked to that specific
#' version. You can override this default behavior by choosing a specific
#' `version`.
#'
#' @return
#' The content of the `plumber.R` file, invisibly.
#'
#' @export
#'
#' @examplesIf rlang::is_installed("plumber")
#' library(pins)
#' tmp <- tempfile()
#' b <- board_temp(versioned = TRUE)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' vetiver_write_plumber(b, "cars_linear", file = tmp)
#'
vetiver_write_plumber <- function(
  board,
  name,
  version = NULL,
  ...,
  file = "plumber.R",
  rsconnect = TRUE,
  additional_pkgs = character(0)
) {
  rlang::check_installed("plumber")
  plumber_dots <- rlang::list2(...)

  if (board$versioned) {
    if (is_null(version)) {
      version <- pins::pin_versions(board, name)
      version <- choose_version(version)
    }
    pin_read <- glue(
      'v <- vetiver_pin_read(b, "{name}", version = "{version}")'
    )
    v <- vetiver_pin_read(board, name, version = version)
  } else {
    pin_read <- glue('v <- vetiver_pin_read(b, "{name}")')
    v <- vetiver_pin_read(board, name)
  }

  write_extra_requirements(v$model, file)

  infra_pkgs <- sort(c(infra_pkgs, additional_pkgs))
  load_infra_pkgs <- glue_collapse(glue("library({infra_pkgs})"), sep = "\n")
  load_required_pkgs <- glue_required_pkgs(v$metadata$required_pkgs, rsconnect)

  ## rlang::expr_deparse won't work for board_url, but
  ## base deparse won't work for S3 and other complex boards:
  if (inherits(board, "pins_board_url")) {
    board <- deparse(pins::board_deparse(board))
  } else {
    board <- rlang::expr_deparse(pins::board_deparse(board))
  }
  board <- glue('b <- {board}')

  if (rlang::is_empty(plumber_dots)) {
    pr_predict <- "pr |> vetiver_api(v)"
  } else {
    pr_predict <- expr(pr |> vetiver_api(v, !!!plumber_dots))
    pr_predict <- expr_deparse(pr_predict)
  }

  plumber <- glue(
    "\n
         #* @plumber
         function(pr) {{
             {pr_predict}
         }}
         "
  )

  ret <- compact(list(
    "# Generated by the vetiver package; edit with care\n",
    load_infra_pkgs,
    load_required_pkgs,
    board,
    pin_read,
    plumber
  ))
  readr::write_lines(ret, file = file)
}

infra_pkgs <- c("pins", "plumber", "rapidoc", "vetiver")


glue_required_pkgs <- function(required_pkgs, rsconnect) {
  if (!is_null(required_pkgs) && rsconnect) {
    required_pkgs <- sort(required_pkgs)
    required_pkgs <- glue_collapse(
      glue("    library({required_pkgs})"),
      sep = "\n"
    )
    load_required_pkgs <- glue(
      "\n
            # Packages needed to generate model predictions
            if (FALSE) {{
            {required_pkgs}
            }}
            "
    )
    return(load_required_pkgs)
  }
  NULL
}

choose_version <- function(df) {
  if (has_name(df, "active")) {
    version <- vec_slice(df, df$active)
  } else if (has_name(df, "created")) {
    idx <- head(order(df$created, decreasing = TRUE), 1)
    version <- vec_slice(df, idx)
  } else {
    version <- vec_slice(df, 1)
    warn(
      c(
        "Pinned vetiver model has no active version and no datetime on versions",
        "Do you need to check your pinned model?",
        glue('Using version {version[["version"]]}')
      )
    )
  }
  version[["version"]]
}

write_extra_requirements <- function(model, file) {
  model <- bundle::unbundle(model)
  path_to_py_requirements <- vetiver_python_requirements(model)
  file_copy_requirements(path_to_py_requirements, file, "requirements.txt")
  path_to_renviron_requirements <- vetiver_renviron_requirements(model)
  file_copy_requirements(path_to_renviron_requirements, file, ".Renviron")
  TRUE
}

file_copy_requirements <- function(requirements, plumber_file, new_name) {
  if (!is.null(requirements)) {
    fs::file_copy(
      requirements,
      fs::path(fs::path_dir(plumber_file), new_name),
      overwrite = TRUE
    )
  }
  requirements
}

#' Use extra files required for deployment
#'
#' Create files required for deploying an app generated via
#' [vetiver_write_plumber()], such as a Python `requirements.txt` or an
#' `.Renviron`
#'
#' @inheritParams vetiver_model
#' @export
#' @keywords internal
vetiver_python_requirements <- function(model) {
  UseMethod("vetiver_python_requirements")
}

#' @rdname vetiver_python_requirements
#' @export
vetiver_python_requirements.default <- function(model) {
  NULL
}

#' @rdname vetiver_python_requirements
#' @export
vetiver_renviron_requirements <- function(model) {
  UseMethod("vetiver_renviron_requirements")
}

#' @rdname vetiver_python_requirements
#' @export
vetiver_renviron_requirements.default <- function(model) {
  NULL
}

skip_if_not_installed("plumber")

test_that("create plumber.R with no packages", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  local_mocked_bindings(
    version_name = function(metadata) "20130104T050607Z-xxxxx",
    .package = "pins"
  )
  tmp <- tempfile()
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R for complicated board", {
  skip_on_cran()
  b <- pins::new_board(
    "pins_board_s3",
    api = 1L,
    cache = board_cache_path("pins"),
    versioned = FALSE,
    bucket = "foo",
    svc = list(
      .internal = list(
        config = list(
          credentials = list(creds = list(profile = 4)),
          region = 6,
          endpoint = 8
        )
      )
    )
  )
  local_mocked_bindings(vetiver_pin_read = function(board, name, version) v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars1", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R for URL board", {
  skip_on_cran()
  b <- pins::new_board(
    "pins_board_url",
    api = 1L,
    cache = board_cache_path("pins"),
    versioned = FALSE,
    bucket = "foo",
    urls = c(foo = "foo", bar = "bar", baz = "baz")
  )
  local_mocked_bindings(vetiver_pin_read = function(board, name, version) v)
  tmp <- tempfile()
  vetiver_write_plumber(b, "cars1", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R with packages", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  local_mocked_bindings(
    version_name = function(metadata) "20130204T050607Z-xxxxx",
    .package = "pins"
  )
  tmp <- tempfile()
  v$metadata$required_pkgs <- c("beepr", "janeaustenr")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = tmp)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R with extra infra packages", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  local_mocked_bindings(
    version_name = function(metadata) "20130204T050607Z-xyxy",
    .package = "pins"
  )
  tmp <- tempfile()
  vetiver_pin_write(b, v)
  vetiver_write_plumber(
    b,
    "cars1",
    file = tmp,
    additional_pkgs = c("beepr", "janeaustenr")
  )
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R with rsconnect = FALSE", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  local_mocked_bindings(
    version_name = function(metadata) "20130304T050607Z-xxxxx",
    .package = "pins"
  )
  tmp <- tempfile()
  v$metadata$required_pkgs <- c("beepr", "janeaustenr")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(b, "cars1", file = tmp, rsconnect = FALSE)
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

test_that("create plumber.R with args in dots", {
  skip_on_cran()
  b <- board_folder(path = tmp_dir)
  local_mocked_bindings(
    version_name = function(metadata) "20130404T050607Z-xxxxx",
    .package = "pins"
  )
  tmp <- tempfile()
  v$metadata$required_pkgs <- c("beepr", "janeaustenr")
  vetiver_pin_write(b, v)
  vetiver_write_plumber(
    b,
    "cars1",
    debug = TRUE,
    endpoint = "/predict2",
    type = "numeric",
    file = tmp
  )
  expect_snapshot(
    cat(readr::read_lines(tmp), sep = "\n"),
    transform = redact_vetiver
  )
})

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.xgb.Booster <- function(model) {
  if (is_null(model$params)) {
    if (is_null(attr(model, "params"))) {
      ret <- "An xgboost model"
    } else {
      ret <- glue("An xgboost {attr(model, 'params')$objective} model")
    }
  } else {
    ret <- glue("An xgboost {model$params$objective} model")
  }

  ret
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.xgb.Booster <- function(model, metadata) {
  vetiver_meta(metadata, required_pkgs = "xgboost")
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.xgb.Booster <- function(model, ...) {
  if (is_null(model$nfeatures)) {
    feature_names <- xgboost::getinfo(model, "feature_name")
    pred_names <- matrix(
      NA_real_,
      ncol = length(feature_names),
      dimnames = list("", feature_names)
    )
  } else {
    pred_names <- matrix(
      NA_real_,
      ncol = model$nfeatures,
      dimnames = list("", model$feature_names)
    )
  }
  ptype <- vctrs::vec_ptype(pred_names)
  tibble::as_tibble(ptype)
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.xgb.Booster <- function(model) {
  bundle::bundle(model)
}

#' @rdname handler_startup
#' @export
handler_startup.xgb.Booster <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.xgb.Booster <- function(vetiver_model, ...) {
  ptype <- vetiver_model$prototype

  function(req) {
    newdata <- req$body
    if (!is_null(ptype)) {
      newdata <- vetiver_type_convert(newdata, ptype)
      newdata <- hardhat::scream(newdata, ptype)
    }
    newdata <- xgboost::xgb.DMatrix(as.matrix(newdata))
    ret <- predict(vetiver_model$model, newdata = newdata, ...)
    list(.pred = ret)
  }
}

#' R Markdown format for model monitoring dashboards
#'
#' @inheritParams pins::pin_write
#' @param pins A list containing `board`, `name`, and `version`, as in
#' [pins::pin_read()]
#' @param display_pins Should the dashboard display a link to the pin(s)?
#' Defaults to `TRUE`, but only creates a link if the pin contains a URL in
#' its metadata.
#' @param ... Arguments passed to [flexdashboard::flex_dashboard()]
#'
#' @details The `vetiver_dashboard()` function is a specialized type of
#' \pkg{flexdashboard}. See the flexdashboard website for additional
#' documentation:
#'  \href{https://rstudio.github.io/flexdashboard/}{https://rstudio.github.io/flexdashboard/}
#'
#' Before knitting the example `vetiver_dashboard()` template, execute the
#' helper function `pin_example_kc_housing_model()` to set up demonstration
#' model and metrics pins needed for the monitoring demo. This function will:
#'
#' - fit an example model to training data
#' - pin the vetiver model to your own [pins::board_local()]
#' - compute metrics from testing data
#' - pin these metrics to the same local board
#'
#' These are the steps you need to complete before setting up monitoring your
#' real model.
#'
#' @export
vetiver_dashboard <- function(pins = list(), display_pins = TRUE, ...) {
  rlang::check_installed("flexdashboard")
  dashboard_dots <- rlang::list2(...)
  v <- dashboard_read_version(
    pins$board,
    pins$name,
    pins$version
  )

  if (display_pins && !is_null(v$metadata$url)) {
    dashboard_dots <- vetiver_modify_navbar(dashboard_dots, v$metadata$url)
  }
  if (is_null(dashboard_dots$favicon)) {
    vetiver_fav <- system.file("favicon.ico", package = "vetiver")
    dashboard_dots <- modifyList(dashboard_dots, list(favicon = vetiver_fav))
  }

  rmarkdown::output_format(
    knitr = rmarkdown::knitr_options(
      opts_knit = list(vetiver_dashboard.pins = pins),
      # require to keep flexdashboard one
      opts_chunk = list(),
      knit_hooks = list(),
      opts_hooks = list(),
      opts_template = list()
    ),
    pandoc = NULL,
    base_format = rlang::inject(flexdashboard::flex_dashboard(
      !!!dashboard_dots
    ))
  )
}


dashboard_read_version <- function(
  board,
  name,
  version,
  call = rlang::caller_env()
) {
  if (!pins::pin_exists(board, name)) {
    rlang::abort(
      c(
        glue("The pin {glue::double_quote(name)} does not exist on your board"),
        "*" = "Check the `pins` params in your dashboard YAML",
        "i" = glue(
          "To knit the example vetiver monitoring dashboard, \\
                       execute `vetiver::pin_example_kc_housing_model()` to \\
                       set up demo model and metrics pins"
        )
      ),
      call = call
    )
  }
  if (board$versioned) {
    if (is_null(version)) {
      version <- pins::pin_versions(board, name)
      version <- choose_version(version)
    }
    v <- vetiver_pin_read(board, name, version = version)
  } else {
    v <- vetiver_pin_read(board, name)
  }
  v
}

vetiver_modify_navbar <- function(dots, url) {
  if (has_name(dots, "navbar")) {
    navbar <- dots$navbar
  } else {
    navbar <- NULL
  }
  pin_navbar <- list(
    title = "Model Pin",
    icon = "fa-map-marker",
    href = url,
    target = "_blank"
  )
  navbar <- append(navbar, list(pin_navbar))
  modifyList(dots, list(navbar = navbar))
}


#' @rdname vetiver_dashboard
#' @export
get_vetiver_dashboard_pins <- function() {
  knitr::opts_knit$get("vetiver_dashboard.pins")
}

#' @rdname vetiver_dashboard
#' @export
pin_example_kc_housing_model <- function(
  board = pins::board_local(),
  name = "seattle_rf"
) {
  df <- mlr3data::kc_housing
  df_train <- dplyr::filter(df, date < "2014-11-01")
  df_test <- dplyr::filter(df, date >= "2014-11-01", date < "2015-01-01")
  rf_spec <- parsnip::rand_forest(trees = 200, mode = "regression")

  rf_fit <-
    workflows::workflow(
      price ~ bedrooms + bathrooms + sqft_living + yr_built,
      rf_spec
    ) |>
    parsnip::fit(df_train)

  v <- vetiver_model(rf_fit, name)
  vetiver_pin_write(board, v)

  old_metrics <-
    parsnip::augment(v, df_test |> dplyr::arrange(date)) |>
    vetiver_compute_metrics(date, "week", price, .pred)

  pins::pin_write(board, old_metrics, paste(name, "metrics", sep = "_"))
}

skip_if_not_installed("rmarkdown", minimum_version = "2.12")

test_that("templates exist", {
  expect_equal(
    rmarkdown::available_templates("vetiver"),
    c("vetiver_dashboard", "vetiver_model_card")
  )
})

describe("vetiver_dashboard()", {
  # don't run on cran because pandoc is required
  skip_on_cran()
  skip_if_not_installed("flexdashboard")

  rmd <- withr::local_tempfile(fileext = ".Rmd")
  html <- withr::local_tempfile(fileext = ".html")
  path <- expect_invisible(rmarkdown::draft(
    rmd,
    template = "vetiver_dashboard",
    package = "vetiver",
    create_dir = FALSE,
    edit = FALSE
  ))

  it("can create dashboard", {
    expect_identical(rmd, path)
    expect_true(file.exists(rmd))
  })

  it("errors without pin", {
    b <- pins::board_temp()
    expect_snapshot(
      vetiver_dashboard(
        pins = list(board = b, name = "seattle_rf", version = NULL)
      ),
      error = TRUE
    )
  })

  it("can render dashboard", {
    b <- pins::board_temp()
    pin_example_kc_housing_model(b)
    output_format <- vetiver_dashboard(
      pins = list(board = b, name = "seattle_rf", version = NULL)
    )
    rmarkdown::render(rmd, output_format = output_format, output_file = html)
    expect_true(file.exists(html))
  })
})

#' Metadata constructors for `vetiver_model()` object
#'
#' These are developer-facing functions, useful for supporting new model types.
#' The metadata stored in a [vetiver_model()] object has four elements:
#' - `$user`, the metadata supplied by the user
#' - `$version`, the version of the pin (which can be `NULL` before pinning)
#' - `$url`, the URL where the pin is located, if any
#' - `$required_pkgs`, a character string of R packages required for prediction
#'
#' @inheritParams vetiver_model
#' @param user Metadata supplied by the user
#' @param version Version of the pin
#' @param url URL for the pin, if any
#' @param required_pkgs Character string of R packages required for prediction
#'
#' @return The `vetiver_meta()` constructor returns a list. The
#' `vetiver_create_meta` function returns a `vetiver_meta()` list.
#'
#' @examples
#' vetiver_meta()
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' vetiver_create_meta(cars_lm, list())
#'
#' @rdname vetiver_create_meta
#' @export
vetiver_meta <- function(
  user = list(),
  version = NULL,
  url = NULL,
  required_pkgs = NULL
) {
  list(user = user, version = version, url = url, required_pkgs = required_pkgs)
}


#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta <- function(model, metadata) {
  UseMethod("vetiver_create_meta")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.default <- function(model, metadata) {
  vetiver_meta(metadata)
}

#' Create a Plumber API to predict with a deployable `vetiver_model()` object
#'
#' Use `vetiver_api()` to add a POST endpoint for predictions from a
#' trained [vetiver_model()] to a Plumber router.
#'
#' @param pr A Plumber router, such as from [plumber::pr()].
#' @param vetiver_model A deployable [vetiver_model()] object
#' @param ... Other arguments passed to `predict()`, such as prediction `type`
#' @param check_prototype Should the input data prototype stored in
#' `vetiver_model` (used for visual API documentation) also be used to check
#' new data at prediction time? Defaults to `TRUE`.
#' @param check_ptype `r lifecycle::badge("deprecated")`
#' @param all_docs Should the interactive visual API documentation be created
#' for _all_ POST endpoints in the router `pr`? This defaults to `TRUE`, and
#' assumes that all POST endpoints use the [vetiver_model()] input data
#' prototype.
#' @inheritParams plumber::pr_post
#' @inheritParams plumber::pr_set_debug
#'
#' @details You can first store and version your [vetiver_model()] with
#' [vetiver_pin_write()], and then create an API endpoint with `vetiver_api()`.
#'
#' Setting `debug = TRUE` may expose any sensitive data from your model in
#' API errors.
#'
#' Several GET endpoints will also be added to the router `pr`, depending on the
#' characteristics of the model object:
#'
#' - a `/pin-url` endpoint to return the URL of the pinned model
#' - a `/metadata` endpoint to return any metadata stored with the model
#' - a `/ping` endpoint for the API health
#' - a `/prototype` endpoint for the model's input data prototype (use
#' [cereal::cereal_from_json()]) to convert this back to a [vctrs ptype](https://vctrs.r-lib.org/articles/type-size.html)
#'
#' The function `vetiver_api()` uses:
#' - `vetiver_pr_post()` for endpoint definition and
#' - `vetiver_pr_docs()` to create visual API documentation
#'
#' These modular functions are available for more advanced use cases.
#'
#' @return A Plumber router with the prediction endpoint added.
#'
#' @examplesIf rlang::is_installed("plumber")
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#'
#' library(plumber)
#' pr() |> vetiver_api(v)
#' ## is the same as:
#' pr() |> vetiver_pr_post(v) |> vetiver_pr_docs(v)
#' ## for either, next, pipe to `pr_run()`
#'
#' @export
vetiver_api <- function(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...
) {
  # `force()` all `...` arguments early; https://github.com/tidymodels/vetiver/pull/20
  rlang::list2(...)
  vetiver_model$model <- bundle::unbundle(vetiver_model$model)
  pr <- vetiver_pr_post(
    pr = pr,
    vetiver_model = vetiver_model,
    path = path,
    debug = debug,
    ...
  )

  pr <- vetiver_pr_docs(pr = pr, vetiver_model = vetiver_model, path = path)

  pr
}

#' @rdname vetiver_api
#' @export
vetiver_pr_post <- function(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...,
  check_prototype = TRUE,
  check_ptype = deprecated()
) {
  rlang::check_installed("plumber")
  # `force()` all `...` arguments early; https://github.com/tidymodels/vetiver/pull/20
  rlang::list2(...)

  if (lifecycle::is_present(check_ptype)) {
    lifecycle::deprecate_soft(
      "0.2.0",
      "vetiver_pr_post(check_ptype)",
      "vetiver_pr_post(check_prototype)"
    )
    check_prototype <- check_ptype
  }

  handler_startup(vetiver_model)
  pr <- plumber::pr_set_debug(pr, debug = debug)
  pr <- plumber::pr_set_serializer(
    pr,
    serializer = plumber::serializer_unboxed_json(null = "null")
  )
  pr <- vetiver_pr_ping(pr)
  pr <- vetiver_pr_pin_url(pr, vetiver_model)
  pr <- vetiver_pr_metadata(pr, vetiver_model)
  if (!check_prototype) {
    vetiver_model$prototype <- NULL
  }
  pr <- vetiver_pr_prototype(pr, vetiver_model)
  pr <- plumber::pr_post(
    pr,
    path = path,
    handler = handler_predict(vetiver_model, ...)
  )
  pr
}

vetiver_pr_ping <- function(pr) {
  plumber::pr_get(
    pr,
    path = "/ping",
    handler = function() {
      list(status = "online", time = Sys.time())
    }
  )
}

vetiver_pr_pin_url <- function(pr, vetiver_model) {
  if (!is_null(vetiver_model$metadata$url)) {
    pr <- plumber::pr_get(
      pr,
      path = "/pin-url",
      handler = function() vetiver_model$metadata$url
    )
  }
  pr
}

vetiver_pr_metadata <- function(pr, vetiver_model) {
  plumber::pr_get(
    pr,
    path = "/metadata",
    handler = function() vetiver_model$metadata
  )
}

vetiver_pr_prototype <- function(pr, vetiver_model) {
  if (!is_null(vetiver_model$prototype)) {
    pr <- plumber::pr_get(
      pr,
      path = "/prototype",
      handler = function() {
        purrr::map(vetiver_model$prototype, cereal::cereal_encode)
      }
    )
  }
  pr
}

#' @rdname vetiver_api
#' @export
vetiver_pr_docs <- function(
  pr,
  vetiver_model,
  path = "/predict",
  all_docs = TRUE
) {
  rlang::check_installed("plumber")
  loadNamespace("rapidoc")
  modify_spec <- function(spec) api_spec(spec, vetiver_model, path, all_docs)
  pr <- plumber::pr_set_api_spec(pr, api = modify_spec)
  logo <-
    '<img slot="logo" src="../logo/vetiver.png"
         width=55px style=\"margin-left:7px\"/>'
  pr <- plumber::pr_static(pr, "/logo", system.file(package = "vetiver"))
  pr <- plumber::pr_set_docs(
    pr,
    "rapidoc",
    slots = logo,
    heading_text = paste("vetiver", utils::packageVersion("vetiver")),
    header_color = "#F2C6AC",
    primary_color = "#8C2D2D"
  )
  pr
}

#' Create a Plumber API to predict with a deployable `vetiver_model()` object
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' This function was deprecated to use [vetiver_api] directly instead.
#'
#' @inheritParams vetiver_api
#' @export
#' @keywords internal
vetiver_pr_predict <- function(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...
) {
  lifecycle::deprecate_stop(
    "0.1.2",
    "vetiver_pr_predict()",
    "vetiver_api()"
  )
}


local_plumber_session <- function(
  pr,
  port,
  docs = FALSE,
  env = parent.frame()
) {
  rlang::check_installed("plumber")
  rs <- callr::r_session$new()
  rs$call(
    function(pr, port, docs) {
      plumber::pr_run(pr = pr, port = port, docs = docs)
    },
    args = list(pr = pr, port = port, docs = docs)
  )
  withr::defer(rs$close(), envir = env)
  rs
}

skip_on_cran()
skip_if_not_installed("pingr")
skip_if_not_installed("httr")
skip_if_not_installed("httpuv")
skip_if_not_installed("plumber")

library(plumber)

pr <- pr() |> vetiver_api(v, debug = TRUE)
rs <- local_plumber_session(pr, port)

## on GH actions, it can take A WHILE for the API to come up on some architectures:
for (i in 1:100) {
  if (pingr::is_up(root_path, port)) {
    break
  }
  Sys.sleep(0.1)
}

test_that("router has health check endpoint", {
  r <- httr::GET(root_path, port = port, path = "ping")
  expect_equal(r$status_code, 200)
})

test_that("router has metadata endpoint", {
  r <- httr::GET(root_path, port = port, path = "metadata")
  metadata <- jsonlite::fromJSON(httr::content(
    r,
    as = "text",
    encoding = "UTF-8"
  ))
  expect_equal(r$status_code, 200)
  expect_equal(names(metadata), c("user", "version", "url", "required_pkgs"))
})

test_that("router has prototype endpoint", {
  r <- httr::GET(root_path, port = port, path = "prototype")
  prototype <- httr::content(r, as = "text", encoding = "UTF-8")
  expect_equal(r$status_code, 200)
  expect_equal(
    jsonlite::fromJSON(prototype),
    purrr::map(v$prototype, cereal::cereal_encode)
  )
  expect_equal(cereal::cereal_from_json(prototype), v$prototype)
})

test_that("can predict on basic vetiver router", {
  endpoint <- vetiver_endpoint(paste0(root_path, ":", port, "/predict"))
  expect_s3_class(endpoint, "vetiver_endpoint")
  expect_snapshot(print(endpoint), transform = redact_port)

  preds <- predict(endpoint, mtcars[10:17, 2:3])
  expect_s3_class(preds, "tbl_df")
  expect_equal(nrow(preds), 8)
  expect_equal(ncol(preds), 1)

  aug <- augment(endpoint, mtcars[10:17, 2:3])
  expect_s3_class(aug, "tbl_df")
  expect_equal(nrow(aug), 8)
  expect_equal(ncol(aug), 3)
})

test_that("can predict with single or NA values", {
  endpoint <- vetiver_endpoint(paste0(root_path, ":", port, "/predict"))

  preds1 <- predict(endpoint, mtcars[10, 2:3])
  expect_s3_class(preds1, "tbl_df")
  expect_equal(nrow(preds1), 1)
  expect_equal(ncol(preds1), 1)

  preds2 <- predict(
    endpoint,
    data.frame(cyl = c(NA_real_, NA_real_), disp = c(100, 200))
  )
  expect_s3_class(preds2, "tbl_df")
  expect_equal(nrow(preds2), 2)
  expect_equal(ncol(preds2), 1)
})

test_that("get correct errors", {
  endpoint <- vetiver_endpoint(paste0(root_path, ":", port, "/predict"))
  expect_snapshot(predict(endpoint, mtcars[, 2:4]), error = TRUE)
  expect_snapshot(predict(endpoint, mtcars[, 3:5]), error = TRUE)
})

skip_if_not_installed("plumber")
library(plumber)

test_that("default endpoint", {
  p <- pr() |> vetiver_api(v)
  expect_equal(
    names(p$routes),
    c("logo", "metadata", "ping", "predict", "prototype")
  )
  expect_equal(
    map_chr(p$routes[-1], "verbs"),
    c(metadata = "GET", ping = "GET", predict = "POST", prototype = "GET")
  )
})

test_that("old function is deprecated", {
  expect_snapshot_error(p <- pr() |> vetiver_pr_predict(v))
})

test_that("default endpoint via modular functions", {
  p1 <- pr() |> vetiver_api(v)
  p2 <- pr() |> vetiver_pr_post(v) |> vetiver_pr_docs(v)
  expect_equal(p1$endpoints, p2$endpoints)
  expect_equal(p1$routes, p2$routes)
})

test_that("pin URL endpoint", {
  v$metadata <- list(url = "potato")
  p <- pr() |> vetiver_api(v)
  expect_equal(
    names(p$routes),
    c("logo", "metadata", "pin-url", "ping", "predict", "prototype")
  )
  expect_equal(
    map_chr(p$routes[-1], "verbs"),
    c(
      metadata = "GET",
      `pin-url` = "GET",
      ping = "GET",
      predict = "POST",
      prototype = "GET"
    )
  )
})

test_that("default OpenAPI spec", {
  v$metadata <- list(url = "potatoes")
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  expect_equal(
    post_spec$requestBody$content$`application/json`$schema$items,
    list(
      type = "object",
      properties = list(
        cyl = list(type = "number"),
        disp = list(type = "number")
      )
    )
  )
  get_spec1 <- car_spec$paths$`/pin-url`$get
  expect_equal(
    as.character(get_spec1$summary),
    "Get URL of pinned vetiver model"
  )
  get_spec2 <- car_spec$paths$`/metadata`$get
  expect_equal(
    as.character(get_spec2$summary),
    "Get all metadata of pinned vetiver model"
  )
  get_spec3 <- car_spec$paths$`/prototype`$get
  expect_equal(
    as.character(get_spec3$summary),
    "Get input data prototype for vetiver model"
  )
})

test_that("OpenAPI spec is the same for modular functions", {
  v$metadata <- list(url = "potatoes")
  p1 <- pr() |> vetiver_api(v)
  p2 <- pr() |> vetiver_pr_post(v) |> vetiver_pr_docs(v)
  spec1 <- p1$getApiSpec()
  spec2 <- p2$getApiSpec()
  expect_equal(spec1, spec2)
})

test_that("OpenAPI spec for check_prototype = FALSE", {
  expect_snapshot_warning(
    p <- pr() |> vetiver_pr_post(v, check_ptype = FALSE) |> vetiver_pr_docs(v)
  )

  p <- pr() |>
    vetiver_pr_post(v, check_prototype = FALSE) |>
    vetiver_pr_docs(v)
  expect_equal(names(p$routes), c("logo", "metadata", "ping", "predict"))
  expect_equal(
    map_chr(p$routes[-1], "verbs"),
    c(metadata = "GET", ping = "GET", predict = "POST")
  )
  car_spec <- p$getApiSpec()
  post_spec <- car_spec$paths$`/predict`$post

  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  expect_equal(
    names(post_spec$requestBody$content$`application/json`$schema$items),
    c("type", "properties")
  )
})

test_that("OpenAPI spec for save_prototype = FALSE", {
  v1 <- vetiver_model(cars_lm, "cars1", save_prototype = FALSE)
  p <- pr() |> vetiver_api(v1)
  expect_equal(names(p$routes), c("logo", "metadata", "ping", "predict"))
  expect_equal(
    map_chr(p$routes[-1], "verbs"),
    c(metadata = "GET", ping = "GET", predict = "POST")
  )
  car_spec <- p$getApiSpec()
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(as.character(post_spec$summary), "Return predictions from model")
  expect_equal(
    names(post_spec$requestBody$content$`application/json`$schema$items),
    c("type", "properties")
  )
})

test_that("OpenAPI spec with custom ptype", {
  car_ptype <- mtcars[15:16, 2:3]
  v <- vetiver_model(cars_lm, "cars1", save_prototype = car_ptype)
  p <- pr() |> vetiver_api(v)
  car_spec <- p$getApiSpec()
  post_spec <- car_spec$paths$`/predict`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  expect_equal(
    post_spec$requestBody$content$`application/json`$schema$items,
    list(
      type = "object",
      properties = list(
        cyl = list(type = "number"),
        disp = list(type = "number")
      )
    )
  )
  expect_equal(
    post_spec$requestBody$content$`application/json`$schema$example,
    purrr::transpose(car_ptype)
  )
  expect_equal(
    car_spec$paths$`/prototype`$get$summary,
    "Get input data prototype for vetiver model"
  )
})

test_that("OpenAPI spec with additional endpoints", {
  v$metadata <- list(url = "potatoes")

  another_handler <- function(req) {
    newdata <- req$body
    sum(newdata[names(v$prototype)])
  }

  p <- pr() |>
    vetiver_pr_post(v) |>
    pr_post(path = "/sum", handler = another_handler) |>
    pr_get(
      "/",
      function() "Hello World",
      ## to test Connect redirect:
      comments = "This endpoint was added to automatically redirect visitors"
    ) |>
    vetiver_pr_docs(v)

  car_spec <- p$getApiSpec()
  expect_equal(car_spec$path$`/`, NULL)
  expect_true(all(names(car_spec$paths) %in% paste0("/", names(p$routes))))

  post_spec <- car_spec$paths$`/predict`$post
  sum_spec <- car_spec$paths$`/sum`$post
  expect_equal(names(post_spec), c("summary", "requestBody", "responses"))
  expect_equal(names(sum_spec), c("summary", "requestBody", "responses"))
  expect_equal(
    as.character(post_spec$summary),
    "Return predictions from model using 2 features"
  )
  expect_equal(
    as.character(sum_spec$summary),
    "Return /sum from model using 2 features"
  )
  items <- list(
    type = "object",
    properties = list(cyl = list(type = "number"), disp = list(type = "number"))
  )
  expect_equal(
    post_spec$requestBody$content$`application/json`$schema$items,
    items
  )
  expect_equal(
    sum_spec$requestBody$content$`application/json`$schema$items,
    items
  )
})

test_that("debug listens to `is_interactive()`", {
  rlang::with_interactive(value = FALSE, {
    p <- pr() |> vetiver_api(v)
    expect_equal(p$getDebug(), FALSE)
  })
  rlang::with_interactive(value = TRUE, {
    p <- pr() |> vetiver_api(v)
    expect_equal(p$getDebug(), TRUE)
  })
})

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.int_conformal_split <- function(model) {
  spec <- workflows::extract_spec_parsnip(model$wflow)
  glue("A Split Conformal inference with a {spec$engine} {spec$mode} model")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.int_conformal_split <- function(model, metadata) {
  reqs <- required_pkgs(model)
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.int_conformal_split <- function(model, ...) {
  mold <- workflows::extract_mold(model$wflow)
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.int_conformal_split <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.int_conformal_split <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.int_conformal_split <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.int_conformal_full <- function(model) {
  spec <- workflows::extract_spec_parsnip(model$wflow)
  glue("A full Conformal inference with a {spec$engine} {spec$mode} model")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.int_conformal_full <- function(model, metadata) {
  reqs <- required_pkgs(model)
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.int_conformal_full <- function(model, ...) {
  mold <- workflows::extract_mold(model$wflow)
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.int_conformal_full <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.int_conformal_full <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.int_conformal_full <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.int_conformal_quantile <- function(model) {
  spec <- workflows::extract_spec_parsnip(model$wflow)
  glue("A quantile Conformal inference with a {spec$engine} {spec$mode} model")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.int_conformal_quantile <- function(model, metadata) {
  reqs <- required_pkgs(model)
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.int_conformal_quantile <- function(model, ...) {
  mold <- workflows::extract_mold(model$wflow)
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.int_conformal_quantile <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.int_conformal_quantile <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.int_conformal_quantile <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.int_conformal_cv <- function(model) {
  spec <- workflows::extract_spec_parsnip(model$models[[1]])
  n <- length(model$models)
  glue(
    "A {n}-fold CV+ Conformal inference with a {spec$engine} {spec$mode} model"
  )
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.int_conformal_cv <- function(model, metadata) {
  reqs <- required_pkgs(model)
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.int_conformal_cv <- function(model, ...) {
  mold <- workflows::extract_mold(model$models[[1]])
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.int_conformal_cv <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.int_conformal_cv <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.int_conformal_cv <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.luz_module_fitted <- function(model) {
  n_parameters <- lapply(model$model$parameters, function(x) prod(x$shape))
  n_parameters <- do.call(sum, n_parameters)
  n_parameters <- formatC(n_parameters, big.mark = ",", format = "d")
  glue("A luz module with {n_parameters} parameters")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.luz_module_fitted <- function(model, metadata) {
  pkgs <- c("luz", "torch", model$model$required_pkgs)
  vetiver_meta(metadata, required_pkgs = pkgs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.luz_module_fitted <- function(model, ...) {
  rlang::check_dots_used()
  dots <- list(...)
  check_ptype_data(dots)
  ptype <- vctrs::vec_ptype(dots$prototype_data)
  tibble::as_tibble(ptype)
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.luz_module_fitted <- function(model) {
  bundle::bundle(model)
}

#' @rdname handler_startup
#' @export
handler_startup.luz_module_fitted <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.luz_module_fitted <- function(vetiver_model, ...) {
  force(vetiver_model)
  function(req) {
    new_data <- vetiver_type_convert(req$body, vetiver_model$prototype)
    new_data <- if (is.data.frame(new_data)) as.matrix(new_data) else new_data
    preds <- tensors_to_array(predict(vetiver_model$model, new_data))
    tibble::tibble(preds)
  }
}

tensors_to_array <- function(x) {
  if (is.list(x)) {
    lapply(x, tensors_to_array)
  } else if (inherits(x, "torch_tensor")) {
    as.array(x$cpu())
  } else {
    x
  }
}

#' @rdname vetiver_python_requirements
#' @export
vetiver_renviron_requirements.luz_module_fitted <- function(model) {
  system.file("requirements/luz-renviron.txt", package = "vetiver")
}

#' Model handler functions for API endpoint
#'
#' These are developer-facing functions, useful for supporting new model types.
#' Each model supported by `vetiver_model()` uses two handler functions
#' in [vetiver_api()]:
#' - The `handler_startup` function executes when the API starts. Use this
#' function for tasks like loading packages. A model can use the default
#' method here, which is `NULL` (to do nothing at startup).
#' - The `handler_predict` function executes at each API call. Use this
#' function for calling `predict()` and any other tasks that must be executed
#' at each API call.
#'
#' @details These are two generics that use the class of `vetiver_model$model`
#' for dispatch.
#'
#' @inheritParams vetiver_api
#'
#' @examples
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' handler_startup(v)
#' handler_predict(v)
#'
#' @return A `handler_startup` function should return invisibly, while a
#' `handler_predict` function should return a function with the signature
#' `function(req)`. The request body (`req$body`) consists of the new data
#' at prediction time; this function should return predictions either as a
#' tibble or as a list coercable to a tibble via [tibble::as_tibble()].
#' @rdname handler_startup
#' @export
handler_startup <- function(vetiver_model) {
  UseMethod("handler_startup", vetiver_model$model)
}

#' @rdname handler_startup
#' @export
handler_startup.default <- function(vetiver_model) invisible(NULL)

#' @rdname handler_startup
#' @export
handler_predict <- function(vetiver_model, ...) {
  UseMethod("handler_predict", vetiver_model$model)
}

#' @rdname handler_startup
#' @export
handler_predict.default <- function(vetiver_model, ...) {
  abort("There is no method available to build a prediction handler for `x`.")
}

#' Convert new data at prediction time using input data prototype
#'
#' This is a developer-facing function, useful for supporting new model types.
#' At prediction time, new observations typically must be checked and sometimes
#' converted to the data types from training time.
#'
#' @examples
#'
#' library(tibble)
#' training_df <- tibble(x = as.Date("2021-01-01") + 0:9,
#'                       y = LETTERS[1:10], z = letters[11:20])
#' training_df
#'
#' prototype <- vctrs::vec_slice(training_df, 0)
#' vetiver_type_convert(tibble(x = "2021-02-01", y = "J", z = "k"), prototype)
#'
#' ## unsuccessful conversion generates an error:
#' try(vetiver_type_convert(tibble(x = "potato", y = "J", z = "k"), prototype))
#'
#' ## error for missing column:
#' try(vetiver_type_convert(tibble(x = "potato", y = "J"), prototype))
#'
#' @inheritParams predict.vetiver_endpoint
#' @param ptype An input data prototype, such as a 0-row slice of the training
#' data
#' @return A converted dataframe
#' @export
vetiver_type_convert <- function(new_data, ptype) {
  if (is.null(ptype)) {
    return(new_data)
  }
  new_data <- hardhat::validate_column_names(new_data, colnames(ptype))
  spec <- readr::as.col_spec(ptype)
  is_character <- vapply(new_data, is.character, logical(1))
  if (any(is_character)) {
    new_data <- type_convert_strict(new_data, col_types = spec)
  }
  new_data
}

type_convert_strict <- function(
  new_data,
  col_types,
  call = rlang::caller_env()
) {
  warn_to_error <- function(e) {
    abort(conditionMessage(e), call = call)
  }

  tryCatch(
    warning = function(e) warn_to_error(e),
    expr = readr::type_convert(new_data, col_types = col_types)
  )
}

test_that("all numeric", {
  expect_equal(
    vetiver_type_convert(mtcars, vctrs::vec_slice(mtcars, 0)),
    tibble::tibble(mtcars)
  )
})

test_that("NULL ptype, for save_ptype = FALSE", {
  expect_equal(
    vetiver_type_convert(chickwts, NULL),
    chickwts
  )
})

test_that("missing variables", {
  expect_snapshot_error(
    vetiver_type_convert(mtcars[, 2:3], vctrs::vec_slice(mtcars, 0)),
  )
}) ## extra variables are caught by hardhat::scream()

test_that("a factor", {
  chicks <- chickwts
  chicks$feed <- as.character(chicks$feed)

  expect_equal(
    vetiver_type_convert(chicks, vctrs::vec_slice(chickwts, 0)),
    tibble::tibble(chickwts)
  )
})

test_that("a factor plus a bad character", {
  teeth <- ToothGrowth
  teeth$supp <- as.character(teeth$supp)
  teeth$dose <- as.character(teeth$dose)

  expect_equal(
    vetiver_type_convert(teeth, vctrs::vec_slice(ToothGrowth, 0)),
    tibble::tibble(ToothGrowth)
  )

  expect_snapshot_error(
    vetiver_type_convert(
      tibble::tibble(len = 4.2, supp = "ZZ", dose = 0.1),
      vctrs::vec_slice(ToothGrowth, 0)
    )
  )
})

test_that("a date", {
  many_dates <- tibble::tibble(
    x = as.Date("2021-01-01") + 0:10,
    y = as.Date("1980-02-14") + 0:10,
    z = as.POSIXct("1950-06-01", tz = "UTC") + 0:10
  )

  new_data <- tibble::tibble(
    x = "2021-01-15",
    y = "1980-03-01",
    z = "1950-06-01 00:00:15"
  )

  bad_data <- tibble::tibble(
    x = "potato",
    y = "1980-03-01",
    z = "1950-06-01 00:00:15"
  )

  expect_equal(
    vetiver_type_convert(new_data, vctrs::vec_slice(many_dates, 0)),
    tibble::tibble(
      x = as.Date("2021-01-15"),
      y = as.Date("1980-03-01"),
      z = as.POSIXct("1950-06-01 00:00:15", tz = "UTC")
    )
  )

  expect_snapshot_error(
    vetiver_type_convert(bad_data, vctrs::vec_slice(many_dates, 0)),
  )
})

#
# renv 1.1.5.9000 [rstudio/renv#4c63ee7]: A dependency management toolkit for R.
# Generated using `renv:::vendor()` at 2025-12-12 13:03:25.497218.
#

renv <- new.env(parent = new.env())

renv$initialize <- function(libname, pkgname) {
  # set up renv + imports environments
  attr(renv, "name") <- "embedded:renv"
  attr(parent.env(renv), "name") <- "imports:renv"

  # get imports
  imports <- list(
    tools = c(
      "file_ext",
      "md5sum",
      "package_dependencies",
      "pskill",
      "psnice",
      "write_PACKAGES"
    ),
    utils = c(
      "Rprof",
      "URLencode",
      "adist",
      "available.packages",
      "browseURL",
      "citation",
      "contrib.url",
      "download.file",
      "download.packages",
      "file.edit",
      "getCRANmirrors",
      "head",
      "help",
      "install.packages",
      "installed.packages",
      "modifyList",
      "old.packages",
      "packageDescription",
      "packageVersion",
      "read.table",
      "remove.packages",
      "sessionInfo",
      "str",
      "summaryRprof",
      "tail",
      "tar",
      "toBibtex",
      "untar",
      "unzip",
      "update.packages",
      "zip"
    )
  )

  # load the imports required by renv
  for (package in names(imports)) {
    namespace <- asNamespace(package)
    functions <- imports[[package]]
    list2env(mget(functions, envir = namespace), envir = parent.env(renv))
  }

  # source renv into the aforementioned environment
  script <- system.file("vendor/renv.R", package = .packageName)
  sys.source(script, envir = renv)

  # initialize metadata
  renv$the$metadata <- list(
    embedded = TRUE,
    version = structure(
      "1.1.5.9000",
      sha = "4c63ee794a768911ea470950dafd20368c862155"
    )
  )

  # run our load / attach hooks so internal state is initialized
  renv$.onLoad(libname, pkgname)

  # remove our initialize method when we're done
  rm(list = "initialize", envir = renv)
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.kproto <- function(model) {
  glue("A k-prototypes clustering model ({length(model$size)} clusters)")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.kproto <- function(model, metadata) {
  vetiver_meta(metadata, required_pkgs = "clustMixType")
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.kproto <- function(model) {
  butcher::butcher(model)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.kproto <- function(model, ...) {
  prototype <- vctrs::vec_ptype(model$data)
  tibble::as_tibble(prototype)
}

#' @rdname handler_startup
#' @export
handler_predict.kproto <- function(vetiver_model, ...) {
  prototype <- vetiver_model$prototype

  function(req) {
    newdata <- req$body
    if (!is_null(prototype)) {
      newdata <- vetiver_type_convert(newdata, prototype)
      newdata <- hardhat::scream(newdata, prototype)
    }
    newdata <- as.data.frame(newdata)
    # clustMixType:::predict.kproto()
    ret <- predict(vetiver_model$model, newdata = newdata, ...)
    list(.pred = ret$cluster)
  }
}

#' Post new data to a deployed model API endpoint and return predictions
#'
#' @param object A model API endpoint object created with [vetiver_endpoint()].
#' @param new_data New data for making predictions, such as a data frame.
#' @param ... Extra arguments passed to [httr::POST()]
#'
#' @return A tibble of model predictions with as many rows as in `new_data`.
#' @seealso [augment.vetiver_endpoint()]
#' @export
#'
#' @examples
#'
#' if (FALSE) {
#' endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
#' predict(endpoint, mtcars[4:7, -1])
#' }
#'
#'
predict.vetiver_endpoint <- function(object, new_data, ...) {
  rlang::check_installed(c("jsonlite", "httr"))
  check_dots_used()
  data_json <- jsonlite::toJSON(new_data, na = "string")
  ret <- httr::POST(object$url, ..., body = data_json)
  resp <- httr::content(ret, "text", encoding = "UTF-8")
  resp <- jsonlite::fromJSON(resp)

  if (httr::status_code(ret) >= 300) {
    if (has_name(resp, "message")) {
      abort(glue("Failed to predict: {resp$message}"))
    } else {
      status <- httr::http_status(ret)
      abort(c("Failed to predict.", status$message))
    }
  }

  tibble::as_tibble(resp)
}

#' Post new data to a deployed model API endpoint and augment with predictions
#'
#' @param x A model API endpoint object created with [vetiver_endpoint()].
#' @inheritParams predict.vetiver_endpoint
#'
#' @return The `new_data` with added prediction column(s).
#' @seealso [predict.vetiver_endpoint()]
#' @export
#'
#' @examples
#'
#' if (FALSE) {
#' endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
#' augment(endpoint, mtcars[4:7, -1])
#' }
#'
augment.vetiver_endpoint <- function(x, new_data, ...) {
  preds <- predict(x, new_data = new_data, ...)
  vctrs::vec_cbind(tibble::as_tibble(new_data), preds)
}


#' Create a model API endpoint object for prediction
#'
#' This function creates a model API endpoint for prediction from a URL. No
#' HTTP calls are made until you actually
#' [`predict()`][predict.vetiver_endpoint()] with your endpoint.
#'
#' @param url An API endpoint URL
#' @return A new `vetiver_endpoint` object
#'
#' @examples
#' vetiver_endpoint("https://colorado.rstudio.com/rsc/seattle-housing/predict")
#'
#' @export
vetiver_endpoint <- function(url) {
  url <- as.character(url)
  url <- gsub("/$", "", url)
  new_vetiver_endpoint(url)
}

new_vetiver_endpoint <- function(url = character()) {
  stopifnot(is.character(url))
  structure(list(url = url), class = "vetiver_endpoint")
}

#' @export
format.vetiver_endpoint <- function(x, ...) {
  cli::cli_format_method({
    cli::cli_h3("A model API endpoint for prediction:")
    cli::cli_text("{x$url}")
  })
}

#' @export
print.vetiver_endpoint <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.train <- function(model) {
  glue("A {tolower(model$modelInfo$label)} {tolower(model$modelType)} model")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.train <- function(model, metadata) {
  reqs <- c("caret", model$modelInfo$library)
  reqs <- sort(unique(reqs))
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.train <- function(model, ...) {
  tibble::as_tibble(model$ptype)
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.train <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.train <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.train <- function(vetiver_model, ...) {
  function(req) {
    newdata <- req$body
    newdata <- vetiver_type_convert(newdata, vetiver_model$prototype)
    predict(vetiver_model$model, newdata = newdata, ...)
  }
}

options(pins.verbose = FALSE)
options(pins.quiet = TRUE)
options(renv.verbose = FALSE)
Sys.setenv(RENV_CONFIG_SNAPSHOT_VALIDATE = FALSE)

clean_python_tmp_dir <- function() {
  if (!rlang::is_installed("reticulate")) {
    return()
  }

  if (!reticulate::py_available()) {
    return()
  }

  tryCatch(
    error = function(cnd) {
      cli::cli_inform("Cannot clean Python temp directory: {cnd}")
    },
    {
      python_temp_dir <- dirname(
        reticulate::py_run_string(
          "import tempfile; x=tempfile.NamedTemporaryFile().name",
          local = TRUE
        )$x
      )
      detritus <- fs::dir_ls(
        python_temp_dir,
        regexp = "__autograph_generated_file|__pycache__"
      )
      fs::file_delete(detritus)
    }
  )
}

withr::defer(clean_python_tmp_dir(), teardown_env())

# Generated by the vetiver package; edit with care

library(pins)
library(plumber)
library(rapidoc)
library(vetiver)

# Packages needed to generate model predictions
if (FALSE) {
    library(dials)
    library(dplyr)
    library(LiblineaR)
    library(parsnip)
    library(purrr)
    library(recipes)
    library(rlang)
    library(rsample)
    library(textrecipes)
    library(themis)
    library(tibble)
    library(tidyr)
    library(tune)
    library(vctrs)
    library(workflows)
    library(yardstick)
}
b <- board_rsconnect("envvar", server = "https://colorado.rstudio.com/rsc")
v <- vetiver_pin_read(b, "julia.silge/netflix_descriptions", version = "53336")

#* @plumber
function(pr) {
    pr |> vetiver_api(v, debug = TRUE)
}

#' Deploy a vetiver model API to Amazon SageMaker
#'
#' @description Use `vetiver_deploy_sagemaker()` to deploy a [vetiver_model()]
#' that has been versioned and stored via [vetiver_pin_write()] as a Plumber API
#' on [Amazon SageMaker](https://aws.amazon.com/sagemaker/).
#'
#' @inheritParams vetiver_sm_build
#' @inheritParams vetiver_sm_model
#' @inheritParams vetiver_sm_endpoint
#' @param ... Not currently used.
#' @param build_args A list of optional arguments passed to
#' [vetiver_sm_build()] such as the model `version` or the `compute_type`.
#' @param endpoint_args A list of optional arguments passed to
#' [vetiver_sm_endpoint()] such as `accelerator_type` or `data_capture_config`.
#' @param repo_name The name for the AWS ECR repository to store the model.
#'
#' @details
#' This function stores your model deployment image in the same bucket used
#' by `board`.
#'
#' The function `vetiver_deploy_sagemaker()` uses:
#' - [vetiver_sm_build()] to build and push a Docker image to ECR,
#' - [vetiver_sm_model()] to create a SageMaker model, and
#' - [vetiver_sm_endpoint()] to deploy a SageMaker model endpoint.
#'
#' These modular functions are available for more advanced use cases.
#'
#' If you are working locally, you will likely need to explicitly set up
#' your execution role to work correctly. Check out
#' ["Execution role requirements"](https://dyfanjones.r-universe.dev/smdocker)
#' in the smdocker documentation, and especially note that the bucket containing
#' your vetiver model needs to be added as a resource in your IAM role policy.
#'
#' @return
#' The deployed [vetiver_endpoint_sagemaker()].
#'
#' @seealso [vetiver_sm_build()], [vetiver_sm_model()], [vetiver_sm_endpoint()]
#' @export
#' @examples
#' if (FALSE) {
#' library(pins)
#' b <- board_s3(bucket = "my-existing-bucket")
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' endpoint <- vetiver_deploy_sagemaker(
#'     board = b,
#'     name = "cars_linear",
#'     instance_type = "ml.t2.medium",
#'     predict_args = list(type = "class", debug = TRUE)
#' )
#' }
#'
vetiver_deploy_sagemaker <- function(
  board,
  name,
  instance_type,
  ...,
  predict_args = list(),
  docker_args = list(),
  build_args = list(),
  endpoint_args = list(),
  repo_name = glue("vetiver-sagemaker-{name}")
) {
  check_dots_empty()
  if (!inherits(board, "pins_board_s3")) {
    stop_input_type(board, "an S3 pins board")
  }

  repo_name <- ifelse(
    grepl(":", repo_name),
    repo_name,
    glue("{repo_name}:{strftime(Sys.time(), '%Y-%m-%d')}")
  )

  # build image and push to aws ecr
  build_args <- compact(c(
    list(
      board = board,
      name = name,
      predict_args = predict_args,
      docker_args = docker_args,
      repository = repo_name,
      bucket = board$bucket
    ),
    build_args
  ))
  image_uri <- do.call(vetiver_sm_build, build_args)

  tags <- sm_check_tags(endpoint_args$tags)
  tags <- list_modify(
    tags,
    "vetiver:pin_board" = glue("s3://{board$bucket}/{board$prefix %||% ''}"),
    "vetiver:r-ver" = getRversion()
  )
  endpoint_args$tags <- NULL

  # create sagemaker model
  model_name <- vetiver_sm_model(image_uri, tags = tags)

  # create sagemaker endpoint
  endpoint_args <- compact(c(
    list(
      model_name = model_name,
      instance_type = instance_type,
      tags = tags
    ),
    endpoint_args
  ))
  endpoint <- do.call(vetiver_sm_endpoint, endpoint_args)
  return(endpoint)
}

#' Deploy a vetiver model API to Amazon SageMaker with modular functions
#'
#' @description
#' Use the function [vetiver_deploy_sagemaker()] for basic deployment on
#' [SageMaker](https://aws.amazon.com/sagemaker/), or these three functions
#' together for more advanced use cases:
#' - `vetiver_sm_build()` generates and builds a Docker image on SageMaker for
#' a vetiver model
#' - `vetiver_sm_model()` creates an Amazon SageMaker model
#' - `vetiver_sm_endpoint()` deploys an Amazon SageMaker model endpoint
#'
#' @inheritParams vetiver_prepare_docker
#' @param board An AWS S3 board created with [pins::board_s3()]. This board
#' must be in the correct region for your SageMaker instance.
#' @param repository The ECR repository and tag for the image as a character.
#' Defaults to `sagemaker-studio-${domain_id}:latest`.
#' @param compute_type The [CodeBuild](https://aws.amazon.com/codebuild/)
#' compute type as a character. Defaults to `BUILD_GENERAL1_SMALL`.
#' @param role The ARN IAM role name (as a character) to be used with:
#' - CodeBuild for `vetiver_sm_build()`
#' - the SageMaker model for `vetiver_sm_model()`
#'
#' Defaults to the SageMaker Studio execution role.
#' @param bucket The S3 bucket to use for sending data to CodeBuild as a
#' character. Defaults to the SageMaker SDK default bucket.
#' @param vpc_id ID of the VPC that will host the CodeBuild project such as
#' `"vpc-05c09f91d48831c8c"`.
#' @param subnet_ids List of subnet IDs for the CodeBuild project, such as
#' `list("subnet-0b31f1863e9d31a67")`.
#' @param security_group_ids List of security group IDs for the CodeBuild
#' project, such as `list("sg-0ce4ec0d0414d2ddc")`.
#' @param log A logical to show the logs of the running CodeBuild build.
#' Defaults to `TRUE`.
#' @param ... [Docker build parameters](https://docs.docker.com/engine/reference/commandline/build/#options>)
#' (Use "_" instead of "-"; for example, Docker optional parameter
#' `build-arg` becomes `build_arg`).
#'
#' @details The function `vetiver_sm_build()` generates the files necessary to
#' build a Docker container to deploy a vetiver model in SageMaker and then
#' builds the image on [AWS CodeBuild](https://aws.amazon.com/codebuild/). The
#' resulting image is stored in [AWS ECR](https://aws.amazon.com/ecr/).
#' This function creates a Plumber file and Dockerfile appropriate for
#' SageMaker, for example, with `path = "/invocations"` and `port = 8080`.
#'
#' If you run into problems with Docker rate limits, then either
#' - authenticate to Docker from SageMaker, or
#' - use a [public ECR base image](https://gallery.ecr.aws/docker/library/r-base),
#' passed through `docker_args`
#'
#' @seealso [vetiver_prepare_docker()], [vetiver_deploy_sagemaker()], [vetiver_endpoint_sagemaker()]
#' @examples
#' if (FALSE) {
#' library(pins)
#' b <- board_s3(bucket = "my-existing-bucket")
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' new_image_uri <- vetiver_sm_build(
#'     board = b,
#'     name = "cars_linear",
#'     predict_args = list(type = "class", debug = TRUE),
#'     docker_args = list(
#'         base_image = "FROM public.ecr.aws/docker/library/r-base:4.2.2"
#'     )
#' )
#'
#' model_name <- vetiver_sm_model(
#'     new_image_uri,
#'     tags = list("my_custom_tag" = "fuel_efficiency")
#' )
#'
#' vetiver_sm_endpoint(model_name, "ml.t2.medium")
#' }
#' @return `vetiver_sm_build()` returns the AWS ECR image URI and
#' `vetiver_sm_model()` returns the model name (both as characters).
#' `vetiver_sm_endpoint()` returns a new [vetiver_endpoint_sagemaker()] object.
#' @export
vetiver_sm_build <- function(
  board,
  name,
  version = NULL,
  path = fs::dir_create(tempdir(), "vetiver"),
  predict_args = list(),
  docker_args = list(),
  repository = NULL,
  compute_type = c(
    "BUILD_GENERAL1_SMALL",
    "BUILD_GENERAL1_MEDIUM",
    "BUILD_GENERAL1_LARGE",
    "BUILD_GENERAL1_2XLARGE"
  ),
  role = NULL,
  bucket = NULL,
  vpc_id = NULL,
  subnet_ids = list(),
  security_group_ids = list(),
  log = TRUE,
  ...
) {
  check_installed("smdocker")
  check_dots_used()
  compute_type <- arg_match(compute_type)
  docker_args <- list_modify(docker_args, port = 8080)
  predict_args <- list_modify(predict_args, path = "/invocations")

  vetiver_prepare_docker(
    board = board,
    name = name,
    version = version,
    path = path,
    predict_args = predict_args,
    docker_args = docker_args
  )

  image_uri <- smdocker::sm_build(
    repository = repository,
    compute_type = compute_type,
    role = role,
    dir = path,
    bucket = bucket,
    vpc_id = vpc_id,
    subnet_ids = subnet_ids,
    security_group_ids = security_group_ids,
    log = log,
    ...
  )

  return(image_uri)
}

#' @param image_uri The AWS ECR image URI for the Amazon SageMaker Model to be
#' created (for example, as returned by [vetiver_sm_build()]).
#' @param model_name The Amazon SageMaker model name to be deployed.
#' @param vpc_config A list containing the VPC configuration for the Amazon
#' SageMaker model [API VpcConfig](https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_VpcConfig.html)
#' (optional).
#' * `Subnets`: List of subnet ids
#' * `SecurityGroupIds`: List of security group ids
#' @param enable_network_isolation A logical to specify whether the container
#' will run in network isolation mode. Defaults to `FALSE`.
#' @param tags A named list of tags for labeling the Amazon SageMaker model or
#' model endpint to be created.
#' @rdname vetiver_sm_build
#' @export
vetiver_sm_model <- function(
  image_uri,
  model_name,
  role = NULL,
  vpc_config = list(),
  enable_network_isolation = FALSE,
  tags = list()
) {
  check_installed(c("smdocker", "paws.machine.learning"))
  config <- smdocker::smdocker_config()
  sagemaker_client <- paws.machine.learning::sagemaker(config)

  if (is_missing(model_name)) {
    model_name <- base_name_from_image(image_uri)
  }

  if (is.null(role)) {
    role <- smdocker::sagemaker_get_execution_role()
  }
  tags <- sm_check_tags(tags)
  tags <- sm_format_tags(tags)

  request <- list(
    "ModelName" = model_name,
    "ExecutionRoleArn" = role,
    "PrimaryContainer" = list("Image" = image_uri)
  )
  request$Tags <- .sm_append_project_tags(tags)
  request$VpcConfig <- sm_check_vpc_config(vpc_config)
  if (isTRUE(enable_network_isolation)) {
    request$EnableNetworkIsolation <- TRUE
  }
  # create model
  do.call(sagemaker_client$create_model, request)

  return(model_name)
}

#' @param instance_type Type of EC2 instance to use; see
#' [Amazon SageMaker pricing](https://aws.amazon.com/sagemaker/pricing/).
#' @param endpoint_name The name to use for the Amazon SageMaker model endpoint
#' to be created, if to be different from `model_name`.
#' @param initial_instance_count The initial number of instances to run
#' in the endpoint.
#' @param accelerator_type Type of Elastic Inference accelerator to
#' attach to an endpoint for model loading and inference, for
#' example, `"ml.eia1.medium"`.
#' @param kms_key The ARN of the KMS key used to encrypt the data on the
#' storage volume attached to the instance hosting the endpoint.
#' @param data_capture_config A list for configuration to control how Amazon
#' SageMaker captures inference data.
#' @param volume_size The size, in GB, of the ML storage volume attached to
#' the individual inference instance associated with the production variant.
#' Currently only Amazon EBS gp2 storage volumes are supported.
#' @param model_data_download_timeout The timeout value, in seconds, to download
#' and extract model data from Amazon S3.
#' @param wait A logical for whether to wait for the endpoint to be deployed.
#' Defaults to `TRUE`.
#' @rdname vetiver_sm_build
#' @export
vetiver_sm_endpoint <- function(
  model_name,
  instance_type,
  endpoint_name = NULL,
  initial_instance_count = 1,
  accelerator_type = NULL,
  tags = list(),
  kms_key = NULL,
  data_capture_config = list(),
  volume_size = NULL,
  model_data_download_timeout = NULL,
  wait = TRUE
) {
  check_installed(c("smdocker", "paws.machine.learning"))

  config <- smdocker::smdocker_config()
  sagemaker_client <- paws.machine.learning::sagemaker(config)

  endpoint_name <- endpoint_name %||% model_name

  tags <- sm_check_tags(tags)
  tags <- sm_format_tags(tags)

  request <- sm_req_endpoint_config(
    model_name,
    endpoint_name,
    instance_type,
    initial_instance_count,
    accelerator_type,
    volume_size,
    model_data_download_timeout,
    .sm_append_project_tags(tags),
    kms_key,
    data_capture_config
  )

  # create endpoint config
  resp <- do.call(
    sagemaker_client$create_endpoint_config,
    request
  )

  # create endpoint
  endpoint_name <- sm_create_endpoint(
    sagemaker_client,
    model_name,
    endpoint_name,
    tags,
    wait
  )

  return(vetiver_endpoint_sagemaker(model_name))
}

#' Delete Amazon SageMaker model, endpoint, and endpoint configuration
#'
#' Use this function to delete the Amazon SageMaker components used in a
#' [vetiver_endpoint_sagemaker()] object. This function does _not_ delete
#' any pinned model object in S3.
#'
#' @param object The model API endpoint object to be deleted, created with
#' [vetiver_endpoint_sagemaker()].
#' @param delete_model Delete the SageMaker model? Defaults to `TRUE`.
#' @param delete_endpoint Delete both the endpoint and endpoint configuration?
#' Defaults to `TRUE`.
#' @return `TRUE`, invisibly
#' @seealso [vetiver_deploy_sagemaker()], [vetiver_sm_build()], [vetiver_endpoint_sagemaker()]
#' @export
vetiver_sm_delete <- function(
  object,
  delete_model = TRUE,
  delete_endpoint = TRUE
) {
  check_installed(c("smdocker", "paws.machine.learning"))

  config <- smdocker::smdocker_config()
  sagemaker_client <- paws.machine.learning::sagemaker(config)

  endpoint_name <- object$model_endpoint

  if (is_true(delete_endpoint)) {
    tryCatch(
      {
        endpoint_describe <- sagemaker_client$describe_endpoint(
          EndpointName = endpoint_name
        )
        endpoint_config_name <- endpoint_describe$EndpointConfigName
        sagemaker_client$delete_endpoint_config(endpoint_config_name)
      },
      error = function(err) {
        cli::cli_warn(
          "Unable to delete {.val {endpoint_name}} endpoint configuration."
        )
      }
    )
    sagemaker_client$delete_endpoint(endpoint_name)
  }
  if (is_true(delete_model)) {
    sagemaker_client$delete_model(endpoint_name)
  }
  return(invisible(TRUE))
}

#' Post new data to a deployed SageMaker model endpoint and return predictions
#'
#' @param object A SageMaker model endpoint object created with [vetiver_endpoint_sagemaker()].
#' @param new_data New data for making predictions, such as a data frame.
#' @param ... Extra arguments passed to [paws.machine.learning::sagemakerruntime_invoke_endpoint()]
#'
#' @return A tibble of model predictions with as many rows as in `new_data`.
#' @seealso [augment.vetiver_endpoint_sagemaker()]
#' @export
#' @examples
#' if (FALSE) {
#'   endpoint <- vetiver_endpoint_sagemaker("sagemaker-demo-model")
#'   predict(endpoint, mtcars[4:7, -1])
#' }
#' @export
predict.vetiver_endpoint_sagemaker <- function(object, new_data, ...) {
  check_installed(c("jsonlite", "smdocker", "paws.machine.learning"))
  data_json <- jsonlite::toJSON(new_data, na = "string")
  config <- smdocker::smdocker_config()
  sm_runtime <- paws.machine.learning::sagemakerruntime(config)
  tryCatch(
    {
      resp <- sm_runtime$invoke_endpoint(object$model_endpoint, data_json, ...)
      resp <- resp$Body
    },
    error = function(error) {
      error_code <- error$error_response$ErrorCode
      if (!is.null(error_code) && error_code == "NO_SUCH_ENDPOINT") {
        cli::cli_abort(
          "Model endpoint {.val {object$model_endpoint}} not found."
        )
      }
      stop(error)
    }
  )
  con <- rawConnection(resp)
  on.exit(close(con))
  resp <- jsonlite::fromJSON(con)
  return(tibble::as_tibble(resp))
}

#' Post new data to a deployed SageMaker model endpoint and augment with predictions
#'
#' @param x A SageMaker model endpoint object created with [vetiver_endpoint_sagemaker()].
#' @inheritParams predict.vetiver_endpoint_sagemaker
#'
#' @return The `new_data` with added prediction column(s).
#' @seealso [predict.vetiver_endpoint_sagemaker()]
#' @export
#' @examples
#'
#' if (FALSE) {
#'   endpoint <- vetiver_endpoint_sagemaker("sagemaker-demo-model")
#'   augment(endpoint, mtcars[4:7, -1])
#' }
#'
augment.vetiver_endpoint_sagemaker <- function(x, new_data, ...) {
  preds <- predict(x, new_data = new_data, ...)
  vctrs::vec_cbind(tibble::as_tibble(new_data), preds)
}


#' Create a SageMaker model API endpoint object for prediction
#'
#' This function creates a model API endpoint for prediction from a Sagemaker Model.
#' No HTTP calls are made until you actually
#' [`predict()`][predict.vetiver_endpoint_sagemaker()] with your endpoint.
#'
#' @param model_endpoint The name of the Amazon SageMaker model endpoint.
#' @return A new `vetiver_endpoint_sagemaker` object
#'
#' @examplesIf rlang::is_installed("smdocker")
#' vetiver_endpoint_sagemaker("vetiver-sagemaker-demo-model")
#' @export
vetiver_endpoint_sagemaker <- function(model_endpoint) {
  check_installed("smdocker")
  config <- smdocker::smdocker_config()
  check_character(model_endpoint)
  check_character(config$region)
  new_vetiver_endpoint_sagemaker(model_endpoint, config$region)
}

new_vetiver_endpoint_sagemaker <- function(
  model_endpoint = character(),
  region = character()
) {
  structure(
    list(model_endpoint = model_endpoint, region = region),
    class = "vetiver_endpoint_sagemaker"
  )
}

#' @export
format.vetiver_endpoint_sagemaker <- function(x, ...) {
  cli::cli_format_method({
    cli::cli_h3("A SageMaker model endpoint for prediction:")
    cli::cli_text("Model endpoint: {x$model_endpoint}")
    cli::cli_text("Region: {x$region}")
  })
}

#' @export
print.vetiver_endpoint_sagemaker <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

skip_if_not_installed("smdocker")

test_that("can deploy via `vetiver_deploy_sagemaker()`", {
  skip_on_cran()
  model_name <- "example_model_name"
  instance_type <- "ml.t2.medium"
  mockery::stub(
    vetiver_deploy_sagemaker,
    "vetiver_sm_build",
    "new_sagemaker_uri"
  )
  mockery::stub(vetiver_deploy_sagemaker, "vetiver_sm_model", model_name)
  mockery::stub(
    vetiver_deploy_sagemaker,
    "vetiver_sm_endpoint",
    vetiver_endpoint_sagemaker(model_name)
  )
  local_mocked_bindings(
    version_name = function(metadata) "20130102T050607Z-xxxxx",
    .package = "pins"
  )

  b <- board_folder(path = tmp_dir)
  cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(cars_lm, "cars1")
  vetiver_pin_write(b, v)

  expect_snapshot(
    vetiver_deploy_sagemaker(b, "cars1", instance_type),
    error = TRUE
  )

  class(b) <- c("pins_board_s3", "pins_board")
  out <- vetiver_deploy_sagemaker(b, "cars1", instance_type)
  expect_equal(out, vetiver_endpoint_sagemaker(model_name))
})

test_that("can create correct files for `vetiver_sm_build()`", {
  skip_on_cran()
  mockery::stub(vetiver_sm_build, "smdocker::sm_build", "new_sagemaker_uri")
  local_mocked_bindings(
    version_name = function(metadata) "20130202T050607Z-xxxxx",
    .package = "pins"
  )

  b <- board_folder(path = tmp_dir)
  cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(cars_lm, "cars1")
  vetiver_pin_write(b, v)

  new_uri <- vetiver_sm_build(b, "cars1", path = tmp_dir)
  expect_equal(new_uri, "new_sagemaker_uri")

  expect_true(fs::file_exists(fs::path(tmp_dir, "vetiver_renv.lock")))
  expect_snapshot(
    cat(readr::read_lines(fs::path(tmp_dir, "plumber.R")), sep = "\n"),
    transform = redact_vetiver
  )
  expect_snapshot(
    cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n"),
    transform = redact_vetiver
  )
})


test_that("can create SageMaker Model", {
  mockery::stub(vetiver_sm_model, "sagemaker_client$create_model", list())
  image_uri <- "999999999999.dkr.ecr.us-east-2.amazonaws.com/vetiver-sagemaker-example-model:2023-03-17"
  model_name <- "vetiver-sagemaker-example_model"
  role <- "arn:aws:iam::999999999999:role/SagemakerExecution"

  expect_snapshot(error = TRUE, {
    vetiver_sm_model(image_uri = image_uri, role = role, tags = "potato")
    vetiver_sm_model(image_uri = image_uri, role = role, tags = list("potato"))
    vetiver_sm_model(
      image_uri = image_uri,
      role = role,
      vpc_config = list("potato")
    )
    vetiver_sm_model(
      image_uri = image_uri,
      role = role,
      vpc_config = list(Subnets = list(1:3), SecurityGroupIds = 1:3)
    )
    vetiver_sm_model(
      image_uri = image_uri,
      role = role,
      vpc_config = list(Subnets = 1:3, SecurityGroupIds = list(1:3))
    )
  })

  out <-
    vetiver_sm_model(
      image_uri = image_uri,
      model_name = model_name,
      role = role,
      tags = list("new-tag" = "amazing")
    )

  expect_equal(out, model_name)
})

test_that("can create SageMaker Endpoint", {
  mockery::stub(
    vetiver_sm_endpoint,
    "sagemaker_client$create_endpoint_config",
    list()
  )
  mockery::stub(
    vetiver_sm_endpoint,
    "sm_create_endpoint",
    "example-endpoint-name"
  )
  model_name <- "vetiver-sagemaker-example_model"
  instance_type <- "ml.t2.medium"

  expect_snapshot(error = TRUE, {
    vetiver_sm_endpoint(
      model_name = model_name,
      instance_type = instance_type,
      tags = "potato"
    )
    vetiver_sm_endpoint(
      model_name = model_name,
      instance_type = instance_type,
      tags = list("potato")
    )
  })

  out <-
    vetiver_sm_endpoint(
      model_name = model_name,
      instance_type = instance_type,
      tags = list("new-tag" = "amazing")
    )

  expect_equal(out, vetiver_endpoint_sagemaker(model_name))
})

test_that("can call sm_create_endpoint", {
  mockery::stub(sm_create_endpoint, "client$create_endpoint", list())

  endpoint_name <- "vetiver-sagemaker-example-model"
  config_name <- "vetiver-sagemaker-config"
  client <- paws.machine.learning::sagemaker(list())

  expect_snapshot(
    out <- sm_create_endpoint(client, endpoint_name, config_name, wait = FALSE)
  )

  expect_equal(out, endpoint_name)
})

test_that("can delete SageMaker endpoints", {
  mock_delete_endpoint_config <- mockery::mock(abort("No config!"), list())
  mock_delete_model <- mockery::mock(list(), cycle = TRUE)
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$describe_endpoint",
    list(EndpointConfigName = "the-name")
  )
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_endpoint_config",
    mock_delete_endpoint_config
  )
  mockery::stub(vetiver_sm_delete, "sagemaker_client$delete_endpoint", list())
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_model",
    mock_delete_model
  )

  object <- vetiver_endpoint_sagemaker("vetiver-sagemaker-example-model")

  expect_snapshot(vetiver_sm_delete(object))
  expect_true(vetiver_sm_delete(object))
  # Check if delete model and endpoint have been called
  mockery::expect_called(mock_delete_endpoint_config, 2)
  mockery::expect_called(mock_delete_model, 2)
})

test_that("can delete SageMaker model but not endpoint", {
  mock_delete_endpoint_config <- mockery::mock(list(
    EndpointConfigName = "dummy-endpoint"
  ))
  mock_delete_model <- mockery::mock(list())
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$describe_endpoint",
    list(EndpointConfigName = "the-name")
  )
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_endpoint_config",
    mock_delete_endpoint_config
  )
  mockery::stub(vetiver_sm_delete, "sagemaker_client$delete_endpoint", list())
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_model",
    mock_delete_model
  )

  object <- vetiver_endpoint_sagemaker("vetiver-sagemaker-example-model")

  expect_true(vetiver_sm_delete(object, delete_endpoint = FALSE))
  # Check if delete model and endpoint have been called
  mockery::expect_called(mock_delete_endpoint_config, 0)
  mockery::expect_called(mock_delete_model, 1)
})

test_that("can delete SageMaker endpoints but not model", {
  mock_delete_endpoint_config <- mockery::mock(list(
    EndpointConfigName = "dummy-endpoint"
  ))
  mock_delete_model <- mockery::mock(list())
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$describe_endpoint",
    list(EndpointConfigName = "the-name")
  )
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_endpoint_config",
    mock_delete_endpoint_config
  )
  mockery::stub(vetiver_sm_delete, "sagemaker_client$delete_endpoint", list())
  mockery::stub(
    vetiver_sm_delete,
    "sagemaker_client$delete_model",
    mock_delete_model
  )

  object <- vetiver_endpoint_sagemaker("vetiver-sagemaker-example-model")

  expect_true(vetiver_sm_delete(object, delete_model = FALSE))
  # Check if delete model and endpoint have been called
  mockery::expect_called(mock_delete_endpoint_config, 1)
  mockery::expect_called(mock_delete_model, 0)
})

test_that("can create vetiver endpoint object", {
  mockery::stub(
    vetiver_endpoint_sagemaker,
    "smdocker::smdocker_config",
    list(region = "my-region")
  )

  model_endpoint <- "vetiver-sagemaker-example-model"

  expect_snapshot(
    vetiver_endpoint_sagemaker(list()),
    error = TRUE
  )
  expect_snapshot(vetiver_endpoint_sagemaker(model_endpoint))
})

test_that("can predict with vetiver endpoint object", {
  res <- tibble::tibble(.pred = mtcars$mpg)
  mockery::stub(
    vetiver_endpoint_sagemaker,
    "smdocker::smdocker_config",
    list(region = "my-region")
  )
  mockery::stub(
    predict.vetiver_endpoint_sagemaker,
    "sm_runtime$invoke_endpoint",
    list(Body = charToRaw(jsonlite::toJSON(res)))
  )
  mockery::stub(augment.vetiver_endpoint_sagemaker, "predict", res)

  endpoint <- vetiver_endpoint_sagemaker("example-model")
  expect_equal(predict(endpoint, mtcars), res)
  expect_equal(
    augment(endpoint, mtcars),
    vctrs::vec_cbind(tibble::tibble(mtcars), res)
  )
})

test_that("can get base name from image", {
  image_uri <- "999999999999.dkr.ecr.us-east-2.amazonaws.com/vetiver-sagemaker-example-model:2023-03-17"

  expect_true(grepl(
    "vetiver-sagemaker-example-model-[0-9]{4,}-[0-9]{2,}-[0-9]{2,}-[0-9]{2,}-[0-9]{2,}-[0-9]{2,}-[0-9]{3,}",
    base_name_from_image(image_uri)
  ))
})

#' @keywords internal
"_PACKAGE"


#' @import rlang
#' @importFrom purrr map map_lgl map_chr
#' @importFrom purrr transpose compact pluck
#' @importFrom purrr pmap safely list_modify
#' @importFrom utils head modifyList flush.console
#' @importFrom stats predict
#' @importFrom vctrs vec_slice vec_sort
#' @importFrom glue glue
#' @importFrom glue glue_collapse
#' @importFrom generics required_pkgs
#' @importFrom lifecycle deprecated
NULL

#' @importFrom generics augment
#' @export
generics::augment

#' @importFrom generics required_pkgs
#' @export
generics::required_pkgs

globalVariables(c(
  "pr",
  ".metric",
  ".pred",
  "price",
  "tidy",
  "term",
  "estimate",
  "terms"
))

## to avoid NOTE about "All declared Imports should be used."
rapidoc_function_for_note <- function() {
  rapidoc::rapidoc_spec()
}

release_bullets <- function() {
  c(
    'Update renv with `renv:::vendor()`'
  )
}

#' Read and write a trained model to a board of models
#'
#' Use `vetiver_pin_write()` to pin a trained model to a board of models,
#' along with an input prototype for new data and other model metadata. Use
#' `vetiver_pin_read()` to retrieve that pinned object.
#'
#' @inheritParams pins::pin_read
#' @inheritParams vetiver_api
#' @param check_renv Use [renv](https://rstudio.github.io/renv/) to record the
#' packages used at training time with `vetiver_pin_write()` and check for
#' differences with `vetiver_pin_read()`. Defaults to `FALSE`.
#'
#' @details These functions read and write a [vetiver_model()] pin on the
#' specified `board` containing the model object itself and other elements
#' needed for prediction, such as the model's input data prototype or which
#' packages are needed at prediction time. You may use [pins::pin_read()] or
#' [pins::pin_meta()] to handle the pin, but `vetiver_pin_read()` returns a
#' [vetiver_model()] object ready for deployment.
#'
#' @return `vetiver_pin_read()` returns a [vetiver_model()]; `vetiver_pin_write()`
#' returns the name of the new pin, invisibly.
#'
#' @examples
#' library(pins)
#' model_board <- board_temp()
#'
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(model_board, v)
#' model_board
#'
#' vetiver_pin_read(model_board, "cars_linear")
#'
#' # can use `version` argument to read a specific version:
#' pin_versions(model_board, "cars_linear")
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#' # can store an renv lockfile as part of the pin:
#' vetiver_pin_write(model_board, v, check_renv = TRUE)
#'
#' @export
vetiver_pin_write <- function(board, vetiver_model, ..., check_renv = FALSE) {
  renv_lock <- NULL

  if (check_renv) {
    pkgs <- c(vetiver_model$metadata$required_pkgs, "vetiver")
    renv_lock <-
      renv$snapshot(
        lockfile = NULL,
        packages = pkgs,
        prompt = FALSE,
        force = TRUE
      )
  }

  metadata <- list_modify(
    vetiver_model$metadata$user,
    required_pkgs = vetiver_model$metadata$required_pkgs,
    renv_lock = renv_lock
  )

  pins::pin_write(
    board = board,
    x = list(model = vetiver_model$model, prototype = vetiver_model$prototype),
    name = vetiver_model$model_name,
    type = "rds",
    description = vetiver_model$description,
    metadata = metadata,
    versioned = vetiver_model$versioned,
    ...,
    force_identical_write = TRUE
  )

  rlang::inform(
    c(
      "\nCreate a Model Card for your published model",
      "Model Cards provide a framework for transparent, responsible reporting",
      "Use the vetiver `.Rmd` template as a place to start"
    ),
    class = "model_card_nudge",
    .frequency = "once",
    .frequency_id = "model_card_nudge"
  )
}

#' @rdname vetiver_pin_write
#' @export
vetiver_pin_read <- function(board, name, version = NULL, check_renv = FALSE) {
  pinned <- pins::pin_read(board = board, name = name, version = version)
  meta <- pins::pin_meta(board = board, name = name, version = version)
  required_pkgs <- meta$user$required_pkgs %||% pinned$required_pkgs

  if (check_renv) {
    if (length(meta$user$renv_lock) > 0) {
      local_lockfile <-
        renv$snapshot(
          lockfile = NULL,
          packages = c(required_pkgs, "vetiver"),
          prompt = FALSE,
          force = TRUE
        )
      orig_lockfile <- structure(meta$user$renv_lock, class = "renv_lockfile")
      renv_report_actions(local_lockfile, orig_lockfile)
    } else {
      cli::cli_warn(c(
        "There is no lockfile stored with {.val {name}}:",
        "i" = "Use {.arg check_renv = TRUE} when you save your model to your board"
      ))
    }
  }

  meta$user <- list_modify(meta$user, required_pkgs = zap(), renv_lock = zap())
  if (is_empty(meta$user)) {
    names(meta$user) <- NULL
  }

  new_vetiver_model(
    model = pinned$model,
    model_name = name,
    description = meta$description,
    metadata = vetiver_meta(
      user = meta$user,
      version = meta$local$version,
      url = meta$local$url,
      required_pkgs = required_pkgs
    ),
    prototype = pinned$prototype %||% pinned$ptype,
    versioned = board$versioned
  )
}


renv_report_actions <- function(current, model) {
  withr::local_options(
    list(renv.pretty.print.emitter = function(text, ...) {
      cli::cli_inform(text)
    })
  )
  diff <- renv$renv_lockfile_diff_packages(current, model)

  if (renv$empty(diff)) {
    return(invisible(NULL))
  }

  lhs <- renv$renv_records(current)
  rhs <- renv$renv_records(model)
  renv$renv_pretty_print_records_pair(
    "The following package(s) do not match your model:",
    lhs[names(lhs) %in% names(diff)],
    rhs[names(rhs) %in% names(diff)],
    "Consider installing the same versions that your model was trained with."
  )
}

skip_if_not_installed("mockery")

test_that("can pin a model", {
  b <- board_temp()
  v <- vetiver_model(cars_lm, "cars1")
  expect_snapshot(v)
  vetiver_pin_write(b, v)
  expect_equal(
    pin_read(b, "cars1"),
    list(
      model = butcher::butcher(cars_lm),
      prototype = vctrs::vec_slice(tibble::as_tibble(mtcars[, 2:3]), 0)
    )
  )
})

test_that("can pin a model with no prototype", {
  b <- board_temp()
  expect_snapshot_warning(
    v <- vetiver_model(cars_lm, "cars_null", save_ptype = FALSE)
  )
  v <- vetiver_model(cars_lm, "cars_null", save_prototype = FALSE)
  vetiver_pin_write(b, v)
  expect_equal(
    pin_read(b, "cars_null"),
    list(
      model = butcher::butcher(cars_lm),
      prototype = NULL
    )
  )
})

test_that("can pin a model with custom prototype", {
  b <- board_temp()
  v <- vetiver_model(cars_lm, "cars_custom", save_prototype = mtcars[3:10, 2:3])
  vetiver_pin_write(b, v)
  expect_equal(
    pin_read(b, "cars_custom"),
    list(
      model = butcher::butcher(cars_lm),
      prototype = mtcars[3:10, 2:3]
    )
  )
})

test_that("default metadata for model", {
  b <- board_temp()
  v <- vetiver_model(cars_lm, "cars2")
  vetiver_pin_write(b, v)
  meta <- pin_meta(b, "cars2")
  expect_equal(meta$user, list(required_pkgs = NULL, renv_lock = NULL))
  expect_equal(meta$description, "An OLS linear regression model")
})

test_that("user can supply metadata for model", {
  b <- board_temp()
  v <- vetiver_model(
    cars_lm,
    "cars3",
    description = "lm model for mtcars",
    metadata = list(metrics = 1:10)
  )
  vetiver_pin_write(b, v)
  meta <- pin_meta(b, "cars3")
  expect_equal(meta$user$metrics, 1:10)
  expect_equal(meta$description, "lm model for mtcars")
})

test_that("can read a pinned model", {
  b <- board_temp()
  cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(cars_lm, "cars1")
  vetiver_pin_write(b, v)
  v1 <- vetiver_pin_read(b, "cars1")
  meta <- pin_meta(b, "cars1")
  expect_equal(v1$model, v$model)
  expect_equal(v1$model_name, v$model_name)
  expect_equal(v1$board, v$board)
  expect_equal(v1$description, v$description)
  expect_equal(
    v1$metadata,
    list(
      user = v$metadata$user,
      version = meta$local$version,
      url = meta$local$url,
      required_pkgs = v$metadata$required_pkgs
    )
  )
  expect_equal(v1$prototype, v$prototype)
  expect_equal(v1$versioned, FALSE)
})

test_that("can read an old pinned model with `required_pkgs` in blob", {
  b <- board_temp()
  pins::pin_write(
    board = b,
    x = list(
      model = v$model,
      ptype = v$ptype,
      required_pkgs = c("janeaustenr", "beepr")
    ),
    name = "cars-plus-pkgs",
    type = "rds",
    description = v$description,
    metadata = v$metadata$user,
    versioned = v$versioned
  )
  v2 <- vetiver_pin_read(b, "cars-plus-pkgs")
  expect_equal(v2$metadata$required_pkgs, c("janeaustenr", "beepr"))
})

test_that("can read a versioned model with metadata", {
  b <- board_temp(versioned = TRUE)
  cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)
  v <- vetiver_model(
    cars_lm,
    "cars4",
    description = "lm model for mtcars",
    metadata = list(metrics = 1:10)
  )
  vetiver_pin_write(b, v)
  v4 <- vetiver_pin_read(b, "cars4")
  meta <- pin_meta(b, "cars4")
  expect_equal(v4$model, v$model)
  expect_equal(v4$model_name, v$model_name)
  expect_equal(v4$board, v$board)
  expect_equal(v4$description, v$description)
  expect_equal(
    v4$metadata,
    list(
      user = v$metadata$user,
      version = meta$local$version,
      url = meta$local$url,
      required_pkgs = v$metadata$required_pkgs
    )
  )
  expect_equal(v4$prototype, v$prototype)
  expect_equal(v4$versioned, TRUE)
})

test_that("right message for reading with `check_renv`", {
  skip_on_cran()

  b <- board_temp(versioned = TRUE)
  mock_version_name <- mockery::mock(
    "20130104T050607Z-xxxxx",
    "20130204T050607Z-yyyyy",
    "20130304T050607Z-zzzzz",
  )
  local_mocked_bindings(version_name = mock_version_name, .package = "pins")
  v <- vetiver_model(cars_lm, "cars5")
  v$metadata$required_pkgs <- "ranger"
  vetiver_pin_write(b, v)
  expect_snapshot_warning(vetiver_pin_read(b, "cars5", check_renv = TRUE))

  vetiver_pin_write(b, v, check_renv = TRUE)
  v1 <- vetiver_pin_read(b, "cars5")
  expect_silent(v2 <- vetiver_pin_read(b, "cars5", check_renv = TRUE))
  expect_equal(v1, v2)

  new_lock <- renv$renv_lockfile_init(project = NULL)
  pins::pin_write(
    board = b,
    x = list(
      model = v$model,
      ptype = v$ptype,
      required_pkgs = v$metadata$required_pkgs
    ),
    name = v$model_name,
    type = "rds",
    description = v$description,
    metadata = list_modify(v$metadata$user, renv_lockfile = new_lock),
    versioned = v$versioned
  )

  expect_message(
    vetiver_pin_read(b, "cars5", check_renv = TRUE),
    regexp = "do not match your model"
  )
})

#' Aggregate model metrics over time for monitoring
#'
#' These three functions can be used for model monitoring (such as in a
#' monitoring dashboard):
#' - `vetiver_compute_metrics()` computes metrics (such as accuracy for a
#' classification model or RMSE for a regression model) at a chosen time
#' aggregation `period`
#' - [vetiver_pin_metrics()] updates an existing pin storing model metrics
#' over time
#' - [vetiver_plot_metrics()] creates a plot of metrics over time
#'
#' @inheritParams yardstick::metrics
#' @inheritParams slider::slide_period
#' @param date_var The column in `data` containing dates or date-times for
#' monitoring, to be aggregated with `.period`
#' @param metric_set A [yardstick::metric_set()] function for computing metrics.
#' Defaults to [yardstick::metrics()].
#'
#' @return A dataframe of metrics.
#'
#' @details
#' For arguments used more than once in your monitoring dashboard,
#' such as `date_var`, consider using
#' [R Markdown parameters](https://pkg.yihui.org/rmarkdown-book/parameterized-reports)
#' to reduce repetition and/or errors.
#'
#' @examplesIf rlang::is_installed(c("dplyr", "parsnip", "modeldata"))
#' library(dplyr)
#' library(parsnip)
#' data(Chicago, package = "modeldata")
#' Chicago <- Chicago |> select(ridership, date, all_of(stations))
#' training_data <- Chicago |> filter(date < "2009-01-01")
#' testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
#' monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
#' lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)
#'
#' library(pins)
#' b <- board_temp()
#'
#' original_metrics <-
#'     augment(lm_fit, new_data = testing_data) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#'
#' new_metrics <-
#'     augment(lm_fit, new_data = monitoring) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#'
#' @seealso [vetiver_pin_metrics()], [vetiver_plot_metrics()]
#' @export
vetiver_compute_metrics <- function(
  data,
  date_var,
  period,
  truth,
  estimate,
  ...,
  metric_set = yardstick::metrics,
  every = 1L,
  origin = NULL,
  before = 0L,
  after = 0L,
  complete = FALSE
) {
  rlang::check_installed("slider")
  truth_quo <- enquo(truth)
  estimate_quo <- enquo(estimate)

  # Figure out which column in `data` corresponds to `date_var`
  date_var <- enquo(date_var)
  date_var <- eval_select_one(date_var, data, "date_var")

  index <- data[[date_var]]

  slider::slide_period_dfr(
    .x = data,
    .i = index,
    .period = period,
    .f = compute_metrics,
    date_var = date_var,
    metric_set = metric_set,
    truth_quo = truth_quo,
    estimate_quo = estimate_quo,
    ...,
    .every = every,
    .origin = origin,
    .before = before,
    .after = after,
    .complete = complete
  )
}


#' Update model metrics over time for monitoring
#'
#' These three functions can be used for model monitoring (such as in a
#' monitoring dashboard):
#' - [vetiver_compute_metrics()] computes metrics (such as accuracy for a
#' classification model or RMSE for a regression model) at a chosen time
#' aggregation `period`
#' - `vetiver_pin_metrics()` updates an existing pin storing model metrics
#' over time
#' - [vetiver_plot_metrics()] creates a plot of metrics over time
#'
#' @inheritParams pins::pin_read
#' @param df_metrics A tidy dataframe of metrics over time, such as created by
#' [vetiver_compute_metrics()].
#' @param metrics_pin_name Pin name for where the *metrics* are stored (as
#' opposed to where the model object is stored with [vetiver_pin_write()]).
#' @param .index The variable in `df_metrics` containing the aggregated dates
#' or date-times (from `time_var` in `data`). Defaults to `.index`.
#' @param overwrite If `FALSE` (the default), error when the new metrics contain
#' overlapping dates with the existing pin.If `TRUE`, overwrite any metrics for
#' dates that exist both in the existing pin and new metrics with the _new_
#' values.
#' @param type File type used to save metrics to disk. With the default `NULL`,
#' uses the `type` of the existing pin. Options are "rds" and "arrow".
#'
#' @return A dataframe of metrics.
#'
#' @details Sometimes when you monitor a model at a given time aggregation, you
#' may end up with dates in your new metrics (like `new_metrics` in the example)
#' that are the same as dates in your existing aggregated metrics (like
#' `original_metrics` in the example). This can happen if you need to re-run a
#' monitoring report because something failed. With `overwrite = FALSE` (the
#' default), `vetiver_pin_metrics()` will error when there are overlapping
#' dates. With `overwrite = TRUE`, `vetiver_pin_metrics()` will replace such
#' metrics with the new values. You probably want `FALSE` for interactive use
#' and `TRUE` for dashboards or reports that run on a schedule.
#'
#' You can initially create your pin with `type = "arrow"` or the default
#' (`type = "rds"`). `vetiver_pin_metrics()` will update the pin using the
#' same `type` by default.
#'
#' @examplesIf rlang::is_installed(c("dplyr", "parsnip", "modeldata", "arrow")) && identical(Sys.getenv("NOT_CRAN"), "true")
#' library(dplyr)
#' library(parsnip)
#' data(Chicago, package = "modeldata")
#' Chicago <- Chicago |> select(ridership, date, all_of(stations))
#' training_data <- Chicago |> filter(date < "2009-01-01")
#' testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
#' monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
#' lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)
#'
#' library(pins)
#' b <- board_temp()
#'
#' ## before starting monitoring, initiate the metrics and pin
#' ## (for example, with the testing data):
#' original_metrics <-
#'     augment(lm_fit, new_data = testing_data) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#' pin_write(b, original_metrics, "lm_fit_metrics", type = "arrow")
#'
#' ## to continue monitoring with new data, compute metrics and update pin:
#' new_metrics <-
#'     augment(lm_fit, new_data = monitoring) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#' vetiver_pin_metrics(b, new_metrics, "lm_fit_metrics")
#'
#' @seealso [vetiver_compute_metrics()], [vetiver_plot_metrics()]
#' @export
vetiver_pin_metrics <- function(
  board,
  df_metrics,
  metrics_pin_name,
  .index = .index,
  overwrite = FALSE,
  type = NULL,
  ...
) {
  pin_type <- type %||% pins::pin_meta(board, metrics_pin_name)$type

  if (!pin_type %in% c("arrow", "rds")) {
    cli::cli_abort(
      'Metrics must be pinned as "arrow" or "rds", not "{pin_type}"'
    )
  }

  .index <- enquo(.index)
  .index <- eval_select_one(.index, df_metrics, "date_var")
  new_dates <- unique(df_metrics[[.index]])

  old_metrics <- pins::pin_read(board, metrics_pin_name)
  overlapping_dates <- old_metrics[[.index]] %in% new_dates
  if (overwrite) {
    old_metrics <- vec_slice(old_metrics, !overlapping_dates)
  } else {
    if (any(overlapping_dates)) {
      cli::cli_abort(c(
        "The new metrics overlap with dates \\
                 already stored in {glue::single_quote(metrics_pin_name)}",
        i = "Check the aggregated dates or use `overwrite = TRUE`"
      ))
    }
  }
  new_metrics <- vctrs::vec_rbind(old_metrics, df_metrics)
  new_metrics <- vec_slice(
    new_metrics,
    vctrs::vec_order(new_metrics[[.index]])
  )

  pins::pin_write(
    board,
    new_metrics,
    basename(metrics_pin_name),
    type = pin_type,
    ...
  )
  new_metrics
}

compute_metrics <- function(
  data,
  date_var,
  metric_set,
  truth_quo,
  estimate_quo,
  ...
) {
  index <- data[[date_var]]
  index <- min(index)

  n <- nrow(data)

  metrics <- metric_set(
    data = data,
    truth = !!truth_quo,
    estimate = !!estimate_quo,
    ...
  )

  tibble::tibble(
    .index = index,
    .n = n,
    metrics
  )
}

eval_select_one <- function(col, data, arg, ..., call = caller_env()) {
  rlang::check_installed("tidyselect")
  check_dots_empty()

  # `col` is a quosure that has its own environment attached
  env <- empty_env()

  loc <- tidyselect::eval_select(
    expr = col,
    data = data,
    env = env,
    error_call = call
  )

  if (length(loc) != 1L) {
    message <- glue::glue(
      "`{arg}` must specify exactly one column from `data`."
    )
    abort(message, call = call)
  }

  loc
}


#' Plot model metrics over time for monitoring
#'
#' These three functions can be used for model monitoring (such as in a
#' monitoring dashboard):
#' - [vetiver_compute_metrics()] computes metrics (such as accuracy for a
#' classification model or RMSE for a regression model) at a chosen time
#' aggregation `period`
#' - [vetiver_pin_metrics()] updates an existing pin storing model metrics
#' over time
#' - `vetiver_plot_metrics()` creates a plot of metrics over time
#'
#' @inheritParams vetiver_pin_metrics
#' @param .estimate The variable in `df_metrics` containing the metric estimate.
#' Defaults to `.estimate`.
#' @param .metric The variable in `df_metrics` containing the metric type.
#' Defaults to `.metric`.
#' @param .n The variable in `df_metrics` containing the number of observations
#' used for estimating the metric.
#'
#' @return A `ggplot2` object.
#'
#' @examplesIf interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")
#' library(dplyr)
#' library(parsnip)
#' data(Chicago, package = "modeldata")
#' Chicago <- Chicago |> select(ridership, date, all_of(stations))
#' training_data <- Chicago |> filter(date < "2009-01-01")
#' testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
#' monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
#' lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)
#'
#' library(pins)
#' b <- board_temp()
#'
#' ## before starting monitoring, initiate the metrics and pin
#' ## (for example, with the testing data):
#' original_metrics <-
#'     augment(lm_fit, new_data = testing_data) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#' pin_write(b, original_metrics, "lm_fit_metrics", type = "arrow")
#'
#' ## to continue monitoring with new data, compute metrics and update pin:
#' new_metrics <-
#'     augment(lm_fit, new_data = monitoring) |>
#'     vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
#' vetiver_pin_metrics(b, new_metrics, "lm_fit_metrics")
#'
#' library(ggplot2)
#' vetiver_plot_metrics(new_metrics) +
#'     scale_size(range = c(2, 4))
#'
#' @seealso [vetiver_compute_metrics()], [vetiver_pin_metrics()]
#' @export
vetiver_plot_metrics <- function(
  df_metrics,
  .index = .index,
  .estimate = .estimate,
  .metric = .metric,
  .n = .n
) {
  rlang::check_installed("ggplot2")
  .metric <- enquo(.metric)

  ggplot2::ggplot(
    data = df_metrics,
    ggplot2::aes({{ .index }}, {{ .estimate }})
  ) +
    ggplot2::geom_line(ggplot2::aes(color = !!.metric), alpha = 0.7) +
    ggplot2::geom_point(
      ggplot2::aes(color = !!.metric, size = {{ .n }}),
      alpha = 0.9
    ) +
    ggplot2::facet_wrap(ggplot2::vars(!!.metric), scales = "free_y", ncol = 1) +
    ggplot2::guides(color = "none") +
    ggplot2::labs(x = NULL, y = NULL, size = NULL)
}

skip_if_not_installed("modeldata")
skip_if_not_installed("dplyr")
skip_if_not_installed("parsnip")
skip_if_not_installed("slider")
skip_if_not_installed("yardstick")

describe("vetiver_compute_metrics()", {
  data(Chicago, package = "modeldata")
  Chicago <- dplyr::select(Chicago, ridership, date, one_of(stations))
  training_data <- dplyr::filter(Chicago, date < "2009-01-01")
  testing_data <- dplyr::filter(
    Chicago,
    date >= "2009-01-01",
    date < "2011-01-01"
  )
  lm_fit <- parsnip::fit(
    parsnip::linear_reg(),
    ridership ~ .,
    data = training_data
  )
  lm_aug <- parsnip::augment(lm_fit, new_data = testing_data)

  it("can compute default metrics", {
    res <- vetiver_compute_metrics(
      lm_aug,
      date,
      "month",
      ridership,
      .pred
    )
    expect_s3_class(res, "tbl_df")
    expect_equal(unique(res$.metric), c("rmse", "rsq", "mae"))
    expect_equal(ncol(res), 5)
    expect_equal(nrow(res), 72)
  })
  it("can compute custom metrics", {
    res <- vetiver_compute_metrics(
      lm_aug,
      date,
      "month",
      ridership,
      .pred,
      metric_set = yardstick::metric_set(yardstick::rmse, yardstick::mape)
    )
    expect_s3_class(res, "tbl_df")
    expect_equal(unique(res$.metric), c("rmse", "mape"))
    expect_equal(ncol(res), 5)
    expect_equal(nrow(res), 48)
  })
  it("can compute rolling metrics", {
    res <- vetiver_compute_metrics(
      lm_aug,
      date,
      "week",
      ridership,
      .pred,
      every = 6L
    )
    expect_s3_class(res, "tbl_df")
    expect_equal(unique(res$.metric), c("rmse", "rsq", "mae"))
    expect_equal(ncol(res), 5)
    expect_equal(nrow(res), 54)
  })
})

describe("vetiver_pin_metrics()", {
  skip_if_not_installed("vdiffr")
  data(Chicago, package = "modeldata")
  Chicago <- dplyr::select(Chicago, ridership, date, one_of(stations))
  training_data <- dplyr::filter(Chicago, date < "2009-01-01")
  testing_data <- dplyr::filter(
    Chicago,
    date >= "2009-01-01",
    date < "2011-01-01"
  )
  lm_fit <- parsnip::fit(
    parsnip::linear_reg(),
    ridership ~ .,
    data = training_data
  )

  df_metrics <-
    parsnip::augment(lm_fit, new_data = testing_data) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)

  it("fails without existing pin", {
    b <- pins::board_temp()
    expect_snapshot_error(
      vetiver_pin_metrics(b, df_metrics, "metrics1", overwrite = TRUE)
    )
  })
  it("fails with `overwrite = FALSE`", {
    b <- pins::board_temp()
    pins::pin_write(b, df_metrics, "metrics2")
    expect_snapshot_error(
      vetiver_pin_metrics(b, df_metrics, "metrics2")
    )
  })
  it("fails with type = csv", {
    b <- pins::board_temp()
    pins::pin_write(b, df_metrics, "metrics2", type = "csv")
    expect_snapshot_error(
      vetiver_pin_metrics(b, df_metrics, "metrics2")
    )

    pins::pin_write(b, df_metrics, "metrics2", type = "rds")
    expect_snapshot_error(
      vetiver_pin_metrics(b, df_metrics, "metrics2", type = "csv")
    )
  })
  it("can update metrics", {
    b <- pins::board_temp()
    pins::pin_write(b, df_metrics, "metrics3")

    new_metrics <- tibble::tibble(
      .index = as.Date("2011-01-12"),
      n = 30,
      .metric = c("rmse", "rsq", "mae"),
      .estimator = "standard",
      .estimate = c(3.0, 0.7, 2.0)
    )
    res2 <- vetiver_pin_metrics(b, new_metrics, "metrics3", overwrite = TRUE)
    expect_equal(
      pins::pin_read(b, "metrics3"),
      dplyr::arrange(vctrs::vec_rbind(df_metrics, new_metrics), .index)
    )
  })
})


describe("vetiver_plot_metrics()", {
  data(Chicago, package = "modeldata")
  Chicago <- dplyr::select(Chicago, ridership, date, one_of(stations))
  training_data <- dplyr::filter(Chicago, date < "2009-01-01")
  testing_data <- dplyr::filter(
    Chicago,
    date >= "2009-01-01",
    date < "2011-01-01"
  )
  lm_fit <- parsnip::fit(
    parsnip::linear_reg(),
    ridership ~ .,
    data = training_data
  )

  df_metrics <-
    parsnip::augment(lm_fit, new_data = testing_data) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)

  it("can plot monitoring metrics", {
    p <- vetiver_plot_metrics(df_metrics)
    expect_s3_class(p, "ggplot")
    vdiffr::expect_doppelganger("default metrics plot", p)
  })
})

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.model_stack <- function(model) {
  num_members <-
    tidy(model$coefs) |>
    dplyr::filter(estimate > 0, term != "(Intercept)") |>
    nrow()
  glue("A {model$mode} stacked ensemble with {num_members} members")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.model_stack <- function(model, metadata) {
  reqs <- map(model$member_fits, required_pkgs)
  reqs <- purrr::flatten_chr(reqs)
  reqs <- sort(unique(c(
    reqs,
    required_pkgs(model$coefs),
    "workflows",
    "stacks"
  )))
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.model_stack <- function(model, ...) {
  mold <- workflows::extract_mold(model$member_fits[[1]])
  mold$blueprint$ptypes$predictors
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.model_stack <- function(model) {
  ret <- butcher::butcher(model)
  ret <- bundle::bundle(ret)
  ret
}

#' @rdname handler_startup
#' @export
handler_startup.model_stack <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.model_stack <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    predict(vetiver_model$model, new_data = new_data, ...)
  }
}

#' @rdname vetiver_create_description
#' @export
vetiver_create_description.Learner <- function(model) {
  glue("A mlr3 {model$id} learner")
}

#' @rdname vetiver_create_meta
#' @export
vetiver_create_meta.Learner <- function(model, metadata) {
  reqs <- model$packages
  reqs <- sort(unique(c(reqs, "mlr3")))
  vetiver_meta(metadata, required_pkgs = reqs)
}

#' @rdname vetiver_create_ptype
#' @export
vetiver_ptype.Learner <- function(model, ...) {
  tibble::as_tibble(model$state$data_prototype)[,
    model$state$train_task$feature_names
  ]
}

#' @rdname vetiver_create_description
#' @export
vetiver_prepare_model.Learner <- function(model) {
  if (is.null(model$state)) {
    rlang::abort("Your `model` object is not a trained learner.")
  }
  model
}

#' @rdname handler_startup
#' @export
handler_startup.Learner <- function(vetiver_model) {
  attach_pkgs(vetiver_model$metadata$required_pkgs)
}

#' @rdname handler_startup
#' @export
handler_predict.Learner <- function(vetiver_model, ...) {
  function(req) {
    new_data <- req$body
    new_data <- vetiver_type_convert(new_data, vetiver_model$prototype)
    pred <- vetiver_model$model$predict_newdata(newdata = new_data)
    stats::setNames(
      list(pred$response),
      vetiver_model$model$state$train_task$target_names
    )
  }
}

#' Deploy a vetiver model API to Posit Connect
#'
#' Use `vetiver_deploy_rsconnect()` to deploy a [vetiver_model()] that has been
#' versioned and stored via [vetiver_pin_write()] as a Plumber API on [Posit
#' Connect](https://docs.posit.co/connect/).
#'
#' @inheritParams vetiver_write_plumber
#' @param predict_args A list of optional arguments passed to [vetiver_api()]
#' such as the prediction `type`.
#' @param appTitle The API title on Posit Connect. Use the default based on
#' `name`, or pass in your own title.
#' @param ... Other arguments passed to [rsconnect::deployApp()] such as
#' `appName`, `account`, or `launch.browser`.
#'
#' @details The two functions `vetiver_deploy_rsconnect()` and
#' [vetiver_create_rsconnect_bundle()] are alternatives to each other, providing
#' different strategies for deploying a vetiver model API to Posit Connect.
#'
#' When you first deploy to Connect, your API will only be accessible to you.
#' You can [change the access settings](https://docs.posit.co/connect/user/content-settings/#set-viewers)
#' so others can also access the API. For all access settings other than
#' "Anyone - no login required", anyone querying your API (including you)
#' will need to pass authentication details with your API call,
#' [as shown in the Connect documentation](https://docs.posit.co/connect/user/vetiver/#predict-from-your-model-endpoint).
#'
#' @return
#' The deployment success (`TRUE` or `FALSE`), invisibly.
#'
#' @seealso [vetiver_write_plumber()], [vetiver_create_rsconnect_bundle()]
#' @export
#'
#' @examplesIf rlang::is_installed("rsconnect")
#' library(pins)
#' b <- board_temp(versioned = TRUE)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' if (FALSE) {
#' ## pass args for predicting:
#' vetiver_deploy_rsconnect(
#'     b,
#'     "user.name/cars_linear",
#'     predict_args = list(debug = TRUE)
#' )
#'
#' ## specify an account name through `...`:
#' vetiver_deploy_rsconnect(
#'     b,
#'     "user.name/cars_linear",
#'     account = "user.name"
#' )
#' }
#'
vetiver_deploy_rsconnect <- function(
  board,
  name,
  version = NULL,
  predict_args = list(),
  appTitle = glue::glue("{name} model API"),
  ...,
  additional_pkgs = character(0)
) {
  check_dots_used()
  tmp <- fs::dir_create(tempdir(), "vetiver")
  vetiver_write_plumber(
    board = board,
    name = name,
    version = version,
    !!!predict_args,
    file = fs::path(tmp, "plumber.R"),
    additional_pkgs = additional_pkgs
  )
  rsconnect::deployAPI(tmp, appTitle = appTitle, ...)
}

#' Create an Posit Connect bundle for a vetiver model API
#'
#' Use `vetiver_create_rsconnect_bundle()` to create a
#' [Posit Connect](https://docs.posit.co/connect/) model API bundle for a
#' [vetiver_model()] that has been versioned and stored via
#' [vetiver_pin_write()].
#'
#' @inheritParams vetiver_write_plumber
#' @inheritParams vetiver_deploy_rsconnect
#' @param filename The path for the model API bundle to be created (can be
#' used as the argument to `connectapi::bundle_path()`)
#'
#' @return The location of the model API bundle `filename`, invisibly.
#' @seealso [vetiver_write_plumber()], [vetiver_deploy_rsconnect()]
#' @details This function creates a deployable bundle. See
#' [Posit Connect docs](https://docs.posit.co/connect/cookbook/deploying/)
#' for how to deploy this bundle, as well as the
#' [connectapi](https://pkgs.rstudio.com/connectapi/) R package for how to
#' integrate with Connect's API from R.
#'
#' The two functions `vetiver_create_rsconnect_bundle()` and
#' [vetiver_deploy_rsconnect()] are alternatives to each other, providing
#' different strategies for deploying a vetiver model API to Posit Connect.
#'
#' @examplesIf rlang::is_installed("connectapi") && identical(Sys.getenv("NOT_CRAN"), "true")
#' library(pins)
#' b <- board_temp(versioned = TRUE)
#' cars_lm <- lm(mpg ~ ., data = mtcars)
#' v <- vetiver_model(cars_lm, "cars_linear")
#' vetiver_pin_write(b, v)
#'
#' ## when you pin to Posit Connect, your pin name will be typically be like:
#' ## "user.name/cars_linear"
#' vetiver_create_rsconnect_bundle(
#'     b,
#'     "cars_linear",
#'     predict_args = list(debug = TRUE)
#' )
#'
#' @export
vetiver_create_rsconnect_bundle <- function(
  board,
  name,
  version = NULL,
  predict_args = list(),
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz"),
  additional_pkgs = character(0)
) {
  tmp <- fs::dir_create(tempdir(), "vetiver")
  vetiver_write_plumber(
    board = board,
    name = name,
    version = version,
    !!!predict_args,
    file = fs::path(tmp, "plumber.R"),
    additional_pkgs = additional_pkgs
  )
  rsconnect::writeManifest(tmp, "plumber.R")
  withr::with_dir(
    tmp,
    utils::tar(
      tarfile = filename,
      files = c("plumber.R", "manifest.json"),
      compression = "gzip",
      tar = "internal"
    )
  )
  inform(c("Your rsconnect bundle has been created at:", filename))
  invisible(filename)
}

mock_write_manifest <- function(appDir, appFiles) {
  fs::file_create(fs::path(appDir, "manifest.json"))
}

---
title: "Version, share, and deploy a model with vetiver"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{vetiver}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = requireNamespace("parsnip", quietly = TRUE) && requireNamespace("recipes", quietly = TRUE) && requireNamespace("workflows", quietly = TRUE)
)
```

The goal of vetiver is to provide fluent tooling for MLOps tasks for your trained model including:

- versioning
- storing
- sharing
- deploying

For more extensive documentation, visit <https://vetiver.posit.co/>.

## Create a `vetiver_model()`

The vetiver package is extensible, with generics that can support many kinds of models. For this example, let's consider one kind of model supported by vetiver, a [tidymodels](https://www.tidymodels.org/) workflow that encompasses both feature engineering and model estimation.

```{r}
library(parsnip)
library(recipes)
library(workflows)
data(bivariate, package = "modeldata")
bivariate_train

biv_rec <-
  recipe(Class ~ ., data = bivariate_train) |>
  step_BoxCox(all_predictors())|>
  step_normalize(all_predictors())

svm_spec <-
  svm_linear(mode = "classification") |>
  set_engine("LiblineaR")

svm_fit <- 
  workflow(biv_rec, svm_spec) |>
  fit(sample_frac(bivariate_train, 0.7))
```

This `svm_fit` object is a fitted workflow, with both feature engineering and model parameters estimated using the training data `bivariate_train`. We can create a `vetiver_model()` from this trained model; a `vetiver_model()` collects the information needed to store, version, and deploy a trained model.

```{r}
library(vetiver)
v <- vetiver_model(svm_fit, "biv_svm")
v
```

Think of this `vetiver_model()` as a deployable model object.

## Store and version your model

You can store and version your model by choosing a [pins](https://pins.rstudio.com) "board" for it, including a local folder, Posit Connect, Amazon S3, and more. Most pins boards have versioning turned on by default, but we can turn it on explicitly for our temporary demo board. When we write the `vetiver_model()` to our board, the binary model object is stored on our board together with necessary metadata, like the packages needed to make a prediction and the model's input data prototype for checking new data at prediction time.

```{r}
#| message: false
library(pins)
model_board <- board_temp(versioned = TRUE)
model_board |> vetiver_pin_write(v)
```

Let's train our model again with a new version of the dataset and write it once more to our board.

```{r}
#| message: false
svm_fit <- 
  workflow(biv_rec, svm_spec) |>
  fit(sample_frac(bivariate_train, 0.7))

v <- vetiver_model(svm_fit, "biv_svm")

model_board |> vetiver_pin_write(v)
```

Both versions are stored, and we have access to both.

```{r}
model_board |> pin_versions("biv_svm")
```

The primary purpose of pins is to make it easy to share data artifacts, so depending on the board you choose, your pinned `vetiver_model()` can be shareable with your collaborators.

## Deploy your model

You can deploy your model by creating a [Plumber](https://www.rplumber.io/) router, and adding a POST endpoint for making predictions.

```{r}
library(plumber)
pr() |>
  vetiver_api(v)
```

To start a server using this object, pipe (`|>`) to `pr_run(port = 8088)` or your port of choice. This allows you to interact with your vetiver API locally and debug it. Plumber APIs such as these can be [hosted in a variety of ways](https://www.rplumber.io/articles/hosting.html). You can use the function `vetiver_write_plumber()` to create a ready-to-go `plumber.R` file that is especially suited for [Posit Connect](https://posit.co/products/enterprise/connect/).

```{r}
#| eval: false
vetiver_write_plumber(model_board, "biv_svm")
```


```{r}
#| echo: false
#| comment: ''
tmp <- tempfile()
vetiver_write_plumber(model_board, "biv_svm", file = tmp)
cat(readr::read_lines(tmp), sep = "\n")
```

In a real-world situation, you would see something like `b <- board_connect()` or `b <- board_s3()` here instead of our temporary demo board. Notice that the deployment is strongly linked to a *specific version* of the pinned model; if you pin another version of the model after you deploy your model, your deployed model will not be affected.

## Predict from your model endpoint

A model deployed via vetiver can be treated as a special `vetiver_endpoint()` object.

```{r}
library(vetiver)
endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
endpoint
```

If such a deployed model endpoint is running via one R process (either remotely on a server or locally, perhaps via a [background job in the RStudio IDE](https://docs.posit.co/ide/user/ide/guide/tools/jobs.html)), you can make predictions with that deployed model and new data in another, separate R process. 

``` r
data(bivariate, package = "modeldata")
predict(endpoint, bivariate_test)
#> # A tibble: 710 × 1
#>    .pred_class
#>    <chr>      
#>  1 One        
#>  2 Two        
#>  3 One        
#>  4 Two        
#>  5 Two        
#>  6 One        
#>  7 Two        
#>  8 Two        
#>  9 Two        
#> 10 One        
#> # … with 700 more rows
```

Being able to `predict()` on a vetiver model endpoint takes advantage of the model's input data prototype and other metadata that is stored with the model.

---
title: "Model Card: your model title"
date: '`r Sys.Date()`'
output: 
  html_document
params:
    board: !r pins::board_connect()
    name: julia.silge/sacramento-rf
    version: NULL
---

```{r setup, include=FALSE}
library(tidyverse)
library(vetiver)
library(pins)
library(yardstick)
knitr::opts_chunk$set(echo = FALSE)
v <- vetiver_pin_read(params$board, params$name, version = params$version)
v_meta <- pin_meta(params$board, params$name)
theme_set(theme_light())
```

A [model card](https://doi.org/10.1145/3287560.3287596) provides brief, transparent, responsible reporting for a trained machine learning model.

## Model details

- Developed by PERSON AND/OR TEAM
- `r cli::pluralize("{v$description} using {ncol(v$prototype)} feature{?s}")`
- More details about how model was developed and what it is predicting
- More details on feature engineering and/or data preprocessing for model
- Version `r v$metadata$version` of this model was published at `r v_meta$created`
- Citation and/or license details for the model
- If you have questions about this model, please contact PERSON@ORG.ORG

## Intended use

- The primary intended uses of this model are ...
- The primary intended users of this model are ...
- Some use cases are out of scope for this model, such as ...

## Important aspects/factors

- Aspects or factors (demographic, environmental, technical) that are relevant to the context of this model are ...
- In evaluating this model, we examined aspects such as ...

## Metrics

- The metrics used to evaluate this model are ...
- These metrics are computed via ...
- We chose these metrics because ...

## Training data & evaluation data

- The training dataset for this model was ...
- The training dataset for this model has the "prototype" or signature:

    ```{r}
    glimpse(v$prototype)
    ```

- The evaluation dataset used in this model card is ...
- We chose this evaluation data because ...

```{r}
## EVALUATION DATA:

data(Sacramento, package = "modeldata")

## consider using a package like skimr or DataExplorer for automated 
## presentation of evaluation data characteristics
```


## Quantitative analyses {.tabset}

```{r}
## compute predictions for your evaluation data
## load packages needed for prediction:
library(parsnip)
library(workflows)
preds <- augment(v, Sacramento)
```


### Overall model performance

```{r}
preds |>
    metrics(price, .pred)
```

### Disaggregated model performance

```{r}
preds |>
    group_by(type) |>
    metrics(price, .pred)
```

### Visualize model performance

```{r, fig.height=3}
preds |>
    ggplot(aes(price, .pred, color = type)) +
    geom_abline(slope = 1, lty = 2, color = "gray60", size = 1.2) +
    geom_point(alpha = 0.5, show.legend = FALSE) +
    facet_wrap(vars(type))
```

### Make a custom plot

```{r}
preds |>
    mutate(.resid = price - .pred) |>
    ggplot(aes(longitude, latitude, color = .resid)) +
    geom_point(alpha = 0.8) +
    scale_color_gradient2() +
    coord_fixed()
```


## Ethical considerations

- We considered ...

## Caveats & recommendations

- This model does ...
- This model does not ...
- We recommend ...

# Custom Handlers

There are two different ways that vetiver supports flexible handling for models that do not work automatically with the vetiver framework. The first way is with new model types where there is no current implementation for the type of model you would like to deploy. The second way is when you would like to implement a current handler, but in a different way. In either case, you should create a custom handler from vetiver's `BaseHandler()`. At a minimum, you must give the type of your model via `model_type` how predictions should be made, via the method `handler_predict()`. Then, initialize your handler with your model, and pass the object into `VetiverModel`.

This example shows a custom handler of `newmodeltype` type.

```python
from vetiver.handlers.base import BaseHandler

class CustomHandler(BaseHandler):
    def __init__(self, model, ptype_data):
        super().__init__(model, ptype_data)

    model_type = staticmethod(lambda: newmodeltype)
    pip_name = "scikit-learn" # pkg name on pip, used for tracking pkg versions

    def handler_predict(self, input_data, check_ptype: bool):
        """
        Generates method for /predict endpoint in VetiverAPI

        The `handler_predict` function executes at each API call. Use this
        function for calling `predict()` and any other tasks that must be executed at each API call.

        Parameters
        ----------
        input_data:
            Test data
        check_ptype: bool
            Whether the ptype should be enforced

        Returns
        -------
        prediction
            Prediction from model
        """
        # your code here
        prediction = model.fancy_new_predict(input_data)

        return prediction

new_model = CustomHandler(model, ptype_data)

VetiverModel(new_model, "custom_model")
```

If your model is a common type, please consider [submitting a pull request](https://github.com/rstudio/vetiver-python/pulls).

---
title: "Monitor model"
output: 
  vetiver::vetiver_dashboard:
    pins:
      board: !expr pins::board_local()
      name: 'seattle_rf'
      version: NULL
    storyboard: true
    theme: 
      version: 4
      bootswatch: cosmo
    display_pins: true
---

```{r setup, include = FALSE}
library(flexdashboard)
library(vetiver)
library(pins)
library(plotly)

knitr::opts_chunk$set(echo = FALSE)
pins <- get_vetiver_dashboard_pins()
metrics_pin_name <- paste(pins$name, "metrics", sep = "_")
```

```{r set-up-demo-pins, include = FALSE, eval = FALSE}
# Before knitting this example monitoring dashboard, execute:
pin_example_kc_housing_model()

# This function will set up demo model and metrics pins 
# Instead of real pin URLs, you will see only demo links
```

```{r load-vetiver-model, include = FALSE}
# Load deployed model from pin:
v <- vetiver_pin_read(pins$board, pins$name, version = pins$version)
meta <- pin_meta(pins$board, pins$name, version = pins$version)
days_old <- difftime(Sys.Date(), as.Date(meta$created), units = "days")

# Attaches packages needed for prediction:
attach_pkgs(v$metadata$required_pkgs)
```

```{r validation, include = FALSE}
# Load new validation data, for example from database or API:
validation_df <- mlr3data::kc_housing |>
    arrange(date) |>
    filter(date >= "2015-01-01") 

validation_aug <- augment(v, validation_df)

new_metrics <-
    validation_aug |>
    vetiver_compute_metrics(date, "week", price, .pred)
    
vetiver_pin_metrics(pins$board, new_metrics, metrics_pin_name, overwrite = TRUE)
```


### Model metrics

```{r}
p1 <- new_metrics |>
    ## you can operate on your metrics as needed:
    filter(.metric %in% c("rmse", "mae"), .n > 20) |>
    vetiver_plot_metrics() + 
    ## you can also operate on the ggplot:
    scale_size(range = c(2, 5))

p1 <- ggplotly(p1)
hide_legend(p1)
```

***

This model was published `r as.numeric(days_old)` days ago.

Plot model metrics over time to *monitor* your model.

### Explore validation data

```{r}
p2 <-
    validation_df |>
    ggplot(aes(price, after_stat(density), fill = waterfront)) +
    geom_histogram(alpha = 0.7, position = "identity")

ggplotly(p2)
```


***

Write your own code to make visualizations or tables with the new validation data, and/or the new predictions.

### API visual documentation

```{r echo=FALSE, out.width="100%"}
## use your own vetiver model API URL here:
knitr::include_url("https://colorado.rstudio.com/rsc/seattle-housing/", height = "600px")
```

***

Interact directly with your model via its visual documentation, and get `curl` examples.

---
title: "Model Card: your model title"
output:
  html_document
---
```{python}
#| echo: false
from IPython.display import display, Markdown
import datetime
display(Markdown(f"""
###### {datetime.date.today()}
"""))
```

```{python}
#| tags: [parameters]
#| include: false
import pins

raw = "https://colorado.rstudio.com/rsc/"
paths = {"v": "sacramento/"}
board = pins.board_urls(raw, paths, allow_pickle_read=True)
```

```{python}
#| include: false
import datetime
import pandas as pd
import plotly.express as px
from sklearn import metrics
import vetiver

v = vetiver.VetiverModel.from_pin(board, "v")
# your pin will most likely have a version already, but we will make one here
v_meta = board.pin_meta("v")
v_meta.version = '5678'
```


A [model card](https://doi.org/10.1145/3287560.3287596) provides brief, transparent, responsible reporting for a trained machine learning model.

## Model details

- Developed by PERSON AND/OR TEAM
```{python}
#| echo: false
model_desc = v.description
num_features = len(v.prototype.construct().dict())

display(Markdown(f"""
- A {model_desc} using {num_features} feature{'s'[:num_features^1]}.
"""))
```
- More details about how model was developed and what it is predicting
- More details on feature engineering and/or data preprocessing for model
```{python}
#| echo: false
version = v_meta.version
time_created =  datetime.datetime
    .strptime(v_meta.created, '%Y%m%dT%H%M%SZ')
    .strftime('%Y-%m-%d %H:%M:%S')

display(Markdown(f"""
- Version {version} was created at {time_created}
"""))
```

- Citation and/or license details for the model
- If you have questions about this model, please contact PERSON@ORG.ORG

## Intended use

- The primary intended uses of this model are ...
- The primary intended users of this model are ...
- Some use cases are out of scope for this model, such as ...

## Important aspects/factors

- Aspects or factors (demographic, environmental, technical) that are relevant to the context of this model are ...
- In evaluating this model, we examined aspects such as ...

## Metrics

- The metrics used to evaluate this model are ...
- These metrics are computed via ...
- We chose these metrics because ...

## Training data & evaluation data

- The training dataset for this model was ...
- The training dataset for this model has the "prototype" or signature:
```{python}
#| echo: false
v.prototype.construct().schema().get("properties")
```

- The evaluation dataset used in this model card is ...
- We chose this evaluation data because ...


## Quantitative analyses
:::{.panel-tabset}
```{python}
#| include: false
## consider using a package like Pandas Profiling for automated EDA

```
## Overall model performance

```{python}
#| echo: false
## compute predictions for your evaluation data
from vetiver.data import sacramento

sacramento["preds"] = v.model.predict(sacramento.drop(columns=["price"]))
metric_set = [metrics.mean_absolute_error, metrics.mean_squared_error, metrics.r2_score]

for metric in metric_set:
    print(metric.__qualname__, ": ", metric(y_pred=sacramento["preds"], y_true=sacramento["price"]))
```

## Disaggregated model performance

```{python}
#| echo: false

dmp = pd.DataFrame()
for metric in metric_set:
    dmp[metric.__qualname__] = sacramento.groupby("type")\
        .apply(lambda x: metric(y_pred=x["preds"], y_true=x["price"]))
dmp
```

## Visualize model performance

```{python}
performance = px.scatter(sacramento, x="price", y = "preds", facet_row="type", trendline="ols")
performance.update_yaxes(matches=None)
performance.show()
```

## Make a custom plot

```{python}
custom = px.violin(sacramento, x="beds", y = "price")
custom.show()
```

:::

## Ethical considerations

- We considered ...

## Caveats & recommendations

- This model does ...
- This model does not ...
- We recommend ...

# Creating and deploying custom code

In some cases, you may need to create and deploy custom code as part of your MLOps workflow using vetiver. This could be necessary when you need to:

-   deploy custom models in vetiver
-   deploy unsupported models in vetiver
-   include custom code in vetiver
-   deploy a vetiver model with a custom pipeline

You may also have custom code in a known framework, such as a column transformer for a scikit-learn model.

In these cases, extra steps will be required to successfully create and deploy a `VetiverModel` object.

# Making a custom model

Vetiver supports basic [scikit-learn](https://scikit-learn.org/), [torch](https://pytorch.org/), [statsmodels](https://www.statsmodels.org/stable/index.html), [xgboost](https://xgboost.readthedocs.io/en/stable/), and [spacy](https://spacy.io/) models. If you need to alter the usage of these models, or deploy a different type of model, you will likely need to create a new model handler.

To create a model handler, you should create a subclass of vetiver's `BaseHandler` class. This handler should include the following:

-   `model_type`: A static method that declares the type of your model.
-   `handler_predict()`: A method that defines how predictions should be made for your model. This method is used at the /predict endpoint in the VetiverAPI.

Here's an example of a handler for a model of `newmodeltype` type. Once you have defined your handler, you can initialize it with your model and pass it to the `VetiverModel` class.

``` python
from vetiver.handlers.base import BaseHandler

class CustomHandler(BaseHandler):
    def __init__(self, model, prototype_data):
        super().__init__(model, prototype_data)

    model_type = staticmethod(lambda: newmodeltype)
    pip_name = "scikit-learn" # package's installation name on pip

    def handler_predict(self, input_data, check_prototype: bool):
        """
        Your code for making predictions using the custom model

        Parameters
        ----------
        input_data:
            Data POSTed to API endpoint
        check_prototype: bool
            Whether the prototype should be enforced
        """
        prediction = model.fancy_new_predict(input_data)

        return prediction

new_model = CustomHandler(model, prototype_data)

VetiverModel(new_model, "custom_model")
```

If your model is a common type, please consider [submitting a pull request](https://github.com/rstudio/vetiver-python/pulls).

To deploy custom code, you need to include the necessary source code in your deployment files. If your model or other elements can be imported from a Python package, you can include the relevant packages in a `requirements.txt` file for deployment. However, if you have custom source code in local files, you will need to include those files in the deployment process.

# Deploying custom elements

If your `VetiverModel` includes custom source code, you need to include that code in your deployment files to build an API in another location. The example below shows a user-created `FeatureSelector`, which is part of a scikit-learn pipeline.

``` {.python filename="model.py"}
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.tree import DecisionTreeClassifier
from sklearn.pipeline import Pipeline

# create custom data preprocessing
class FeatureSelector(BaseEstimator, TransformerMixin):
    def __init__(self, columns):
        self.columns = columns

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X[self.columns]

# create model
model = Pipeline(steps=[
    ('feature_selector', FeatureSelector(features)),
    ('decision_tree', DecisionTreeClassifier())
])

# create deployable model object
from vetiver import VetiverModel, vetiver_pin_write

v = VetiverModel(model, "selected_decision_tree", protoype_data = X)

# pin model to some location, eg, Posit Connect
import pins

board = pins.board_connect(allow_pickle_read=True)
vetiver_pin_write(board, v)
```

::: panel-tabset
## Docker

To generate files needed to start a Docker container, you can use the command `vetiver.prepare_docker`.

``` python
vetiver.prepare_docker(board, "selected_decision_tree")
```

When you run this line, 3 files are generated: a Dockerfile, an `app.py` file, and a `vetiver_requirements.txt`. In the `app.py` file, you'll need to add an import statement that is formatted `from {name of file, excluding .py, that has custom element} import {name of custom element}`.

``` {.python filename="app.py"}
from vetiver import VetiverModel
import vetiver
import pins
from model import FeatureSelector # add this line to import your custom feature engineering


b = pins.board_connect(allow_pickle_read=True)
v = VetiverModel.from_pin(b, 'selected_decision_tree')

vetiver_api = vetiver.VetiverAPI(v)
api = vetiver_api.app
```

Add a line to your Dockerfile to copy your source file(s) into your Docker container. The format will be `COPY path/to/your/filename.py /vetiver/app/filename.py`, where the destination is always in the `/vetiver/app/` directory.

``` {.bash filename="Dockerfile"}
# # Generated by the vetiver package; edit with care
# start with python base image
FROM python:3.10

# create directory in container for vetiver files
WORKDIR /vetiver

# copy  and install requirements
COPY vetiver_requirements.txt /vetiver/requirements.txt

#
RUN pip install --no-cache-dir --upgrade -r /vetiver/requirements.txt

# copy app file
COPY app.py /vetiver/app/app.py

# ADD THIS LINE to copy model source code
COPY model.py /vetiver/app/model.py

# expose port
EXPOSE 8080

# run vetiver API
CMD ["uvicorn", "app.app:api", "--host", "0.0.0.0", "--port", "8080"]
```

## Posit Connect

To deploy custom code to Posit Connect, you'll first start with the command `vetiver.write_app`.

``` python
vetiver.write_app(board, 'selected_decision_tree')
```

This will generate an `app.py` file, where you'll need to add an import statement that is formatted `from {name of file, excluding .py, that has custom element} import {name of custom element}`.

``` {.python filename="=\"app.py\""}
from vetiver import VetiverModel
import vetiver
import pins
from model import FeatureSelector # add this line to import your custom feature engineering


b = pins.board_connect(allow_pickle_read=True)
v = VetiverModel.from_pin(b, 'selected_decision_tree')

vetiver_api = vetiver.VetiverAPI(v)
api = vetiver_api.app
```

After editing the `app.py` file, you can deploy it to Posit Connect using the `rsconnect` package. Use the `rsconnect.api.actions.deploy_python_fastapi()` function to deploy the API, specifying the Connect server URL, API key, directory containing the `app.py` and `model.py` files, and the entry point of the API.

``` python
from rsconnect.api.actions import deploy_python_fastapi
import rsconnect

url = "example.connect.com" # your Posit Connect server url
api_key = os.environ(CONNECT_API_KEY) # your Posit Connect API key

connect_server = rsconnect.api.RSConnectServer(
    url = url,
    api_key = api_key
)

rsconnect.actions.deploy_python_fastapi(
        connect_server = connect_server,
        directory = "./", # path to the directory containing the app.py and model.py files
        entry_point = "app:api" # the API is the app.py file, in a variable named api
    )
```
:::

## Common Pitfalls

When deploying custom code, the most common error is something similar to `AttributeError: Can't get attribute 'ExampleModel' on <module '__main__' (built-in)>`. There are a few possible causes for this error:

1.  The original `ExampleModel` may have been pinned from inside a Jupyter Notebook. Because pickling only saves a reference for how to read a class, not the source code, a custom model transformer pinned from a Jupyter Notebook cannot be imported and resolved later. To fix this, repin your model/transformer from inside a Python script.

2.  You may not be uploading the custom code to be used later. When deploying, you'll want to add the files containing your custom code to the `extra_files` argument so that it can be imported, eg, `vetiver.deploy_rsconnect(connect_server, board, model_name, extra_files=['custom_model.py', 'requirements.txt'])` .

Please note that the above steps are a general guide, and you may need to adapt them to your specific use case and deployment environment. If you have any questions, please consider [opening an issue](https://github.com/rstudio/vetiver-python/issues/new).

---
title: "Monitoring dashboard"
format:
  dashboard:
    orientation: columns
logo: https://github.com/rstudio/vetiver-python/blob/main/docs/figures/logo.png?raw=true
output: asis
jupyter: python3
---

```{python}
#| include: false
#| tags: [parameters]

# import model and metadata
import pins
from IPython.display import display, Markdown, IFrame
from datetime import datetime, timedelta
import pandas as pd
import plotly.express as px
from sklearn import metrics
from vetiver import VetiverModel, compute_metrics, plot_metrics
from sklearn.metrics import recall_score, accuracy_score

raw = "https://pub.palm.ptd.posit.it/public"
paths = {"chicago-model-python": "chicago-model-python/"}
board = pins.board_url(raw, paths, allow_pickle_read=True)
v = VetiverModel.from_pin(board, "chicago-model-python")
v_meta = board.pin_meta("chicago-model-python")
days_old = datetime.today() - datetime.strptime(v_meta.created, "%Y%m%dT%H%M%SZ")
```

```{python}
## the next few lines are an example model, here is a place to
## add any code you need to import new data and make predictions

# import new data to track performance over time
raw = "https://pub.palm.ptd.posit.it/public"
paths = {"new-data": "inspections-monitor/"}
board = pins.board_url(raw, paths, allow_pickle_read=True)
inspections_new = board.pin_read("new-data")

# make predictions
inspections_new["preds"] = v.model.predict(
    inspections_new.drop(columns=["results", "aka_name", "inspection_date"])
)

# map results
inspections_new["preds"] = inspections_new["preds"].map({"PASS": 0, "FAIL": 1})
inspections_new["results"] = inspections_new["results"].map({"PASS": 0, "FAIL": 1})
```

# Model info

## Column
### Row {height="33%"}
::: {.valuebox}
`{python} v.description`

`{python} v.model_name`
:::

::: {.valuebox}
Model age

`{python} days_old.days` days old
:::

### Row

Model details

- This model has the prototype:

```
`{python} v.prototype.model_json_schema()`
```

- The model was created by ...

# Model metrics

## Column
```{python}
import itables

td = timedelta(weeks = 4)
metric_set = [accuracy_score, recall_score]

metrics_df = compute_metrics(
    data = inspections_new,
    date_var = "inspection_date",
    period = td,
    metric_set = metric_set,
    truth = "results",
    estimate = "preds"
  )
itables.show(metrics_df)
```

```{python}
plot_metrics(metrics_df).show()
```

## Column {.sidebar}

This tab is used to see model performance over time. In this context, _performance_ is the statistical properties of the model, eg, accuracy and recall.

You can add custom information and metrics here.

# Explore validation data

```{python}
fig = px.histogram(inspections_new, x = "facility_type")
fig.show()
```

## Column {.sidebar}

Write your own code to make visualizations or tables with the new validation data, and/or the new predictions.


# API visual documentation

## Column

```{python}
# For your own model, you can use
# the following code with your URL
# to display your model API's visual documentation

from IPython.display import IFrame
url = "your-model-api-url"
IFrame(url, width=750, height=350)
```
---


## Column {.sidebar}

Interact directly with your model via its visual documentation, and get `curl` examples.

---
name: Release checklist
about: prepare for relase
title: Release vetiver X.X.X
labels: release
assignees: ''

---

Prepare for release:

- [ ] `pre-commit run`
- [ ] `pytest vetiver/tests/`
- [ ] bump release number

Submit to PyPI:
- [ ] submit 🎉

---
title: "Interlinks Example"
filters:
  - interlinks.py
interlinks:
  sources:
    siuba:
      url: https://siuba.org/
      inv: null
      fallback: objects_siuba.json
    vetiver:
      url: https://vetiver.rstudio.com/
      inv: null
      fallback: objects_vetiver.json
---

## Heading {#sec-hello}

## Testing


| style | syntax | output |
| ----- | ------ | ------ |
| md custom text  | `[some explanation](`vetiver.SKLearnHandler`)` | [some explanation](`vetiver.SKLearnHandler`) |
| md blank text   | `[](`vetiver.SKLearnHandler`)` | [](`vetiver.SKLearnHandler`) |
| md blank text (~shortened)| `[](`~vetiver.SKLearnHandler`)` | [](`~vetiver.SKLearnHandler`) |
| rst custom text | `` :ref:`some explanation <vetiver.SKLearnHandler>` `` | :ref:`some explanation <vetiver.SKLearnHandler>` |
| rst blank text  | `` :ref:`vetiver.SKLearnHandler` `` | :ref:`vetiver.SKLearnHandler` |
| rst blank text (~shortened) | `` :ref:`~vetiver.SKLearnHandler` `` | :ref:`~vetiver.SKLearnHandler` |

# Who maintains Vetiver

The vetiver package was created by and is currently maintained by Isabel Zimmerman <isabel.zimmerman@posit.co>. [Posit Software, PBC](https://posit.co/products/open-source/) is a copyright holder and funder of the vetiver package.

Several individuals in the community have taken an active role in helping to maintain vetiver and submit fixes. Those individuals are shown in the git changelog.

---
title: "Version, share, and deploy a model with vetiver"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{vetiver}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = requireNamespace("parsnip", quietly = TRUE) &&
    requireNamespace("recipes", quietly = TRUE) &&
    requireNamespace("workflows", quietly = TRUE)
)
```

The goal of vetiver is to provide fluent tooling for MLOps tasks for your trained model including:

- versioning
- storing
- sharing
- deploying

For more extensive documentation, visit <https://vetiver.posit.co/>.

## Create a `vetiver_model()`

The vetiver package is extensible, with generics that can support many kinds of models. For this example, let's consider one kind of model supported by vetiver, a [tidymodels](https://www.tidymodels.org/) workflow that encompasses both feature engineering and model estimation.

```{r}
library(parsnip)
library(recipes)
library(workflows)
data(bivariate, package = "modeldata")
bivariate_train

biv_rec <-
  recipe(Class ~ ., data = bivariate_train) |>
  step_BoxCox(all_predictors()) |>
  step_normalize(all_predictors())

svm_spec <-
  svm_linear(mode = "classification") |>
  set_engine("LiblineaR")

svm_fit <-
  workflow(biv_rec, svm_spec) |>
  fit(sample_frac(bivariate_train, 0.7))
```

This `svm_fit` object is a fitted workflow, with both feature engineering and model parameters estimated using the training data `bivariate_train`. We can create a `vetiver_model()` from this trained model; a `vetiver_model()` collects the information needed to store, version, and deploy a trained model.

```{r}
library(vetiver)
v <- vetiver_model(svm_fit, "biv_svm")
v
```

Think of this `vetiver_model()` as a deployable model object.

## Store and version your model

You can store and version your model by choosing a [pins](https://pins.rstudio.com) "board" for it, including a local folder, Posit Connect, Amazon S3, and more. Most pins boards have versioning turned on by default, but we can turn it on explicitly for our temporary demo board. When we write the `vetiver_model()` to our board, the binary model object is stored on our board together with necessary metadata, like the packages needed to make a prediction and the model's input data prototype for checking new data at prediction time.

```{r}
#| message: false
library(pins)
model_board <- board_temp(versioned = TRUE)
model_board |> vetiver_pin_write(v)
```

Let's train our model again with a new version of the dataset and write it once more to our board.

```{r}
#| message: false
svm_fit <-
  workflow(biv_rec, svm_spec) |>
  fit(sample_frac(bivariate_train, 0.7))

v <- vetiver_model(svm_fit, "biv_svm")

model_board |> vetiver_pin_write(v)
```

Both versions are stored, and we have access to both.

```{r}
model_board |> pin_versions("biv_svm")
```

The primary purpose of pins is to make it easy to share data artifacts, so depending on the board you choose, your pinned `vetiver_model()` can be shareable with your collaborators.

## Deploy your model

You can deploy your model by creating a [Plumber](https://www.rplumber.io/) router, and adding a POST endpoint for making predictions.

```{r}
library(plumber)
pr() |>
  vetiver_api(v)
```

To start a server using this object, pipe (`|>`) to `pr_run(port = 8088)` or your port of choice. This allows you to interact with your vetiver API locally and debug it. Plumber APIs such as these can be [hosted in a variety of ways](https://www.rplumber.io/articles/hosting.html). You can use the function `vetiver_write_plumber()` to create a ready-to-go `plumber.R` file that is especially suited for [Posit Connect](https://posit.co/products/enterprise/connect).

```{r}
#| eval: false
vetiver_write_plumber(model_board, "biv_svm")
```


```{r}
#| echo: false
#| comment: ''
tmp <- tempfile()
vetiver_write_plumber(model_board, "biv_svm", file = tmp)
cat(readr::read_lines(tmp), sep = "\n")
```

In a real-world situation, you would see something like `b <- board_connect()` or `b <- board_s3()` here instead of our temporary demo board. Notice that the deployment is strongly linked to a *specific version* of the pinned model; if you pin another version of the model after you deploy your model, your deployed model will not be affected.

## Predict from your model endpoint

A model deployed via vetiver can be treated as a special `vetiver_endpoint()` object.

```{r}
library(vetiver)
endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
endpoint
```

If such a deployed model endpoint is running via one R process (either remotely on a server or locally, perhaps via a [background job in the RStudio IDE](https://docs.posit.co/ide/user/ide/guide/tools/jobs.html)), you can make predictions with that deployed model and new data in another, separate R process. 

``` r
data(bivariate, package = "modeldata")
predict(endpoint, bivariate_test)
#> # A tibble: 710 × 1
#>    .pred_class
#>    <chr>      
#>  1 One        
#>  2 Two        
#>  3 One        
#>  4 Two        
#>  5 Two        
#>  6 One        
#>  7 Two        
#>  8 Two        
#>  9 Two        
#> 10 One        
#> # … with 700 more rows
```

Being able to `predict()` on a vetiver model endpoint takes advantage of the model's input data prototype and other metadata that is stored with the model.

# mlr3 learner description can be printed

    Code
      v
    Output
      
      -- rpart_penguins - <LearnerClassifRpart> model for deployment 
      A mlr3 classif.rpart learner using 7 features

# create plumber.R for mlr3

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(mlr3)
          library(rpart)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "rpart_penguins")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R with no packages

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R for complicated board

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_s3(bucket = "foo", region = 6, endpoint = 8)
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R for URL board

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_url(c(foo = "foo", bar = "bar", baz = "baz"))
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R with packages

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(beepr)
          library(janeaustenr)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R with extra infra packages

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(beepr)
      library(janeaustenr)
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R with rsconnect = FALSE

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create plumber.R with args in dots

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(beepr)
          library(janeaustenr)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          vetiver_api(pr, v, debug = TRUE, endpoint = "/predict2", type = "numeric")
      }

# can print glm model

    Code
      v
    Output
      
      -- cars_glm - <butchered_glm> model for deployment 
      A generalized linear model (gaussian family, identity link) using 10 features

# create plumber.R for glm

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_glm")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print gam model

    Code
      v
    Output
      
      -- cars_gam - <butchered_gam> model for deployment 
      A generalized additive model (gaussian family, identity link) using 2 features

# create plumber.R for gam

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(mgcv)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_gam")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print ranger model

    Code
      v
    Output
      
      -- cars3 - <ranger> model for deployment 
      A ranger regression model using 10 features

# error for no prototype_data with ranger

    Code
      vetiver_model(cars_rf, "cars3")
    Condition
      Error in `vetiver_ptype()`:
      ! No `prototype_data` available to create an input data prototype
      * Pass at least one row of training features as `prototype_data`
      * See the documentation for `vetiver_ptype()`

# create plumber.R for ranger

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(ranger)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars3")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print tidymodels model

    Code
      v
    Output
      
      -- cars_wf - <bundled_workflow> model for deployment 
      A ranger regression modeling workflow using 10 features

# create plumber.R for tidymodels

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(parsnip)
          library(ranger)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_wf")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# create Dockerfile with packages

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY <redacted>/plumber.R /opt/ml/plumber.R
      EXPOSE 8000
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8000)"]

# create Dockerfile with 'additional' packages

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY <redacted>/plumber.R /opt/ml/plumber.R
      EXPOSE 8000
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8000)"]

# create Dockerfile with no RSPM

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY <redacted>/plumber.R /opt/ml/plumber.R
      EXPOSE 8000
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8000)"]

# create Dockerfile with no packages

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY <redacted>/plumber.R /opt/ml/plumber.R
      EXPOSE 8000
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8000)"]

# create Dockerfile with specific port

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY <redacted>/plumber.R /opt/ml/plumber.R
      
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = as.numeric(Sys.getenv("PORT")))"]

# create all files needed for Docker

    Code
      cat(readr::read_lines(file.path(tmp_dir, "plumber.R")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          vetiver_api(pr, v, path = "cars")
      }

---

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY plumber.R /opt/ml/plumber.R
      EXPOSE 8000
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8000)"]

# can print stacks model

    Code
      v
    Output
      
      -- frog-stack - <bundled_model_stack> model for deployment 
      A regression stacked ensemble with 3 members using 4 features

# create plumber.R for stacks

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(glmnet)
          library(parsnip)
          library(recipes)
          library(stacks)
          library(stats)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "frog-stack")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print caret model

    Code
      v
    Output
      
      -- cars_rf - <bundled_train> model for deployment 
      A random forest regression model using 3 features

# create plumber.R for xgboost

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(caret)
          library(dplyr)
          library(e1071)
          library(ranger)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_rf")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print recipe

    Code
      v
    Output
      
      -- car-splines - <bundled_recipe> model for deployment 
      A feature engineering recipe with 1 step using 2 features

# create plumber.R for recipe

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(recipes)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "car-splines")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can deploy via `vetiver_deploy_sagemaker()`

    Code
      vetiver_deploy_sagemaker(b, "cars1", instance_type)
    Condition
      Error in `vetiver_deploy_sagemaker()`:
      ! `board` must be an S3 pins board, not a <pins_board_folder/pins_board> object.

# can create correct files for `vetiver_sm_build()`

    Code
      cat(readr::read_lines(fs::path(tmp_dir, "plumber.R")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          vetiver_api(pr, v, path = "/invocations")
      }

---

    Code
      cat(readr::read_lines(file.path(tmp_dir, "Dockerfile")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      FROM rocker/r-ver:<r_version>
      ENV RENV_CONFIG_REPOS_OVERRIDE https://packagemanager.rstudio.com/cran/latest
      
      RUN apt-get update -qq && apt-get install -y --no-install-recommends \
        cmake \
        libcurl4-openssl-dev \
        libicu-dev \
        libsodium-dev \
        libssl-dev \
        libuv1-dev \
        libx11-dev \
        make \
        zlib1g-dev \
        && apt-get clean
      
      COPY vetiver_renv.lock renv.lock
      RUN Rscript -e "install.packages('renv')"
      RUN Rscript -e "renv::restore()"
      COPY plumber.R /opt/ml/plumber.R
      EXPOSE 8080
      ENTRYPOINT ["R", "-e", "pr <- plumber::plumb('/opt/ml/plumber.R'); pr$run(host = '0.0.0.0', port = 8080)"]

# can create SageMaker Model

    Code
      vetiver_sm_model(image_uri = image_uri, role = role, tags = "potato")
    Condition
      Error in `vetiver_sm_model()`:
      ! `tags` must be a list, not the string "potato".
    Code
      vetiver_sm_model(image_uri = image_uri, role = role, tags = list("potato"))
    Condition
      Error in `vetiver_sm_model()`:
      ! `tags` must have valid names, like `list("my-tag" = "my-value")`
    Code
      vetiver_sm_model(image_uri = image_uri, role = role, vpc_config = list("potato"))
    Condition
      Error in `vetiver_sm_model()`:
      ! `subnet` must be a list, not `NULL`.
    Code
      vetiver_sm_model(image_uri = image_uri, role = role, vpc_config = list(Subnets = list(
        1:3), SecurityGroupIds = 1:3))
    Condition
      Error in `vetiver_sm_model()`:
      ! `security_group_ids` must be a list, not an integer vector.
    Code
      vetiver_sm_model(image_uri = image_uri, role = role, vpc_config = list(Subnets = 1:
        3, SecurityGroupIds = list(1:3)))
    Condition
      Error in `vetiver_sm_model()`:
      ! `subnet` must be a list, not an integer vector.

# can create SageMaker Endpoint

    Code
      vetiver_sm_endpoint(model_name = model_name, instance_type = instance_type,
        tags = "potato")
    Condition
      Error in `vetiver_sm_endpoint()`:
      ! `tags` must be a list, not the string "potato".
    Code
      vetiver_sm_endpoint(model_name = model_name, instance_type = instance_type,
        tags = list("potato"))
    Condition
      Error in `vetiver_sm_endpoint()`:
      ! `tags` must have valid names, like `list("my-tag" = "my-value")`

# can call sm_create_endpoint

    Code
      out <- sm_create_endpoint(client, endpoint_name, config_name, wait = FALSE)
    Message
      Creating endpoint with name "vetiver-sagemaker-example-model"

# can delete SageMaker endpoints

    Code
      vetiver_sm_delete(object)
    Condition
      Warning:
      Unable to delete "vetiver-sagemaker-example-model" endpoint configuration.

# can create vetiver endpoint object

    Code
      vetiver_endpoint_sagemaker(list())
    Condition
      Error in `vetiver_endpoint_sagemaker()`:
      ! `model_endpoint` must be a character vector, not an empty list.

---

    Code
      vetiver_endpoint_sagemaker(model_endpoint)
    Output
      
      -- A SageMaker model endpoint for prediction: 
      Model endpoint: vetiver-sagemaker-example-model
      Region: my-region

# can print kproto model

    Code
      v
    Output
      
      -- kproto-example - <butchered_kproto> model for deployment 
      A k-prototypes clustering model (3 clusters) using 3 features

# create plumber.R for kproto

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(clustMixType)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "kproto-example")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can attach pkgs

    Code
      attach_pkgs(c("knitr", "readr"))

# can fail on a single pkg

    Code
      attach_pkgs(c("potato"))
    Condition
      Error in `attach_pkgs()`:
      ! Package(s) could not be attached:
      * potato

# can fail on multiple pkgs

    Code
      attach_pkgs(c("potato", "bloopy"))
    Condition
      Error in `attach_pkgs()`:
      ! Package(s) could not be attached:
      * potato
      * bloopy

---

    Code
      attach_pkgs(c("potato", "readr"))
    Condition
      Error in `attach_pkgs()`:
      ! Package(s) could not be attached:
      * potato

# can predict on basic vetiver router

    Code
      print(endpoint)
    Output
      
      -- A model API endpoint for prediction: 
      http://localhost:<port>/predict

# get correct errors

    Code
      predict(endpoint, mtcars[, 2:4])
    Condition
      Error in `predict()`:
      ! Failed to predict: Error in `hardhat::scream()`:
      ! Can't convert from `data` <tbl_df<
        cyl : integer
        disp: double
        hp  : integer
      >> to <tbl_df<
        cyl : double
        disp: double
      >> due to loss of precision.

---

    Code
      predict(endpoint, mtcars[, 3:5])
    Condition
      Error in `predict()`:
      ! Failed to predict: Error in `hardhat::validate_column_names()`:
      ! The required column "cyl" is missing.

# vetiver_pin_metrics() / fails without existing pin

    Can't find pin called "metrics1"
    i Use `pin_list()` to see all available pins in this board

# vetiver_pin_metrics() / fails with `overwrite = FALSE`

    The new metrics overlap with dates already stored in 'metrics2'
    i Check the aggregated dates or use `overwrite = TRUE`

# vetiver_pin_metrics() / fails with type = csv

    Metrics must be pinned as "arrow" or "rds", not "csv"

---

    Metrics must be pinned as "arrow" or "rds", not "csv"

# can print xgboost model

    Code
      v
    Output
      
      -- cars2 - <bundled_xgboost> model for deployment 
      An xgboost reg:squarederror model using 10 features

# create plumber.R for xgboost

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(xgboost)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars2")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# missing variables

    The required columns "mpg", "hp", "drat", "wt", "qsec", "vs", "am", "gear", and "carb" are missing.

# a factor plus a bad character

    [0, 2]: expected value in level set, but got 'ZZ'

# a date

    [0, 1]: expected date like , but got 'potato'

# can print int_conformal_split model

    Code
      v_split
    Output
      
      -- cars_int_split - <butchered_int_conformal_split> model for deployment 
      A Split Conformal inference with a ranger regression model using 10 features

# create plumber.R for int_conformal_split

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(parsnip)
          library(probably)
          library(ranger)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_int_split")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print int_conformal_full model

    Code
      v_full
    Output
      
      -- cars_int_full - <butchered_int_conformal_full> model for deployment 
      A full Conformal inference with a ranger regression model using 10 features

# create plumber.R for int_conformal_full

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(parsnip)
          library(probably)
          library(ranger)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_int_full")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print int_conformal_quantile model

    Code
      v_quantile
    Output
      
      -- cars_int_quantile - <butchered_int_conformal_quantile> model for deployment 
      A quantile Conformal inference with a ranger regression model using 10 features

# create plumber.R for int_conformal_quantile

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(parsnip)
          library(probably)
          library(ranger)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_int_quantile")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print int_conformal_cv model

    Code
      v_cv
    Output
      
      -- cars_int_cv - <butchered_int_conformal_cv> model for deployment 
      A 10-fold CV+ Conformal inference with a ranger regression model using 10
      features

# create plumber.R for int_conformal_cv

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(parsnip)
          library(probably)
          library(ranger)
          library(workflows)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars_int_cv")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can pin a model

    Code
      v
    Output
      
      -- cars1 - <butchered_lm> model for deployment 
      An OLS linear regression model using 2 features

# can pin a model with no prototype

    The `save_ptype` argument of `vetiver_model()` is deprecated as of vetiver 0.2.0.
    i Please use the `save_prototype` argument instead.

# right message for reading with `check_renv`

    There is no lockfile stored with "cars5":
    i Use `check_renv = TRUE` when you save your model to your board

# can print keras model

    Code
      v
    Output
      
      -- cars-keras - <bundled_keras> model for deployment 
      A sequential keras model with 2 layers using 10 features

# create plumber.R for keras

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(keras)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars-keras")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

---

    Code
      cat(readr::read_lines(fs::path(fs::path_dir(tmp), "requirements.txt")), sep = "\n")
    Output
      keras
      tensorflow

---

    Code
      vetiver_write_docker(v, tmp, tmp_dir)
    Condition
      Warning:
      Your `vetiver_model` object contains a keras model
      i Be sure to use an appropriate `base_image` such as `rocker/cuda`

# can warn for strange setups

    Code
      p <- choose_version(p)
    Condition
      Warning:
      Pinned vetiver model has no active version and no datetime on versions
      * Do you need to check your pinned model?
      * Using version 4500

# create rsconnect bundle / contains correct plumber file

    Code
      cat(readr::read_lines(fs::path(tmp_dir, "plumber.R")), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars1")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

# can print a `vetiver`ed luz model

    Code
      v
    Output
      
      -- cars-luz - <bundled_luz_module_fitted> model for deployment 
      A luz module with 11 parameters using 10 features

# create plumber.R for keras

    Code
      cat(readr::read_lines(tmp), sep = "\n")
    Output
      # Generated by the vetiver package; edit with care
      
      library(pins)
      library(plumber)
      library(rapidoc)
      library(vetiver)
      
      # Packages needed to generate model predictions
      if (FALSE) {
          library(luz)
          library(torch)
      }
      b <- board_folder(path = "<redacted>")
      v <- vetiver_pin_read(b, "cars-luz")
      
      #* @plumber
      function(pr) {
          pr |> vetiver_api(v)
      }

---

    Code
      cat(readr::read_lines(fs::path(fs::path_dir(tmp), ".Renviron")), sep = "\n")
    Output
      TORCH_INSTALL=1
      TORCH_HOME="libtorch/"

# vetiver_dashboard() / errors without pin

    Code
      vetiver_dashboard(pins = list(board = b, name = "seattle_rf", version = NULL))
    Condition
      Error in `vetiver_dashboard()`:
      ! The pin "seattle_rf" does not exist on your board
      * Check the `pins` params in your dashboard YAML
      i To knit the example vetiver monitoring dashboard, execute `vetiver::pin_example_kc_housing_model()` to set up demo model and metrics pins

# old function is deprecated

    `vetiver_pr_predict()` was deprecated in vetiver 0.1.2 and is now defunct.
    i Please use `vetiver_api()` instead.

# OpenAPI spec for check_prototype = FALSE

    The `check_ptype` argument of `vetiver_pr_post()` is deprecated as of vetiver 0.2.0.
    i Please use the `check_prototype` argument instead.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vetiver-package.R
\docType{package}
\name{vetiver-package}
\alias{vetiver}
\alias{vetiver-package}
\title{vetiver: Version, Share, Deploy, and Monitor Models}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

The goal of 'vetiver' is to provide fluent tooling to version, share, deploy, and monitor a trained model. Functions handle both recording and checking the model's input data prototype, and predicting from a remote API endpoint. The 'vetiver' package is extensible, with generics that can support many kinds of models.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://vetiver.posit.co}
  \item \url{https://rstudio.github.io/vetiver-r/}
  \item \url{https://github.com/rstudio/vetiver-r/}
  \item Report bugs at \url{https://github.com/rstudio/vetiver-r/issues}
}

}
\author{
\strong{Maintainer}: Julia Silge \email{julia.silge@posit.co} (\href{https://orcid.org/0000-0002-3671-836X}{ORCID})

Authors:
\itemize{
  \item Julia Silge \email{julia.silge@posit.co} (\href{https://orcid.org/0000-0002-3671-836X}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-plumber.R
\name{vetiver_write_plumber}
\alias{vetiver_write_plumber}
\title{Write a deployable Plumber file for a vetiver model}
\usage{
vetiver_write_plumber(
  board,
  name,
  version = NULL,
  ...,
  file = "plumber.R",
  rsconnect = TRUE,
  additional_pkgs = character(0)
)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}

\item{...}{Other arguments passed to \code{\link[=vetiver_api]{vetiver_api()}} such as the endpoint
\code{path} or prediction \code{type}.}

\item{file}{A path to write the Plumber file. Defaults to \code{plumber.R} in the
working directory. See \code{\link[plumber:plumb]{plumber::plumb()}} for naming precedence rules.}

\item{rsconnect}{Create a Plumber file with features needed for \href{https://posit.co/products/enterprise/connect}{Posit Connect}? Defaults to \code{TRUE}.}

\item{additional_pkgs}{Any additional R packages that need to be \strong{attached}
via \code{\link[=library]{library()}} to run your API, as a character vector.}
}
\value{
The content of the \code{plumber.R} file, invisibly.
}
\description{
Use \code{vetiver_write_plumber()} to create a \code{plumber.R} file for a
\code{\link[=vetiver_model]{vetiver_model()}} that has been versioned and stored via
\code{\link[=vetiver_pin_write]{vetiver_pin_write()}}.
}
\details{
By default, this function will find and use the latest version of your
vetiver model; the model API (when deployed) will be linked to that specific
version. You can override this default behavior by choosing a specific
\code{version}.
}
\examples{
\dontshow{if (rlang::is_installed("plumber")) withAutoprint(\{ # examplesIf}
library(pins)
tmp <- tempfile()
b <- board_temp(versioned = TRUE)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

vetiver_write_plumber(b, "cars_linear", file = tmp)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keras.R, R/luz.R, R/write-plumber.R
\name{vetiver_python_requirements.keras.engine.training.Model}
\alias{vetiver_python_requirements.keras.engine.training.Model}
\alias{vetiver_renviron_requirements.luz_module_fitted}
\alias{vetiver_python_requirements}
\alias{vetiver_python_requirements.default}
\alias{vetiver_renviron_requirements}
\alias{vetiver_renviron_requirements.default}
\title{Use extra files required for deployment}
\usage{
\method{vetiver_python_requirements}{keras.engine.training.Model}(model)

\method{vetiver_renviron_requirements}{luz_module_fitted}(model)

vetiver_python_requirements(model)

\method{vetiver_python_requirements}{default}(model)

vetiver_renviron_requirements(model)

\method{vetiver_renviron_requirements}{default}(model)
}
\arguments{
\item{model}{A trained model, such as an \code{lm()} model or a tidymodels
\code{\link[workflows:workflow]{workflows::workflow()}}.}
}
\description{
Create files required for deploying an app generated via
\code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}}, such as a Python \code{requirements.txt} or an
\code{.Renviron}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsconnect.R
\name{vetiver_deploy_rsconnect}
\alias{vetiver_deploy_rsconnect}
\title{Deploy a vetiver model API to Posit Connect}
\usage{
vetiver_deploy_rsconnect(
  board,
  name,
  version = NULL,
  predict_args = list(),
  appTitle = glue::glue("{name} model API"),
  ...,
  additional_pkgs = character(0)
)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}

\item{predict_args}{A list of optional arguments passed to \code{\link[=vetiver_api]{vetiver_api()}}
such as the prediction \code{type}.}

\item{appTitle}{The API title on Posit Connect. Use the default based on
\code{name}, or pass in your own title.}

\item{...}{Other arguments passed to \code{\link[rsconnect:deployApp]{rsconnect::deployApp()}} such as
\code{appName}, \code{account}, or \code{launch.browser}.}

\item{additional_pkgs}{Any additional R packages that need to be \strong{attached}
via \code{\link[=library]{library()}} to run your API, as a character vector.}
}
\value{
The deployment success (\code{TRUE} or \code{FALSE}), invisibly.
}
\description{
Use \code{vetiver_deploy_rsconnect()} to deploy a \code{\link[=vetiver_model]{vetiver_model()}} that has been
versioned and stored via \code{\link[=vetiver_pin_write]{vetiver_pin_write()}} as a Plumber API on \href{https://docs.posit.co/connect/}{Posit Connect}.
}
\details{
The two functions \code{vetiver_deploy_rsconnect()} and
\code{\link[=vetiver_create_rsconnect_bundle]{vetiver_create_rsconnect_bundle()}} are alternatives to each other, providing
different strategies for deploying a vetiver model API to Posit Connect.

When you first deploy to Connect, your API will only be accessible to you.
You can \href{https://docs.posit.co/connect/user/content-settings/#set-viewers}{change the access settings}
so others can also access the API. For all access settings other than
"Anyone - no login required", anyone querying your API (including you)
will need to pass authentication details with your API call,
\href{https://docs.posit.co/connect/user/vetiver/#predict-from-your-model-endpoint}{as shown in the Connect documentation}.
}
\examples{
\dontshow{if (rlang::is_installed("rsconnect")) withAutoprint(\{ # examplesIf}
library(pins)
b <- board_temp(versioned = TRUE)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

if (FALSE) {
## pass args for predicting:
vetiver_deploy_rsconnect(
    b,
    "user.name/cars_linear",
    predict_args = list(debug = TRUE)
)

## specify an account name through `...`:
vetiver_deploy_rsconnect(
    b,
    "user.name/cars_linear",
    account = "user.name"
)
}
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}}, \code{\link[=vetiver_create_rsconnect_bundle]{vetiver_create_rsconnect_bundle()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open-api-spec.R
\name{api_spec}
\alias{api_spec}
\alias{glue_spec_summary}
\alias{glue_spec_summary.default}
\alias{glue_spec_summary.data.frame}
\alias{glue_spec_summary.array}
\title{Update the OpenAPI specification using model metadata}
\usage{
api_spec(spec, vetiver_model, path, all_docs = TRUE)

glue_spec_summary(prototype, return_type)

\method{glue_spec_summary}{default}(prototype, return_type = NULL)

\method{glue_spec_summary}{data.frame}(prototype, return_type = "predictions")

\method{glue_spec_summary}{array}(prototype, return_type = "predictions")
}
\arguments{
\item{spec}{An OpenAPI Specification formatted list object}

\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{path}{The endpoint path}

\item{all_docs}{Should the interactive visual API documentation be created
for \emph{all} POST endpoints in the router \code{pr}? This defaults to \code{TRUE}, and
assumes that all POST endpoints use the \code{\link[=vetiver_model]{vetiver_model()}} input data
prototype.}

\item{prototype}{An input data prototype from a model}

\item{return_type}{Character string to describe what endpoint returns, such
as "predictions"}
}
\value{
\code{api_spec()} returns the updated OpenAPI Specification object. This
function uses \code{glue_spec_summary()} internally, which returns a \code{glue}
character string.
}
\description{
Update the OpenAPI specification using model metadata
}
\examples{
\dontshow{if (rlang::is_installed("plumber")) withAutoprint(\{ # examplesIf}
library(plumber)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")

glue_spec_summary(v$prototype)

modify_spec <- function(spec) api_spec(spec, v, "/predict")
pr() |> pr_set_api_spec(api = modify_spec)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vetiver-model.R
\name{vetiver_model}
\alias{vetiver_model}
\alias{new_vetiver_model}
\title{Create a vetiver object for deployment of a trained model}
\usage{
vetiver_model(
  model,
  model_name,
  ...,
  description = NULL,
  metadata = list(),
  save_prototype = TRUE,
  save_ptype = deprecated(),
  versioned = NULL
)

new_vetiver_model(
  model,
  model_name,
  description,
  metadata,
  prototype,
  versioned
)
}
\arguments{
\item{model}{A trained model, such as an \code{lm()} model or a tidymodels
\code{\link[workflows:workflow]{workflows::workflow()}}.}

\item{model_name}{Model name or ID.}

\item{...}{Other method-specific arguments passed to \code{\link[=vetiver_ptype]{vetiver_ptype()}}
to compute an input data prototype, such as \code{prototype_data} (a sample of
training features).}

\item{description}{A detailed description of the model. If omitted, a brief
description of the model will be generated.}

\item{metadata}{A list containing additional metadata to store with the pin.
When retrieving the pin, this will be stored in the \code{user} key, to
avoid potential clashes with the metadata that pins itself uses.}

\item{save_prototype}{Should an input data prototype be stored with the model?
The options are \code{TRUE} (the default, which stores a zero-row slice of the
training data), \code{FALSE} (no input data prototype for visual documentation or
checking), or a dataframe to be used for both checking at prediction time
\emph{and} examples in API visual documentation.}

\item{save_ptype}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}

\item{versioned}{Should the model object be versioned when stored with
\code{\link[=vetiver_pin_write]{vetiver_pin_write()}}? The default, \code{NULL}, will use the default for the
\code{board} where you store the model.}

\item{prototype}{An input data prototype. If \code{NULL}, there is no checking of
new data at prediction time.}
}
\value{
A new \code{vetiver_model} object.
}
\description{
A \code{vetiver_model()} object collects the information needed to store, version,
and deploy a trained model. Once your \code{vetiver_model()} object has been
created, you can:
\itemize{
\item store and version it as a pin with \code{\link[=vetiver_pin_write]{vetiver_pin_write()}}
\item create an API endpoint for it with \code{\link[=vetiver_api]{vetiver_api()}}
}
}
\details{
You can provide your own data to \code{save_prototype} to use as examples in the
visual documentation created by \code{\link[=vetiver_api]{vetiver_api()}}. If you do this,
consider checking that your input data prototype has the same structure
as your training data (perhaps with \code{\link[hardhat:scream]{hardhat::scream()}}) and/or simulating
data to avoid leaking PII via your deployed model.

Some models, like \code{\link[ranger:ranger]{ranger::ranger()}}, \href{https://tensorflow.rstudio.com/}{keras},
and \href{https://torch.mlverse.org/}{luz (torch)},
\emph{require} that you pass in example training data as \code{prototype_data}
or else explicitly set \code{save_prototype = FALSE}. For non-rectangular data
input to models, such as image input for a keras or torch model, we currently
recommend that you turn off prototype checking via \code{save_prototype = FALSE}.
}
\examples{

cars_lm <- lm(mpg ~ ., data = mtcars)
vetiver_model(cars_lm, "cars-linear")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoint.R
\name{predict.vetiver_endpoint}
\alias{predict.vetiver_endpoint}
\title{Post new data to a deployed model API endpoint and return predictions}
\usage{
\method{predict}{vetiver_endpoint}(object, new_data, ...)
}
\arguments{
\item{object}{A model API endpoint object created with \code{\link[=vetiver_endpoint]{vetiver_endpoint()}}.}

\item{new_data}{New data for making predictions, such as a data frame.}

\item{...}{Extra arguments passed to \code{\link[httr:POST]{httr::POST()}}}
}
\value{
A tibble of model predictions with as many rows as in \code{new_data}.
}
\description{
Post new data to a deployed model API endpoint and return predictions
}
\examples{

if (FALSE) {
endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
predict(endpoint, mtcars[4:7, -1])
}


}
\seealso{
\code{\link[=augment.vetiver_endpoint]{augment.vetiver_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{vetiver_deploy_sagemaker}
\alias{vetiver_deploy_sagemaker}
\title{Deploy a vetiver model API to Amazon SageMaker}
\usage{
vetiver_deploy_sagemaker(
  board,
  name,
  instance_type,
  ...,
  predict_args = list(),
  docker_args = list(),
  build_args = list(),
  endpoint_args = list(),
  repo_name = glue("vetiver-sagemaker-{name}")
)
}
\arguments{
\item{board}{An AWS S3 board created with \code{\link[pins:board_s3]{pins::board_s3()}}. This board
must be in the correct region for your SageMaker instance.}

\item{name}{Pin name.}

\item{instance_type}{Type of EC2 instance to use; see
\href{https://aws.amazon.com/sagemaker/pricing/}{Amazon SageMaker pricing}.}

\item{...}{Not currently used.}

\item{predict_args}{A list of optional arguments passed to \code{\link[=vetiver_api]{vetiver_api()}}
such as the prediction \code{type}.}

\item{docker_args}{A list of optional arguments passed to
\code{\link[=vetiver_write_docker]{vetiver_write_docker()}} such as the \code{lockfile} name or whether to use
\code{rspm}. Do not pass \code{additional_pkgs} here, as this function uses
\code{additional_pkgs = required_pkgs(board)}.}

\item{build_args}{A list of optional arguments passed to
\code{\link[=vetiver_sm_build]{vetiver_sm_build()}} such as the model \code{version} or the \code{compute_type}.}

\item{endpoint_args}{A list of optional arguments passed to
\code{\link[=vetiver_sm_endpoint]{vetiver_sm_endpoint()}} such as \code{accelerator_type} or \code{data_capture_config}.}

\item{repo_name}{The name for the AWS ECR repository to store the model.}
}
\value{
The deployed \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}.
}
\description{
Use \code{vetiver_deploy_sagemaker()} to deploy a \code{\link[=vetiver_model]{vetiver_model()}}
that has been versioned and stored via \code{\link[=vetiver_pin_write]{vetiver_pin_write()}} as a Plumber API
on \href{https://aws.amazon.com/sagemaker/}{Amazon SageMaker}.
}
\details{
This function stores your model deployment image in the same bucket used
by \code{board}.

The function \code{vetiver_deploy_sagemaker()} uses:
\itemize{
\item \code{\link[=vetiver_sm_build]{vetiver_sm_build()}} to build and push a Docker image to ECR,
\item \code{\link[=vetiver_sm_model]{vetiver_sm_model()}} to create a SageMaker model, and
\item \code{\link[=vetiver_sm_endpoint]{vetiver_sm_endpoint()}} to deploy a SageMaker model endpoint.
}

These modular functions are available for more advanced use cases.

If you are working locally, you will likely need to explicitly set up
your execution role to work correctly. Check out
\href{https://dyfanjones.r-universe.dev/smdocker}{"Execution role requirements"}
in the smdocker documentation, and especially note that the bucket containing
your vetiver model needs to be added as a resource in your IAM role policy.
}
\examples{
if (FALSE) {
library(pins)
b <- board_s3(bucket = "my-existing-bucket")
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

endpoint <- vetiver_deploy_sagemaker(
    board = b,
    name = "cars_linear",
    instance_type = "ml.t2.medium",
    predict_args = list(type = "class", debug = TRUE)
)
}

}
\seealso{
\code{\link[=vetiver_sm_build]{vetiver_sm_build()}}, \code{\link[=vetiver_sm_model]{vetiver_sm_model()}}, \code{\link[=vetiver_sm_endpoint]{vetiver_sm_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open-api-spec.R
\name{map_request_body}
\alias{map_request_body}
\title{Identify data types for each column in an input data prototype}
\usage{
map_request_body(prototype)
}
\arguments{
\item{prototype}{An input data prototype from a model}
}
\value{
A list to be used within \code{\link[plumber:pr_set_api_spec]{plumber::pr_set_api_spec()}}
}
\description{
The OpenAPI specification of a Plumber API created via \code{\link[plumber:pr]{plumber::pr()}} can
be modified via \code{\link[plumber:pr_set_api_spec]{plumber::pr_set_api_spec()}}, and this helper function will
identify data types of predictors and create a list to use in this
specification. These are \emph{not} R data types, but instead basic JSON data
types. For example, factors in R will be documented as strings in the
OpenAPI specification.
}
\details{
This is a developer-facing function, useful for supporting new model types.
It is called by \code{\link[=api_spec]{api_spec()}}.
}
\examples{
map_request_body(vctrs::vec_slice(chickwts, 0))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caret.R, R/gam.R, R/keras.R, R/kproto.R,
%   R/luz.R, R/meta.R, R/mlr3.R, R/probably.R, R/ranger.R, R/recipe.R,
%   R/stacks.R, R/tidymodels.R, R/xgboost.R
\name{vetiver_create_meta.train}
\alias{vetiver_create_meta.train}
\alias{vetiver_create_meta.gam}
\alias{vetiver_create_meta.keras.engine.training.Model}
\alias{vetiver_create_meta.kproto}
\alias{vetiver_create_meta.luz_module_fitted}
\alias{vetiver_meta}
\alias{vetiver_create_meta}
\alias{vetiver_create_meta.default}
\alias{vetiver_create_meta.Learner}
\alias{vetiver_create_meta.int_conformal_split}
\alias{vetiver_create_meta.int_conformal_full}
\alias{vetiver_create_meta.int_conformal_quantile}
\alias{vetiver_create_meta.int_conformal_cv}
\alias{vetiver_create_meta.ranger}
\alias{vetiver_create_meta.recipe}
\alias{vetiver_create_meta.model_stack}
\alias{vetiver_create_meta.workflow}
\alias{vetiver_create_meta.xgb.Booster}
\title{Metadata constructors for \code{vetiver_model()} object}
\usage{
\method{vetiver_create_meta}{train}(model, metadata)

\method{vetiver_create_meta}{gam}(model, metadata)

\method{vetiver_create_meta}{keras.engine.training.Model}(model, metadata)

\method{vetiver_create_meta}{kproto}(model, metadata)

\method{vetiver_create_meta}{luz_module_fitted}(model, metadata)

vetiver_meta(user = list(), version = NULL, url = NULL, required_pkgs = NULL)

vetiver_create_meta(model, metadata)

\method{vetiver_create_meta}{default}(model, metadata)

\method{vetiver_create_meta}{Learner}(model, metadata)

\method{vetiver_create_meta}{int_conformal_split}(model, metadata)

\method{vetiver_create_meta}{int_conformal_full}(model, metadata)

\method{vetiver_create_meta}{int_conformal_quantile}(model, metadata)

\method{vetiver_create_meta}{int_conformal_cv}(model, metadata)

\method{vetiver_create_meta}{ranger}(model, metadata)

\method{vetiver_create_meta}{recipe}(model, metadata)

\method{vetiver_create_meta}{model_stack}(model, metadata)

\method{vetiver_create_meta}{workflow}(model, metadata)

\method{vetiver_create_meta}{xgb.Booster}(model, metadata)
}
\arguments{
\item{model}{A trained model, such as an \code{lm()} model or a tidymodels
\code{\link[workflows:workflow]{workflows::workflow()}}.}

\item{metadata}{A list containing additional metadata to store with the pin.
When retrieving the pin, this will be stored in the \code{user} key, to
avoid potential clashes with the metadata that pins itself uses.}

\item{user}{Metadata supplied by the user}

\item{version}{Version of the pin}

\item{url}{URL for the pin, if any}

\item{required_pkgs}{Character string of R packages required for prediction}
}
\value{
The \code{vetiver_meta()} constructor returns a list. The
\code{vetiver_create_meta} function returns a \code{vetiver_meta()} list.
}
\description{
These are developer-facing functions, useful for supporting new model types.
The metadata stored in a \code{\link[=vetiver_model]{vetiver_model()}} object has four elements:
\itemize{
\item \verb{$user}, the metadata supplied by the user
\item \verb{$version}, the version of the pin (which can be \code{NULL} before pinning)
\item \verb{$url}, the URL where the pin is located, if any
\item \verb{$required_pkgs}, a character string of R packages required for prediction
}
}
\examples{
vetiver_meta()

cars_lm <- lm(mpg ~ ., data = mtcars)
vetiver_create_meta(cars_lm, list())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsconnect.R
\name{vetiver_create_rsconnect_bundle}
\alias{vetiver_create_rsconnect_bundle}
\title{Create an Posit Connect bundle for a vetiver model API}
\usage{
vetiver_create_rsconnect_bundle(
  board,
  name,
  version = NULL,
  predict_args = list(),
  filename = fs::file_temp(pattern = "bundle", ext = ".tar.gz"),
  additional_pkgs = character(0)
)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}

\item{predict_args}{A list of optional arguments passed to \code{\link[=vetiver_api]{vetiver_api()}}
such as the prediction \code{type}.}

\item{filename}{The path for the model API bundle to be created (can be
used as the argument to \code{connectapi::bundle_path()})}

\item{additional_pkgs}{Any additional R packages that need to be \strong{attached}
via \code{\link[=library]{library()}} to run your API, as a character vector.}
}
\value{
The location of the model API bundle \code{filename}, invisibly.
}
\description{
Use \code{vetiver_create_rsconnect_bundle()} to create a
\href{https://docs.posit.co/connect/}{Posit Connect} model API bundle for a
\code{\link[=vetiver_model]{vetiver_model()}} that has been versioned and stored via
\code{\link[=vetiver_pin_write]{vetiver_pin_write()}}.
}
\details{
This function creates a deployable bundle. See
\href{https://docs.posit.co/connect/cookbook/deploying/}{Posit Connect docs}
for how to deploy this bundle, as well as the
\href{https://pkgs.rstudio.com/connectapi/}{connectapi} R package for how to
integrate with Connect's API from R.

The two functions \code{vetiver_create_rsconnect_bundle()} and
\code{\link[=vetiver_deploy_rsconnect]{vetiver_deploy_rsconnect()}} are alternatives to each other, providing
different strategies for deploying a vetiver model API to Posit Connect.
}
\examples{
\dontshow{if (rlang::is_installed("connectapi") && identical(Sys.getenv("NOT_CRAN"), "true")) withAutoprint(\{ # examplesIf}
library(pins)
b <- board_temp(versioned = TRUE)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

## when you pin to Posit Connect, your pin name will be typically be like:
## "user.name/cars_linear"
vetiver_create_rsconnect_bundle(
    b,
    "cars_linear",
    predict_args = list(debug = TRUE)
)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}}, \code{\link[=vetiver_deploy_rsconnect]{vetiver_deploy_rsconnect()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{vetiver_api}
\alias{vetiver_api}
\alias{vetiver_pr_post}
\alias{vetiver_pr_docs}
\title{Create a Plumber API to predict with a deployable \code{vetiver_model()} object}
\usage{
vetiver_api(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...
)

vetiver_pr_post(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...,
  check_prototype = TRUE,
  check_ptype = deprecated()
)

vetiver_pr_docs(pr, vetiver_model, path = "/predict", all_docs = TRUE)
}
\arguments{
\item{pr}{A Plumber router, such as from \code{\link[plumber:pr]{plumber::pr()}}.}

\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{path}{The endpoint path}

\item{debug}{\code{TRUE} provides more insight into your API errors.}

\item{...}{Other arguments passed to \code{predict()}, such as prediction \code{type}}

\item{check_prototype}{Should the input data prototype stored in
\code{vetiver_model} (used for visual API documentation) also be used to check
new data at prediction time? Defaults to \code{TRUE}.}

\item{check_ptype}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}}

\item{all_docs}{Should the interactive visual API documentation be created
for \emph{all} POST endpoints in the router \code{pr}? This defaults to \code{TRUE}, and
assumes that all POST endpoints use the \code{\link[=vetiver_model]{vetiver_model()}} input data
prototype.}
}
\value{
A Plumber router with the prediction endpoint added.
}
\description{
Use \code{vetiver_api()} to add a POST endpoint for predictions from a
trained \code{\link[=vetiver_model]{vetiver_model()}} to a Plumber router.
}
\details{
You can first store and version your \code{\link[=vetiver_model]{vetiver_model()}} with
\code{\link[=vetiver_pin_write]{vetiver_pin_write()}}, and then create an API endpoint with \code{vetiver_api()}.

Setting \code{debug = TRUE} may expose any sensitive data from your model in
API errors.

Several GET endpoints will also be added to the router \code{pr}, depending on the
characteristics of the model object:
\itemize{
\item a \verb{/pin-url} endpoint to return the URL of the pinned model
\item a \verb{/metadata} endpoint to return any metadata stored with the model
\item a \verb{/ping} endpoint for the API health
\item a \verb{/prototype} endpoint for the model's input data prototype (use
\code{\link[cereal:cereal_from_json]{cereal::cereal_from_json()}}) to convert this back to a \href{https://vctrs.r-lib.org/articles/type-size.html}{vctrs ptype}
}

The function \code{vetiver_api()} uses:
\itemize{
\item \code{vetiver_pr_post()} for endpoint definition and
\item \code{vetiver_pr_docs()} to create visual API documentation
}

These modular functions are available for more advanced use cases.
}
\examples{
\dontshow{if (rlang::is_installed("plumber")) withAutoprint(\{ # examplesIf}

cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")

library(plumber)
pr() |> vetiver_api(v)
## is the same as:
pr() |> vetiver_pr_post(v) |> vetiver_pr_docs(v)
## for either, next, pipe to `pr_run()`
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{vetiver_pr_predict}
\alias{vetiver_pr_predict}
\title{Create a Plumber API to predict with a deployable \code{vetiver_model()} object}
\usage{
vetiver_pr_predict(
  pr,
  vetiver_model,
  path = "/predict",
  debug = is_interactive(),
  ...
)
}
\arguments{
\item{pr}{A Plumber router, such as from \code{\link[plumber:pr]{plumber::pr()}}.}

\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{path}{The endpoint path}

\item{debug}{\code{TRUE} provides more insight into your API errors.}

\item{...}{Other arguments passed to \code{predict()}, such as prediction \code{type}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function was deprecated to use \link{vetiver_api} directly instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handlers.R
\name{vetiver_type_convert}
\alias{vetiver_type_convert}
\title{Convert new data at prediction time using input data prototype}
\usage{
vetiver_type_convert(new_data, ptype)
}
\arguments{
\item{new_data}{New data for making predictions, such as a data frame.}

\item{ptype}{An input data prototype, such as a 0-row slice of the training
data}
}
\value{
A converted dataframe
}
\description{
This is a developer-facing function, useful for supporting new model types.
At prediction time, new observations typically must be checked and sometimes
converted to the data types from training time.
}
\examples{

library(tibble)
training_df <- tibble(x = as.Date("2021-01-01") + 0:9,
                      y = LETTERS[1:10], z = letters[11:20])
training_df

prototype <- vctrs::vec_slice(training_df, 0)
vetiver_type_convert(tibble(x = "2021-02-01", y = "J", z = "k"), prototype)

## unsuccessful conversion generates an error:
try(vetiver_type_convert(tibble(x = "potato", y = "J", z = "k"), prototype))

## error for missing column:
try(vetiver_type_convert(tibble(x = "potato", y = "J"), prototype))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{augment.vetiver_endpoint_sagemaker}
\alias{augment.vetiver_endpoint_sagemaker}
\title{Post new data to a deployed SageMaker model endpoint and augment with predictions}
\usage{
\method{augment}{vetiver_endpoint_sagemaker}(x, new_data, ...)
}
\arguments{
\item{x}{A SageMaker model endpoint object created with \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}.}

\item{new_data}{New data for making predictions, such as a data frame.}

\item{...}{Extra arguments passed to \code{\link[paws.machine.learning:sagemakerruntime_invoke_endpoint]{paws.machine.learning::sagemakerruntime_invoke_endpoint()}}}
}
\value{
The \code{new_data} with added prediction column(s).
}
\description{
Post new data to a deployed SageMaker model endpoint and augment with predictions
}
\examples{

if (FALSE) {
  endpoint <- vetiver_endpoint_sagemaker("sagemaker-demo-model")
  augment(endpoint, mtcars[4:7, -1])
}

}
\seealso{
\code{\link[=predict.vetiver_endpoint_sagemaker]{predict.vetiver_endpoint_sagemaker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-docker.R
\name{vetiver_write_docker}
\alias{vetiver_write_docker}
\title{Write a Dockerfile for a vetiver model}
\usage{
vetiver_write_docker(
  vetiver_model,
  plumber_file = "plumber.R",
  path = ".",
  ...,
  lockfile = "vetiver_renv.lock",
  rspm = TRUE,
  base_image = glue::glue("FROM rocker/r-ver:{getRversion()}"),
  port = 8000,
  expose = TRUE,
  additional_pkgs = character(0)
)
}
\arguments{
\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{plumber_file}{A path for your Plumber file, created via
\code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}}. Defaults to \code{plumber.R} in the working directory.}

\item{path}{A path to write the Dockerfile and \code{lockfile}, capturing the
model's package dependencies. Defaults to the working directory.}

\item{...}{Not currently used.}

\item{lockfile}{The generated lockfile in \code{path}. Defaults to
\code{"vetiver_renv.lock"}.}

\item{rspm}{A logical to use the
\href{https://packagemanager.rstudio.com/}{RStudio Public Package Manager} for
\code{renv::restore()} in the Docker image. Defaults to \code{TRUE}.}

\item{base_image}{The base Docker image to start with. Defaults to
\code{rocker/r-ver} for the version of R you are working with, but models like
keras will require a different base image.}

\item{port}{The server port for listening: a number such as 8080 or an
expression like \code{'as.numeric(Sys.getenv("PORT"))'} when the port is injected
as an environment variable.}

\item{expose}{Add \code{EXPOSE} to the Dockerfile? This is helpful for using
Docker Desktop but does not work with an expression for \code{port}.}

\item{additional_pkgs}{A character vector of additional package names to add
to the Docker image. For example, some boards like \code{\link[pins:board_s3]{pins::board_s3()}} require
additional software; you can use \code{required_pkgs(board)} here.}
}
\value{
The content of the Dockerfile, invisibly.
}
\description{
After creating a Plumber file with \code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}}, use
\code{vetiver_write_docker()} to create a Dockerfile plus a \code{vetiver_renv.lock}
file for a pinned \code{\link[=vetiver_model]{vetiver_model()}}.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}

library(pins)
tmp_plumber <- tempfile()
b <- board_temp(versioned = TRUE)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)
vetiver_write_plumber(b, "cars_linear", file = tmp_plumber)

## default port
vetiver_write_docker(v, tmp_plumber, tempdir())
## install more pkgs, like those required to access board
vetiver_write_docker(v, tmp_plumber, tempdir(),
                     additional_pkgs = required_pkgs(b))
## port from env variable
vetiver_write_docker(v, tmp_plumber, tempdir(),
                     port = 'as.numeric(Sys.getenv("PORT"))',
                     expose = FALSE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caret.R, R/gam.R, R/glm.R, R/keras.R,
%   R/kproto.R, R/lm.R, R/luz.R, R/mlr3.R, R/probably.R, R/prototype.R,
%   R/ranger.R, R/recipe.R, R/stacks.R, R/tidymodels.R, R/xgboost.R
\name{vetiver_ptype.train}
\alias{vetiver_ptype.train}
\alias{vetiver_ptype.gam}
\alias{vetiver_ptype.glm}
\alias{vetiver_ptype.keras.engine.training.Model}
\alias{vetiver_ptype.kproto}
\alias{vetiver_ptype.lm}
\alias{vetiver_ptype.luz_module_fitted}
\alias{vetiver_ptype.Learner}
\alias{vetiver_ptype.int_conformal_split}
\alias{vetiver_ptype.int_conformal_full}
\alias{vetiver_ptype.int_conformal_quantile}
\alias{vetiver_ptype.int_conformal_cv}
\alias{vetiver_ptype}
\alias{vetiver_ptype.default}
\alias{vetiver_create_ptype}
\alias{vetiver_ptype.ranger}
\alias{vetiver_ptype.recipe}
\alias{vetiver_ptype.model_stack}
\alias{vetiver_ptype.workflow}
\alias{vetiver_ptype.xgb.Booster}
\title{Create a vetiver input data prototype}
\usage{
\method{vetiver_ptype}{train}(model, ...)

\method{vetiver_ptype}{gam}(model, ...)

\method{vetiver_ptype}{glm}(model, ...)

\method{vetiver_ptype}{keras.engine.training.Model}(model, ...)

\method{vetiver_ptype}{kproto}(model, ...)

\method{vetiver_ptype}{lm}(model, ...)

\method{vetiver_ptype}{luz_module_fitted}(model, ...)

\method{vetiver_ptype}{Learner}(model, ...)

\method{vetiver_ptype}{int_conformal_split}(model, ...)

\method{vetiver_ptype}{int_conformal_full}(model, ...)

\method{vetiver_ptype}{int_conformal_quantile}(model, ...)

\method{vetiver_ptype}{int_conformal_cv}(model, ...)

vetiver_ptype(model, ...)

\method{vetiver_ptype}{default}(model, ...)

vetiver_create_ptype(model, save_prototype, ...)

\method{vetiver_ptype}{ranger}(model, ...)

\method{vetiver_ptype}{recipe}(model, ...)

\method{vetiver_ptype}{model_stack}(model, ...)

\method{vetiver_ptype}{workflow}(model, ...)

\method{vetiver_ptype}{xgb.Booster}(model, ...)
}
\arguments{
\item{model}{A trained model, such as an \code{lm()} model or a tidymodels
\code{\link[workflows:workflow]{workflows::workflow()}}.}

\item{...}{Other method-specific arguments passed to \code{\link[=vetiver_ptype]{vetiver_ptype()}}
to compute an input data prototype, such as \code{prototype_data} (a sample of
training features).}

\item{save_prototype}{Should an input data prototype be stored with the model?
The options are \code{TRUE} (the default, which stores a zero-row slice of the
training data), \code{FALSE} (no input data prototype for visual documentation or
checking), or a dataframe to be used for both checking at prediction time
\emph{and} examples in API visual documentation.}
}
\value{
A \code{vetiver_ptype} method returns a zero-row dataframe, and
\code{vetiver_create_ptype()} returns either such a zero-row dataframe, \code{NULL},
or the dataframe passed to \code{save_prototype}.
}
\description{
Optionally find and return an input data prototype for a model.
}
\details{
These are developer-facing functions, useful for supporting new model types.
A \code{\link[=vetiver_model]{vetiver_model()}} object optionally stores an input data prototype for
checking at prediction time.
\itemize{
\item The default for \code{save_prototype}, \code{TRUE}, finds an input data prototype (a
zero-row slice of the training data) via \code{\link[=vetiver_ptype]{vetiver_ptype()}}.
\item \code{save_prototype = FALSE} opts out of storing any input data prototype.
\item You may pass your own data to \code{save_prototype}, but be sure to check that it
has the same structure as your training data, perhaps with
\code{\link[hardhat:scream]{hardhat::scream()}}.
}
}
\examples{

cars_lm <- lm(mpg ~ cyl + disp, data = mtcars)

vetiver_create_ptype(cars_lm, TRUE)

## calls the right method for `model` via:
vetiver_ptype(cars_lm)

## can also turn off prototype
vetiver_create_ptype(cars_lm, FALSE)
\dontshow{if (rlang::is_installed("ranger")) withAutoprint(\{ # examplesIf}
## some models require that you pass in training features
cars_rf <- ranger::ranger(mpg ~ ., data = mtcars)
vetiver_ptype(cars_rf, prototype_data = mtcars[,-1])
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoint.R
\name{augment.vetiver_endpoint}
\alias{augment.vetiver_endpoint}
\title{Post new data to a deployed model API endpoint and augment with predictions}
\usage{
\method{augment}{vetiver_endpoint}(x, new_data, ...)
}
\arguments{
\item{x}{A model API endpoint object created with \code{\link[=vetiver_endpoint]{vetiver_endpoint()}}.}

\item{new_data}{New data for making predictions, such as a data frame.}

\item{...}{Extra arguments passed to \code{\link[httr:POST]{httr::POST()}}}
}
\value{
The \code{new_data} with added prediction column(s).
}
\description{
Post new data to a deployed model API endpoint and augment with predictions
}
\examples{

if (FALSE) {
endpoint <- vetiver_endpoint("http://127.0.0.1:8088/predict")
augment(endpoint, mtcars[4:7, -1])
}

}
\seealso{
\code{\link[=predict.vetiver_endpoint]{predict.vetiver_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attach-pkgs.R
\name{attach_pkgs}
\alias{attach_pkgs}
\alias{load_pkgs}
\title{Fully attach or load packages for making model predictions}
\usage{
attach_pkgs(pkgs)

load_pkgs(pkgs)
}
\arguments{
\item{pkgs}{A character vector of package names to load or fully attach.}
}
\value{
An invisible \code{TRUE}.
}
\description{
These are developer-facing functions, useful for supporting new model types.
Some models require one or more R packages to be fully attached to make
predictions, and some require only that the namespace of one or more R
packages is loaded.
}
\details{
These two functions will attempt either to:
\itemize{
\item fully attach or
\item load
}

the namespace of the \code{pkgs} vector of package names, preserving the current
random seed.

To learn more about load vs. attach, read the \href{https://r-pkgs.org/dependencies-mindset-background.html#sec-dependencies-attach-vs-load}{"Dependencies" chapter of \emph{R Packages}}.
For deploying a model, it is likely safer to fully attach needed packages
but that comes with the risk of naming conflicts between packages.
}
\examples{
## succeed
load_pkgs(c("knitr", "readr"))
attach_pkgs(c("knitr", "readr"))

## fail
try(load_pkgs(c("bloopy", "readr")))
try(attach_pkgs(c("bloopy", "readr")))

}
\concept{namespace}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoint.R
\name{vetiver_endpoint}
\alias{vetiver_endpoint}
\title{Create a model API endpoint object for prediction}
\usage{
vetiver_endpoint(url)
}
\arguments{
\item{url}{An API endpoint URL}
}
\value{
A new \code{vetiver_endpoint} object
}
\description{
This function creates a model API endpoint for prediction from a URL. No
HTTP calls are made until you actually
\code{\link[=predict.vetiver_endpoint]{predict()}} with your endpoint.
}
\examples{
vetiver_endpoint("https://colorado.rstudio.com/rsc/seattle-housing/predict")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pin-read-write.R
\name{vetiver_pin_write}
\alias{vetiver_pin_write}
\alias{vetiver_pin_read}
\title{Read and write a trained model to a board of models}
\usage{
vetiver_pin_write(board, vetiver_model, ..., check_renv = FALSE)

vetiver_pin_read(board, name, version = NULL, check_renv = FALSE)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{...}{Additional arguments passed on to methods for a specific board.}

\item{check_renv}{Use \href{https://rstudio.github.io/renv/}{renv} to record the
packages used at training time with \code{vetiver_pin_write()} and check for
differences with \code{vetiver_pin_read()}. Defaults to \code{FALSE}.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}
}
\value{
\code{vetiver_pin_read()} returns a \code{\link[=vetiver_model]{vetiver_model()}}; \code{vetiver_pin_write()}
returns the name of the new pin, invisibly.
}
\description{
Use \code{vetiver_pin_write()} to pin a trained model to a board of models,
along with an input prototype for new data and other model metadata. Use
\code{vetiver_pin_read()} to retrieve that pinned object.
}
\details{
These functions read and write a \code{\link[=vetiver_model]{vetiver_model()}} pin on the
specified \code{board} containing the model object itself and other elements
needed for prediction, such as the model's input data prototype or which
packages are needed at prediction time. You may use \code{\link[pins:pin_read]{pins::pin_read()}} or
\code{\link[pins:pin_meta]{pins::pin_meta()}} to handle the pin, but \code{vetiver_pin_read()} returns a
\code{\link[=vetiver_model]{vetiver_model()}} object ready for deployment.
}
\examples{
library(pins)
model_board <- board_temp()

cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(model_board, v)
model_board

vetiver_pin_read(model_board, "cars_linear")

# can use `version` argument to read a specific version:
pin_versions(model_board, "cars_linear")
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
# can store an renv lockfile as part of the pin:
vetiver_pin_write(model_board, v, check_renv = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{vetiver_compute_metrics}
\alias{vetiver_compute_metrics}
\title{Aggregate model metrics over time for monitoring}
\usage{
vetiver_compute_metrics(
  data,
  date_var,
  period,
  truth,
  estimate,
  ...,
  metric_set = yardstick::metrics,
  every = 1L,
  origin = NULL,
  before = 0L,
  after = 0L,
  complete = FALSE
)
}
\arguments{
\item{data}{A \code{data.frame} containing the columns specified by \code{truth},
\code{estimate}, and \code{...}.}

\item{date_var}{The column in \code{data} containing dates or date-times for
monitoring, to be aggregated with \code{.period}}

\item{period}{\verb{[character(1)]}

A string defining the period to group by. Valid inputs can be roughly
broken into:
\itemize{
\item \code{"year"}, \code{"quarter"}, \code{"month"}, \code{"week"}, \code{"day"}
\item \code{"hour"}, \code{"minute"}, \code{"second"}, \code{"millisecond"}
\item \code{"yweek"}, \code{"mweek"}
\item \code{"yday"}, \code{"mday"}
}}

\item{truth}{The column identifier for the true results (that
is \code{numeric} or \code{factor}). This should be an unquoted column name
although this argument is passed by expression and support
\link[rlang:topic-inject]{quasiquotation} (you can unquote column
names).}

\item{estimate}{The column identifier for the predicted results
(that is also \code{numeric} or \code{factor}). As with \code{truth} this can be
specified different ways but the primary method is to use an
unquoted variable name.}

\item{...}{A set of unquoted column names or one or more
\code{dplyr} selector functions to choose which variables contain the
class probabilities. If \code{truth} is binary, only 1 column should be selected,
and it should correspond to the value of \code{event_level}. Otherwise, there
should be as many columns as factor levels of \code{truth} and the ordering of
the columns should be the same as the factor levels of \code{truth}.}

\item{metric_set}{A \code{\link[yardstick:metric_set]{yardstick::metric_set()}} function for computing metrics.
Defaults to \code{\link[yardstick:metrics]{yardstick::metrics()}}.}

\item{every}{\verb{[positive integer(1)]}

The number of periods to group together.

For example, if the period was set to \code{"year"} with an every value of \code{2},
then the years 1970 and 1971 would be placed in the same group.}

\item{origin}{\verb{[Date(1) / POSIXct(1) / POSIXlt(1) / NULL]}

The reference date time value. The default when left as \code{NULL} is the
epoch time of \verb{1970-01-01 00:00:00}, \emph{in the time zone of the index}.

This is generally used to define the anchor time to count from, which is
relevant when the every value is \verb{> 1}.}

\item{before, after}{\verb{[integer(1) / Inf]}

The number of values before or after the current element to
include in the sliding window. Set to \code{Inf} to select all elements
before or after the current element. Negative values are allowed, which
allows you to "look forward" from the current element if used as the
\code{.before} value, or "look backwards" if used as \code{.after}.}

\item{complete}{\verb{[logical(1)]}

Should the function be evaluated on complete windows only? If \code{FALSE},
the default, then partial computations will be allowed.}
}
\value{
A dataframe of metrics.
}
\description{
These three functions can be used for model monitoring (such as in a
monitoring dashboard):
\itemize{
\item \code{vetiver_compute_metrics()} computes metrics (such as accuracy for a
classification model or RMSE for a regression model) at a chosen time
aggregation \code{period}
\item \code{\link[=vetiver_pin_metrics]{vetiver_pin_metrics()}} updates an existing pin storing model metrics
over time
\item \code{\link[=vetiver_plot_metrics]{vetiver_plot_metrics()}} creates a plot of metrics over time
}
}
\details{
For arguments used more than once in your monitoring dashboard,
such as \code{date_var}, consider using
\href{https://pkg.yihui.org/rmarkdown-book/parameterized-reports}{R Markdown parameters}
to reduce repetition and/or errors.
}
\examples{
\dontshow{if (rlang::is_installed(c("dplyr", "parsnip", "modeldata"))) withAutoprint(\{ # examplesIf}
library(dplyr)
library(parsnip)
data(Chicago, package = "modeldata")
Chicago <- Chicago |> select(ridership, date, all_of(stations))
training_data <- Chicago |> filter(date < "2009-01-01")
testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)

library(pins)
b <- board_temp()

original_metrics <-
    augment(lm_fit, new_data = testing_data) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)

new_metrics <-
    augment(lm_fit, new_data = monitoring) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=vetiver_pin_metrics]{vetiver_pin_metrics()}}, \code{\link[=vetiver_plot_metrics]{vetiver_plot_metrics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dashboard.R
\name{vetiver_dashboard}
\alias{vetiver_dashboard}
\alias{get_vetiver_dashboard_pins}
\alias{pin_example_kc_housing_model}
\title{R Markdown format for model monitoring dashboards}
\usage{
vetiver_dashboard(pins = list(), display_pins = TRUE, ...)

get_vetiver_dashboard_pins()

pin_example_kc_housing_model(board = pins::board_local(), name = "seattle_rf")
}
\arguments{
\item{pins}{A list containing \code{board}, \code{name}, and \code{version}, as in
\code{\link[pins:pin_read]{pins::pin_read()}}}

\item{display_pins}{Should the dashboard display a link to the pin(s)?
Defaults to \code{TRUE}, but only creates a link if the pin contains a URL in
its metadata.}

\item{...}{Arguments passed to \code{\link[flexdashboard:flex_dashboard]{flexdashboard::flex_dashboard()}}}

\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{name}{Pin name.}
}
\description{
R Markdown format for model monitoring dashboards
}
\details{
The \code{vetiver_dashboard()} function is a specialized type of
\pkg{flexdashboard}. See the flexdashboard website for additional
documentation:
\href{https://rstudio.github.io/flexdashboard/}{https://rstudio.github.io/flexdashboard/}

Before knitting the example \code{vetiver_dashboard()} template, execute the
helper function \code{pin_example_kc_housing_model()} to set up demonstration
model and metrics pins needed for the monitoring demo. This function will:
\itemize{
\item fit an example model to training data
\item pin the vetiver model to your own \code{\link[pins:board_local]{pins::board_local()}}
\item compute metrics from testing data
\item pin these metrics to the same local board
}

These are the steps you need to complete before setting up monitoring your
real model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{vetiver_sm_build}
\alias{vetiver_sm_build}
\alias{vetiver_sm_model}
\alias{vetiver_sm_endpoint}
\title{Deploy a vetiver model API to Amazon SageMaker with modular functions}
\usage{
vetiver_sm_build(
  board,
  name,
  version = NULL,
  path = fs::dir_create(tempdir(), "vetiver"),
  predict_args = list(),
  docker_args = list(),
  repository = NULL,
  compute_type = c("BUILD_GENERAL1_SMALL", "BUILD_GENERAL1_MEDIUM",
    "BUILD_GENERAL1_LARGE", "BUILD_GENERAL1_2XLARGE"),
  role = NULL,
  bucket = NULL,
  vpc_id = NULL,
  subnet_ids = list(),
  security_group_ids = list(),
  log = TRUE,
  ...
)

vetiver_sm_model(
  image_uri,
  model_name,
  role = NULL,
  vpc_config = list(),
  enable_network_isolation = FALSE,
  tags = list()
)

vetiver_sm_endpoint(
  model_name,
  instance_type,
  endpoint_name = NULL,
  initial_instance_count = 1,
  accelerator_type = NULL,
  tags = list(),
  kms_key = NULL,
  data_capture_config = list(),
  volume_size = NULL,
  model_data_download_timeout = NULL,
  wait = TRUE
)
}
\arguments{
\item{board}{An AWS S3 board created with \code{\link[pins:board_s3]{pins::board_s3()}}. This board
must be in the correct region for your SageMaker instance.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}

\item{path}{A path to write the Plumber file, Dockerfile, and lockfile,
capturing the model's dependencies.}

\item{predict_args}{A list of optional arguments passed to \code{\link[=vetiver_api]{vetiver_api()}}
such as the prediction \code{type}.}

\item{docker_args}{A list of optional arguments passed to
\code{\link[=vetiver_write_docker]{vetiver_write_docker()}} such as the \code{lockfile} name or whether to use
\code{rspm}. Do not pass \code{additional_pkgs} here, as this function uses
\code{additional_pkgs = required_pkgs(board)}.}

\item{repository}{The ECR repository and tag for the image as a character.
Defaults to \verb{sagemaker-studio-$\{domain_id\}:latest}.}

\item{compute_type}{The \href{https://aws.amazon.com/codebuild/}{CodeBuild}
compute type as a character. Defaults to \code{BUILD_GENERAL1_SMALL}.}

\item{role}{The ARN IAM role name (as a character) to be used with:
\itemize{
\item CodeBuild for \code{vetiver_sm_build()}
\item the SageMaker model for \code{vetiver_sm_model()}
}

Defaults to the SageMaker Studio execution role.}

\item{bucket}{The S3 bucket to use for sending data to CodeBuild as a
character. Defaults to the SageMaker SDK default bucket.}

\item{vpc_id}{ID of the VPC that will host the CodeBuild project such as
\code{"vpc-05c09f91d48831c8c"}.}

\item{subnet_ids}{List of subnet IDs for the CodeBuild project, such as
\code{list("subnet-0b31f1863e9d31a67")}.}

\item{security_group_ids}{List of security group IDs for the CodeBuild
project, such as \code{list("sg-0ce4ec0d0414d2ddc")}.}

\item{log}{A logical to show the logs of the running CodeBuild build.
Defaults to \code{TRUE}.}

\item{...}{\href{https://docs.docker.com/engine/reference/commandline/build/#options>}{Docker build parameters}
(Use "_" instead of "-"; for example, Docker optional parameter
\code{build-arg} becomes \code{build_arg}).}

\item{image_uri}{The AWS ECR image URI for the Amazon SageMaker Model to be
created (for example, as returned by \code{\link[=vetiver_sm_build]{vetiver_sm_build()}}).}

\item{model_name}{The Amazon SageMaker model name to be deployed.}

\item{vpc_config}{A list containing the VPC configuration for the Amazon
SageMaker model \href{https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_VpcConfig.html}{API VpcConfig}
(optional).
\itemize{
\item \code{Subnets}: List of subnet ids
\item \code{SecurityGroupIds}: List of security group ids
}}

\item{enable_network_isolation}{A logical to specify whether the container
will run in network isolation mode. Defaults to \code{FALSE}.}

\item{tags}{A named list of tags for labeling the Amazon SageMaker model or
model endpint to be created.}

\item{instance_type}{Type of EC2 instance to use; see
\href{https://aws.amazon.com/sagemaker/pricing/}{Amazon SageMaker pricing}.}

\item{endpoint_name}{The name to use for the Amazon SageMaker model endpoint
to be created, if to be different from \code{model_name}.}

\item{initial_instance_count}{The initial number of instances to run
in the endpoint.}

\item{accelerator_type}{Type of Elastic Inference accelerator to
attach to an endpoint for model loading and inference, for
example, \code{"ml.eia1.medium"}.}

\item{kms_key}{The ARN of the KMS key used to encrypt the data on the
storage volume attached to the instance hosting the endpoint.}

\item{data_capture_config}{A list for configuration to control how Amazon
SageMaker captures inference data.}

\item{volume_size}{The size, in GB, of the ML storage volume attached to
the individual inference instance associated with the production variant.
Currently only Amazon EBS gp2 storage volumes are supported.}

\item{model_data_download_timeout}{The timeout value, in seconds, to download
and extract model data from Amazon S3.}

\item{wait}{A logical for whether to wait for the endpoint to be deployed.
Defaults to \code{TRUE}.}
}
\value{
\code{vetiver_sm_build()} returns the AWS ECR image URI and
\code{vetiver_sm_model()} returns the model name (both as characters).
\code{vetiver_sm_endpoint()} returns a new \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}} object.
}
\description{
Use the function \code{\link[=vetiver_deploy_sagemaker]{vetiver_deploy_sagemaker()}} for basic deployment on
\href{https://aws.amazon.com/sagemaker/}{SageMaker}, or these three functions
together for more advanced use cases:
\itemize{
\item \code{vetiver_sm_build()} generates and builds a Docker image on SageMaker for
a vetiver model
\item \code{vetiver_sm_model()} creates an Amazon SageMaker model
\item \code{vetiver_sm_endpoint()} deploys an Amazon SageMaker model endpoint
}
}
\details{
The function \code{vetiver_sm_build()} generates the files necessary to
build a Docker container to deploy a vetiver model in SageMaker and then
builds the image on \href{https://aws.amazon.com/codebuild/}{AWS CodeBuild}. The
resulting image is stored in \href{https://aws.amazon.com/ecr/}{AWS ECR}.
This function creates a Plumber file and Dockerfile appropriate for
SageMaker, for example, with \code{path = "/invocations"} and \code{port = 8080}.

If you run into problems with Docker rate limits, then either
\itemize{
\item authenticate to Docker from SageMaker, or
\item use a \href{https://gallery.ecr.aws/docker/library/r-base}{public ECR base image},
passed through \code{docker_args}
}
}
\examples{
if (FALSE) {
library(pins)
b <- board_s3(bucket = "my-existing-bucket")
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

new_image_uri <- vetiver_sm_build(
    board = b,
    name = "cars_linear",
    predict_args = list(type = "class", debug = TRUE),
    docker_args = list(
        base_image = "FROM public.ecr.aws/docker/library/r-base:4.2.2"
    )
)

model_name <- vetiver_sm_model(
    new_image_uri,
    tags = list("my_custom_tag" = "fuel_efficiency")
)

vetiver_sm_endpoint(model_name, "ml.t2.medium")
}
}
\seealso{
\code{\link[=vetiver_prepare_docker]{vetiver_prepare_docker()}}, \code{\link[=vetiver_deploy_sagemaker]{vetiver_deploy_sagemaker()}}, \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{vetiver_pin_metrics}
\alias{vetiver_pin_metrics}
\title{Update model metrics over time for monitoring}
\usage{
vetiver_pin_metrics(
  board,
  df_metrics,
  metrics_pin_name,
  .index = .index,
  overwrite = FALSE,
  type = NULL,
  ...
)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{df_metrics}{A tidy dataframe of metrics over time, such as created by
\code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}}.}

\item{metrics_pin_name}{Pin name for where the \emph{metrics} are stored (as
opposed to where the model object is stored with \code{\link[=vetiver_pin_write]{vetiver_pin_write()}}).}

\item{.index}{The variable in \code{df_metrics} containing the aggregated dates
or date-times (from \code{time_var} in \code{data}). Defaults to \code{.index}.}

\item{overwrite}{If \code{FALSE} (the default), error when the new metrics contain
overlapping dates with the existing pin.If \code{TRUE}, overwrite any metrics for
dates that exist both in the existing pin and new metrics with the \emph{new}
values.}

\item{type}{File type used to save metrics to disk. With the default \code{NULL},
uses the \code{type} of the existing pin. Options are "rds" and "arrow".}

\item{...}{Additional arguments passed on to methods for a specific board.}
}
\value{
A dataframe of metrics.
}
\description{
These three functions can be used for model monitoring (such as in a
monitoring dashboard):
\itemize{
\item \code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}} computes metrics (such as accuracy for a
classification model or RMSE for a regression model) at a chosen time
aggregation \code{period}
\item \code{vetiver_pin_metrics()} updates an existing pin storing model metrics
over time
\item \code{\link[=vetiver_plot_metrics]{vetiver_plot_metrics()}} creates a plot of metrics over time
}
}
\details{
Sometimes when you monitor a model at a given time aggregation, you
may end up with dates in your new metrics (like \code{new_metrics} in the example)
that are the same as dates in your existing aggregated metrics (like
\code{original_metrics} in the example). This can happen if you need to re-run a
monitoring report because something failed. With \code{overwrite = FALSE} (the
default), \code{vetiver_pin_metrics()} will error when there are overlapping
dates. With \code{overwrite = TRUE}, \code{vetiver_pin_metrics()} will replace such
metrics with the new values. You probably want \code{FALSE} for interactive use
and \code{TRUE} for dashboards or reports that run on a schedule.

You can initially create your pin with \code{type = "arrow"} or the default
(\code{type = "rds"}). \code{vetiver_pin_metrics()} will update the pin using the
same \code{type} by default.
}
\examples{
\dontshow{if (rlang::is_installed(c("dplyr", "parsnip", "modeldata", "arrow")) && identical(Sys.getenv("NOT_CRAN"), "true")) withAutoprint(\{ # examplesIf}
library(dplyr)
library(parsnip)
data(Chicago, package = "modeldata")
Chicago <- Chicago |> select(ridership, date, all_of(stations))
training_data <- Chicago |> filter(date < "2009-01-01")
testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)

library(pins)
b <- board_temp()

## before starting monitoring, initiate the metrics and pin
## (for example, with the testing data):
original_metrics <-
    augment(lm_fit, new_data = testing_data) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
pin_write(b, original_metrics, "lm_fit_metrics", type = "arrow")

## to continue monitoring with new data, compute metrics and update pin:
new_metrics <-
    augment(lm_fit, new_data = monitoring) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
vetiver_pin_metrics(b, new_metrics, "lm_fit_metrics")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}}, \code{\link[=vetiver_plot_metrics]{vetiver_plot_metrics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{predict.vetiver_endpoint_sagemaker}
\alias{predict.vetiver_endpoint_sagemaker}
\title{Post new data to a deployed SageMaker model endpoint and return predictions}
\usage{
\method{predict}{vetiver_endpoint_sagemaker}(object, new_data, ...)
}
\arguments{
\item{object}{A SageMaker model endpoint object created with \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}.}

\item{new_data}{New data for making predictions, such as a data frame.}

\item{...}{Extra arguments passed to \code{\link[paws.machine.learning:sagemakerruntime_invoke_endpoint]{paws.machine.learning::sagemakerruntime_invoke_endpoint()}}}
}
\value{
A tibble of model predictions with as many rows as in \code{new_data}.
}
\description{
Post new data to a deployed SageMaker model endpoint and return predictions
}
\examples{
if (FALSE) {
  endpoint <- vetiver_endpoint_sagemaker("sagemaker-demo-model")
  predict(endpoint, mtcars[4:7, -1])
}
}
\seealso{
\code{\link[=augment.vetiver_endpoint_sagemaker]{augment.vetiver_endpoint_sagemaker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{vetiver_endpoint_sagemaker}
\alias{vetiver_endpoint_sagemaker}
\title{Create a SageMaker model API endpoint object for prediction}
\usage{
vetiver_endpoint_sagemaker(model_endpoint)
}
\arguments{
\item{model_endpoint}{The name of the Amazon SageMaker model endpoint.}
}
\value{
A new \code{vetiver_endpoint_sagemaker} object
}
\description{
This function creates a model API endpoint for prediction from a Sagemaker Model.
No HTTP calls are made until you actually
\code{\link[=predict.vetiver_endpoint_sagemaker]{predict()}} with your endpoint.
}
\examples{
\dontshow{if (rlang::is_installed("smdocker")) withAutoprint(\{ # examplesIf}
vetiver_endpoint_sagemaker("vetiver-sagemaker-demo-model")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-docker.R
\name{vetiver_prepare_docker}
\alias{vetiver_prepare_docker}
\title{Generate files necessary to build a Docker container for a vetiver model}
\usage{
vetiver_prepare_docker(
  board,
  name,
  version = NULL,
  path = ".",
  predict_args = list(),
  docker_args = list()
)
}
\arguments{
\item{board}{A pin board, created by \code{\link[pins:board_folder]{board_folder()}}, \code{\link[pins:board_connect]{board_connect()}},
\code{\link[pins:board_url]{board_url()}} or another \code{board_} function.}

\item{name}{Pin name.}

\item{version}{Retrieve a specific version of a pin. Use \code{\link[pins:pin_versions]{pin_versions()}} to
find out which versions are available and when they were created.}

\item{path}{A path to write the Plumber file, Dockerfile, and lockfile,
capturing the model's dependencies.}

\item{predict_args}{A list of optional arguments passed to \code{\link[=vetiver_api]{vetiver_api()}}
such as the prediction \code{type}.}

\item{docker_args}{A list of optional arguments passed to
\code{\link[=vetiver_write_docker]{vetiver_write_docker()}} such as the \code{lockfile} name or whether to use
\code{rspm}. Do not pass \code{additional_pkgs} here, as this function uses
\code{additional_pkgs = required_pkgs(board)}.}
}
\value{
An invisible \code{TRUE}.
}
\description{
Deploying a vetiver model via Docker requires several files. Use this
function to create these needed files in the directory located at \code{path}.
}
\details{
The function \code{vetiver_prepare_docker()} uses:
\itemize{
\item \code{\link[=vetiver_write_plumber]{vetiver_write_plumber()}} to create a Plumber file and
\item \code{\link[=vetiver_write_docker]{vetiver_write_docker()}} to create a Dockerfile and renv lockfile
}

These modular functions are available for more advanced use cases. For
models such as keras and torch, you will need to edit the generated
Dockerfile to, for example, \verb{COPY requirements.txt requirements.txt} or
similar.
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
library(pins)
b <- board_temp(versioned = TRUE)
cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
vetiver_pin_write(b, v)

vetiver_prepare_docker(b, "cars_linear", path = tempdir())
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caret.R, R/gam.R, R/glm.R, R/keras.R,
%   R/kproto.R, R/lm.R, R/luz.R, R/mlr3.R, R/prepare.R, R/probably.R,
%   R/ranger.R, R/recipe.R, R/stacks.R, R/tidymodels.R, R/xgboost.R
\name{vetiver_create_description.train}
\alias{vetiver_create_description.train}
\alias{vetiver_prepare_model.train}
\alias{vetiver_create_description.gam}
\alias{vetiver_prepare_model.gam}
\alias{vetiver_create_description.glm}
\alias{vetiver_prepare_model.glm}
\alias{vetiver_create_description.keras.engine.training.Model}
\alias{vetiver_prepare_model.keras.engine.training.Model}
\alias{vetiver_create_description.kproto}
\alias{vetiver_prepare_model.kproto}
\alias{vetiver_create_description.lm}
\alias{vetiver_prepare_model.lm}
\alias{vetiver_create_description.luz_module_fitted}
\alias{vetiver_prepare_model.luz_module_fitted}
\alias{vetiver_create_description.Learner}
\alias{vetiver_prepare_model.Learner}
\alias{vetiver_create_description}
\alias{vetiver_create_description.default}
\alias{vetiver_prepare_model}
\alias{vetiver_prepare_model.default}
\alias{vetiver_create_description.int_conformal_split}
\alias{vetiver_prepare_model.int_conformal_split}
\alias{vetiver_create_description.int_conformal_full}
\alias{vetiver_prepare_model.int_conformal_full}
\alias{vetiver_create_description.int_conformal_quantile}
\alias{vetiver_prepare_model.int_conformal_quantile}
\alias{vetiver_create_description.int_conformal_cv}
\alias{vetiver_prepare_model.int_conformal_cv}
\alias{vetiver_create_description.ranger}
\alias{vetiver_prepare_model.ranger}
\alias{vetiver_create_description.recipe}
\alias{vetiver_prepare_model.recipe}
\alias{vetiver_create_description.model_stack}
\alias{vetiver_prepare_model.model_stack}
\alias{vetiver_create_description.workflow}
\alias{vetiver_prepare_model.workflow}
\alias{vetiver_create_description.xgb.Booster}
\alias{vetiver_prepare_model.xgb.Booster}
\title{Model constructor methods}
\usage{
\method{vetiver_create_description}{train}(model)

\method{vetiver_prepare_model}{train}(model)

\method{vetiver_create_description}{gam}(model)

\method{vetiver_prepare_model}{gam}(model)

\method{vetiver_create_description}{glm}(model)

\method{vetiver_prepare_model}{glm}(model)

\method{vetiver_create_description}{keras.engine.training.Model}(model)

\method{vetiver_prepare_model}{keras.engine.training.Model}(model)

\method{vetiver_create_description}{kproto}(model)

\method{vetiver_prepare_model}{kproto}(model)

\method{vetiver_create_description}{lm}(model)

\method{vetiver_prepare_model}{lm}(model)

\method{vetiver_create_description}{luz_module_fitted}(model)

\method{vetiver_prepare_model}{luz_module_fitted}(model)

\method{vetiver_create_description}{Learner}(model)

\method{vetiver_prepare_model}{Learner}(model)

vetiver_create_description(model)

\method{vetiver_create_description}{default}(model)

vetiver_prepare_model(model)

\method{vetiver_prepare_model}{default}(model)

\method{vetiver_create_description}{int_conformal_split}(model)

\method{vetiver_prepare_model}{int_conformal_split}(model)

\method{vetiver_create_description}{int_conformal_full}(model)

\method{vetiver_prepare_model}{int_conformal_full}(model)

\method{vetiver_create_description}{int_conformal_quantile}(model)

\method{vetiver_prepare_model}{int_conformal_quantile}(model)

\method{vetiver_create_description}{int_conformal_cv}(model)

\method{vetiver_prepare_model}{int_conformal_cv}(model)

\method{vetiver_create_description}{ranger}(model)

\method{vetiver_prepare_model}{ranger}(model)

\method{vetiver_create_description}{recipe}(model)

\method{vetiver_prepare_model}{recipe}(model)

\method{vetiver_create_description}{model_stack}(model)

\method{vetiver_prepare_model}{model_stack}(model)

\method{vetiver_create_description}{workflow}(model)

\method{vetiver_prepare_model}{workflow}(model)

\method{vetiver_create_description}{xgb.Booster}(model)

\method{vetiver_prepare_model}{xgb.Booster}(model)
}
\arguments{
\item{model}{A trained model, such as an \code{lm()} model or a tidymodels
\code{\link[workflows:workflow]{workflows::workflow()}}.}
}
\description{
These are developer-facing functions, useful for supporting new model types.
Each model supported by \code{\link[=vetiver_model]{vetiver_model()}} uses up to four methods when the
deployable object is created:
\itemize{
\item The \code{vetiver_create_description()} function generates a helpful description
of the model based on its characteristics. This method is required.
\item The \code{\link[=vetiver_create_meta]{vetiver_create_meta()}} function creates the correct \code{\link[=vetiver_meta]{vetiver_meta()}}
for the model. This is especially helpful for specifying which packages are
needed for prediction. A model can use the default method here, which is
to have no special metadata.
\item The \code{\link[=vetiver_ptype]{vetiver_ptype()}} function finds an input data prototype from the
training data (a zero-row slice) to use for checking at prediction time.
This method is required.
\item The \code{vetiver_prepare_model()} function executes last. Use this function
for tasks like checking if the model is trained and reducing the size of the
model via \code{\link[butcher:butcher]{butcher::butcher()}}. A model can use the default method here,
which is to return the model without changes.
}
}
\details{
These are four generics that use the class of \code{model} for dispatch.
}
\examples{

cars_lm <- lm(mpg ~ ., data = mtcars)
vetiver_create_description(cars_lm)
vetiver_prepare_model(cars_lm)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caret.R, R/gam.R, R/glm.R, R/handlers.R,
%   R/keras.R, R/kproto.R, R/lm.R, R/luz.R, R/mlr3.R, R/probably.R, R/ranger.R,
%   R/recipe.R, R/stacks.R, R/tidymodels.R, R/xgboost.R
\name{handler_startup.train}
\alias{handler_startup.train}
\alias{handler_predict.train}
\alias{handler_startup.gam}
\alias{handler_predict.gam}
\alias{handler_predict.glm}
\alias{handler_startup}
\alias{handler_startup.default}
\alias{handler_predict}
\alias{handler_predict.default}
\alias{handler_startup.keras.engine.training.Model}
\alias{handler_predict.keras.engine.training.Model}
\alias{handler_predict.kproto}
\alias{handler_predict.lm}
\alias{handler_startup.luz_module_fitted}
\alias{handler_predict.luz_module_fitted}
\alias{handler_startup.Learner}
\alias{handler_predict.Learner}
\alias{handler_startup.int_conformal_split}
\alias{handler_predict.int_conformal_split}
\alias{handler_startup.int_conformal_full}
\alias{handler_predict.int_conformal_full}
\alias{handler_startup.int_conformal_quantile}
\alias{handler_predict.int_conformal_quantile}
\alias{handler_startup.int_conformal_cv}
\alias{handler_predict.int_conformal_cv}
\alias{handler_startup.ranger}
\alias{handler_predict.ranger}
\alias{handler_startup.recipe}
\alias{handler_predict.recipe}
\alias{handler_startup.model_stack}
\alias{handler_predict.model_stack}
\alias{handler_startup.workflow}
\alias{handler_predict.workflow}
\alias{handler_startup.xgb.Booster}
\alias{handler_predict.xgb.Booster}
\title{Model handler functions for API endpoint}
\usage{
\method{handler_startup}{train}(vetiver_model)

\method{handler_predict}{train}(vetiver_model, ...)

\method{handler_startup}{gam}(vetiver_model)

\method{handler_predict}{gam}(vetiver_model, ...)

\method{handler_predict}{glm}(vetiver_model, ...)

handler_startup(vetiver_model)

\method{handler_startup}{default}(vetiver_model)

handler_predict(vetiver_model, ...)

\method{handler_predict}{default}(vetiver_model, ...)

\method{handler_startup}{keras.engine.training.Model}(vetiver_model)

\method{handler_predict}{keras.engine.training.Model}(vetiver_model, ...)

\method{handler_predict}{kproto}(vetiver_model, ...)

\method{handler_predict}{lm}(vetiver_model, ...)

\method{handler_startup}{luz_module_fitted}(vetiver_model)

\method{handler_predict}{luz_module_fitted}(vetiver_model, ...)

\method{handler_startup}{Learner}(vetiver_model)

\method{handler_predict}{Learner}(vetiver_model, ...)

\method{handler_startup}{int_conformal_split}(vetiver_model)

\method{handler_predict}{int_conformal_split}(vetiver_model, ...)

\method{handler_startup}{int_conformal_full}(vetiver_model)

\method{handler_predict}{int_conformal_full}(vetiver_model, ...)

\method{handler_startup}{int_conformal_quantile}(vetiver_model)

\method{handler_predict}{int_conformal_quantile}(vetiver_model, ...)

\method{handler_startup}{int_conformal_cv}(vetiver_model)

\method{handler_predict}{int_conformal_cv}(vetiver_model, ...)

\method{handler_startup}{ranger}(vetiver_model)

\method{handler_predict}{ranger}(vetiver_model, ...)

\method{handler_startup}{recipe}(vetiver_model)

\method{handler_predict}{recipe}(vetiver_model, ...)

\method{handler_startup}{model_stack}(vetiver_model)

\method{handler_predict}{model_stack}(vetiver_model, ...)

\method{handler_startup}{workflow}(vetiver_model)

\method{handler_predict}{workflow}(vetiver_model, ...)

\method{handler_startup}{xgb.Booster}(vetiver_model)

\method{handler_predict}{xgb.Booster}(vetiver_model, ...)
}
\arguments{
\item{vetiver_model}{A deployable \code{\link[=vetiver_model]{vetiver_model()}} object}

\item{...}{Other arguments passed to \code{predict()}, such as prediction \code{type}}
}
\value{
A \code{handler_startup} function should return invisibly, while a
\code{handler_predict} function should return a function with the signature
\verb{function(req)}. The request body (\code{req$body}) consists of the new data
at prediction time; this function should return predictions either as a
tibble or as a list coercable to a tibble via \code{\link[tibble:as_tibble]{tibble::as_tibble()}}.
}
\description{
These are developer-facing functions, useful for supporting new model types.
Each model supported by \code{vetiver_model()} uses two handler functions
in \code{\link[=vetiver_api]{vetiver_api()}}:
\itemize{
\item The \code{handler_startup} function executes when the API starts. Use this
function for tasks like loading packages. A model can use the default
method here, which is \code{NULL} (to do nothing at startup).
\item The \code{handler_predict} function executes at each API call. Use this
function for calling \code{predict()} and any other tasks that must be executed
at each API call.
}
}
\details{
These are two generics that use the class of \code{vetiver_model$model}
for dispatch.
}
\examples{

cars_lm <- lm(mpg ~ ., data = mtcars)
v <- vetiver_model(cars_lm, "cars_linear")
handler_startup(v)
handler_predict(v)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vetiver-package.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{augment}
\alias{required_pkgs}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{generics}{\code{\link[generics:augment]{augment()}}, \code{\link[generics:required_pkgs]{required_pkgs()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sagemaker.R
\name{vetiver_sm_delete}
\alias{vetiver_sm_delete}
\title{Delete Amazon SageMaker model, endpoint, and endpoint configuration}
\usage{
vetiver_sm_delete(object, delete_model = TRUE, delete_endpoint = TRUE)
}
\arguments{
\item{object}{The model API endpoint object to be deleted, created with
\code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}.}

\item{delete_model}{Delete the SageMaker model? Defaults to \code{TRUE}.}

\item{delete_endpoint}{Delete both the endpoint and endpoint configuration?
Defaults to \code{TRUE}.}
}
\value{
\code{TRUE}, invisibly
}
\description{
Use this function to delete the Amazon SageMaker components used in a
\code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}} object. This function does \emph{not} delete
any pinned model object in S3.
}
\seealso{
\code{\link[=vetiver_deploy_sagemaker]{vetiver_deploy_sagemaker()}}, \code{\link[=vetiver_sm_build]{vetiver_sm_build()}}, \code{\link[=vetiver_endpoint_sagemaker]{vetiver_endpoint_sagemaker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{vetiver_plot_metrics}
\alias{vetiver_plot_metrics}
\title{Plot model metrics over time for monitoring}
\usage{
vetiver_plot_metrics(
  df_metrics,
  .index = .index,
  .estimate = .estimate,
  .metric = .metric,
  .n = .n
)
}
\arguments{
\item{df_metrics}{A tidy dataframe of metrics over time, such as created by
\code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}}.}

\item{.index}{The variable in \code{df_metrics} containing the aggregated dates
or date-times (from \code{time_var} in \code{data}). Defaults to \code{.index}.}

\item{.estimate}{The variable in \code{df_metrics} containing the metric estimate.
Defaults to \code{.estimate}.}

\item{.metric}{The variable in \code{df_metrics} containing the metric type.
Defaults to \code{.metric}.}

\item{.n}{The variable in \code{df_metrics} containing the number of observations
used for estimating the metric.}
}
\value{
A \code{ggplot2} object.
}
\description{
These three functions can be used for model monitoring (such as in a
monitoring dashboard):
\itemize{
\item \code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}} computes metrics (such as accuracy for a
classification model or RMSE for a regression model) at a chosen time
aggregation \code{period}
\item \code{\link[=vetiver_pin_metrics]{vetiver_pin_metrics()}} updates an existing pin storing model metrics
over time
\item \code{vetiver_plot_metrics()} creates a plot of metrics over time
}
}
\examples{
\dontshow{if (interactive() || identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
library(dplyr)
library(parsnip)
data(Chicago, package = "modeldata")
Chicago <- Chicago |> select(ridership, date, all_of(stations))
training_data <- Chicago |> filter(date < "2009-01-01")
testing_data <- Chicago |> filter(date >= "2009-01-01", date < "2011-01-01")
monitoring <- Chicago |> filter(date >= "2011-01-01", date < "2012-12-31")
lm_fit <- linear_reg() |> fit(ridership ~ ., data = training_data)

library(pins)
b <- board_temp()

## before starting monitoring, initiate the metrics and pin
## (for example, with the testing data):
original_metrics <-
    augment(lm_fit, new_data = testing_data) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
pin_write(b, original_metrics, "lm_fit_metrics", type = "arrow")

## to continue monitoring with new data, compute metrics and update pin:
new_metrics <-
    augment(lm_fit, new_data = monitoring) |>
    vetiver_compute_metrics(date, "week", ridership, .pred, every = 4L)
vetiver_pin_metrics(b, new_metrics, "lm_fit_metrics")

library(ggplot2)
vetiver_plot_metrics(new_metrics) +
    scale_size(range = c(2, 4))
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=vetiver_compute_metrics]{vetiver_compute_metrics()}}, \code{\link[=vetiver_pin_metrics]{vetiver_pin_metrics()}}
}
